//#####################################################################
// Copyright 2022.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __ANALYTIC_TESTS__
#define __ANALYTIC_TESTS__

#include <Core/Arrays/ARRAYS_ND_BASE.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Log/LOG.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>

namespace PhysBAM{

//#####################################################################
// Function Test_With_Analytic (scalar Cell-based)
//#####################################################################
template<class TV,typename T> void Test_With_Analytic(const GRID<TV>& grid,const ARRAYS_ND_BASE<T,VECTOR<int,TV::m> >& input,
    std::string equation_name,std::function<T(TV)> eq,T time,bool draw=false,int ghost=0,bool sub_avg=false)
{
    T L2_error=0;
    T total_error=0;
    T Linf_error=0;
    int n=0;
    for(CELL_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
    {
        TV X=grid.Cell_X(it.index);
        T analytic=eq(X);
        T error=analytic-input(it.index);
        if(abs(error)>Linf_error) Linf_error=abs(error);
        if(draw) Add_Ball(X,error,((error>0)?VECTOR<T,3>(1,0,0):VECTOR<T,3>(1,0,1)));
        total_error+=error;
        L2_error+=sqr(error);
        n++;
    }
    if(sub_avg)
    {
        L2_error=sqrt(max((T)0,L2_error/n-sqr(total_error/n)));
        LOG::cout<<"Avg difference ("<<equation_name<<") at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<total_error/n<<std::endl;
        Linf_error-=total_error/n;
    }
    else L2_error=sqrt(L2_error/n);
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<equation_name<<") at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<L2_error;
    LOG::cout<<"   max abs error: "<<Linf_error<<std::endl;
}
//#####################################################################
// Function Test_With_Analytic (scalar Face-based)
//#####################################################################
template<class TV,typename T> void Test_With_Analytic(const GRID<TV>& grid,const ARRAY<T,FACE_INDEX<TV::m> >& input,
    std::string equation_name,std::function<TV(TV)> eq,T time,bool draw=false,int ghost=0,bool sub_avg=false)
{
    TV L2_error;
    TV total_error;
    TV Linf_error;
    VECTOR<int,TV::m> counts;
    for(FACE_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
    {
        TV X=grid.Face_X(it.face);
        T analytic=eq(X)(it.Axis());
        T error=analytic-input(it.face);
        if(abs(error)>Linf_error(it.Axis())) Linf_error(it.Axis())=abs(error);
        if(draw)
        {
            TV v;
            v(it.Axis())=error;
            Add_Arrow(X.Add_Axis(1-it.Axis(),0.1*grid.dX(it.Axis())),v,VECTOR<T,3>(1,0,0));
        }
        total_error(it.Axis())+=error;
        L2_error(it.Axis())+=sqr(error);
        counts(it.Axis())++;
    }
    TV L2_error_uv;
    if(sub_avg)
    {
        for(int a=0;a<TV::m;a++) L2_error_uv(a)=sqrt(max(L2_error(a)/counts(a)-sqr(total_error(a)/counts(a)),(T)0));
        LOG::cout<<"Avg difference ("<<equation_name<<") at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<total_error/TV(counts)<<std::endl;
        Linf_error-=total_error/TV(counts);
    }
    else L2_error_uv=sqrt(L2_error/TV(counts));
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<equation_name<<"_x)   at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<L2_error_uv(0);
    LOG::cout<<"   max abs error: "<<Linf_error(0)<<std::endl;
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<equation_name<<"_y)   at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<L2_error_uv(1);
    LOG::cout<<"   max abs error: "<<Linf_error(1)<<std::endl;
}
template<class TV,typename T> void Test_With_Analytic(const GRID<TV>& grid,const ARRAY<T,FACE_INDEX<TV::m> >& input,
    std::string equation_name,std::function<T(TV)> eq,T time,bool draw=false,int ghost=0,bool sub_avg=false)
{
    TV L2_error;
    TV total_error;
    TV Linf_error;
    VECTOR<int,TV::m> counts;
    for(FACE_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
    {
        TV X=grid.Face_X(it.face);
        T analytic=eq(X);
        T error=analytic-input(it.face);
        if(abs(error)>Linf_error(it.Axis())) Linf_error(it.Axis())=abs(error);
        if(draw)
        {
            TV v;
            v(it.Axis())=error;
            Add_Arrow(X.Add_Axis(1-it.Axis(),0.1*grid.dX(it.Axis())),v,VECTOR<T,3>(1,0,0));
        }
        total_error(it.Axis())+=error;
        L2_error(it.Axis())+=sqr(error);
        counts(it.Axis())++;
    }
    TV L2_error_uv;
    if(sub_avg)
    {
        for(int a=0;a<TV::m;a++) L2_error_uv(a)=sqrt(max(L2_error(a)/TV(counts)(a)-sqr(total_error(a)/TV(counts)(a)),(T)0));
        LOG::cout<<"Avg difference ("<<equation_name<<") at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<total_error/TV(counts)<<std::endl;
        Linf_error-=total_error/TV(counts);
    }
    else L2_error_uv=sqrt(L2_error/TV(counts));
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<equation_name<<"_x)   at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<L2_error_uv(0);
    LOG::cout<<"   max abs error: "<<Linf_error(0)<<std::endl;
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<equation_name<<"_y)   at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<L2_error_uv(1);
    LOG::cout<<"   max abs error: "<<Linf_error(1)<<std::endl;
}
//#####################################################################
// Function Test_With_Analytic (vector Cell-based)
//#####################################################################
template<class TV,typename T> void Test_With_Analytic(const GRID<TV>& grid,const ARRAY<TV,VECTOR<int,TV::m> >& input,
    std::string equation_name,std::function<TV(TV)> eq,T time,bool draw=false,int ghost=0,bool sub_avg=false)
{
    TV L2_error;
    TV total_error;
    TV Linf_error;
    int counts=0;
    for(CELL_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
    {
        TV X=grid.Cell_X(it.index);
        TV analytic=eq(X);
        TV error=analytic-input(it.index);
        for(int axis=0;axis<TV::m;axis++)
            if(abs(error(axis))>Linf_error(axis)) Linf_error(axis)=abs(error(axis));
        if(draw) Add_Arrow(X,error,VECTOR<T,3>(1,0,0));
        total_error+=error;
        L2_error+=sqr(error);
        counts++;
    }
    TV L2_error_uv;
    if(sub_avg)
    {
        for(int a=0;a<TV::m;a++) L2_error_uv(a)=sqrt(max(L2_error(a)/TV(counts)-sqr(total_error(a)/TV(counts)),(T)0));
        LOG::cout<<"Avg difference ("<<equation_name<<") at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<total_error/TV(counts)<<std::endl;
        Linf_error-=total_error/TV(counts);
    }
    else L2_error_uv=sqrt(L2_error/TV(counts));
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<equation_name<<"_x)   at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<L2_error_uv(0);
    LOG::cout<<"   max abs error: "<<Linf_error(0)<<std::endl;
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<equation_name<<"_y)   at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<L2_error_uv(1);
    LOG::cout<<"   max abs error: "<<Linf_error(1)<<std::endl;
}
//#####################################################################
// Function Test_With_Analytic (vector of scalar Cell-based)
//#####################################################################
template<class TV,typename T> void Test_With_Analytic(const GRID<TV>& grid,const VECTOR<ARRAY<T,VECTOR<int,TV::m> >,TV::m>& input,
    std::string equation_name,std::function<TV(TV)> eq,T time,bool draw=false,int ghost=0,bool sub_avg=false)
{
    TV L2_error;
    TV total_error;
    TV Linf_error;
    int counts=0;
    for(CELL_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
    {
        TV X=grid.Cell_X(it.index);
        TV analytic=eq(X);
        TV error;
        for(int axis=0;axis<TV::m;axis++)
        {
            error(axis)=analytic(axis)-input(axis)(it.index);
            if(abs(error(axis))>Linf_error(axis)) Linf_error(axis)=abs(error(axis));
        }
        if(draw) Add_Arrow(X,error,VECTOR<T,3>(1,0,0));
        total_error+=error;
        L2_error+=sqr(error);
        counts++;
    }
    TV L2_error_uv;
    if(sub_avg)
    {
        for(int a=0;a<TV::m;a++) L2_error_uv(a)=sqrt(max(L2_error(a)/counts-sqr(total_error(a)/counts),(T)0));
        LOG::cout<<"Avg difference ("<<equation_name<<") at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<total_error/counts<<std::endl;
        Linf_error-=total_error/counts;
    }
    else L2_error_uv=sqrt(L2_error/counts);
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<equation_name<<"_x)   at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<L2_error_uv(0);
    LOG::cout<<"   max abs error: "<<Linf_error(0)<<std::endl;
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<equation_name<<"_y)   at time "<<std::left<<std::setw(6)<<time<<" : "<<std::setw(16)<<L2_error_uv(1);
    LOG::cout<<"   max abs error: "<<Linf_error(1)<<std::endl;
}

}
//#####################################################################
#endif
