//#####################################################################
// Copyright 2022
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAYS_ND_BASE.h>
#include <Core/Arrays/ARRAYS_ND_VIEW.h>
#include <Core/Boundaries/REFLECTION_HELPER.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Interpolation/INTERPOLATED_COLOR_MAP.h>
#include <Core/Interpolation/WENO_INTERPOLATION_CELL.h>
#include <Core/Level_Sets/LEVELSET.h>
#include <Core/Log/DEBUG_SUBSTEPS.h>
#include <Core/Log/LOG.h>
#include <Core/Log/SCOPE.h>
//#include <Core/Log/TEST_UNITS.h>
#include <Core/Ordinary_Differential_Equations/RUNGEKUTTA.h>
#include <Core/Poisson/HELMHOLTZ.h>
#include "CAHN_HILLIARD.h"
#include "CH_SYMBOLIC.h"

using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> CAHN_HILLIARD<TV>::
CAHN_HILLIARD(const GRID<TV>& g,int ghost,PARSE_ARGS& parse_args,T kg,T s,T m)
    :mass_scale(kg),time_scale(s),distance_scale(m),ghost(ghost),grid(g)
{
    // mass scaling (unit tests) (constructor input)
    // time scaling (unit tests) (constructor input)
    // distance scaling (unit tests) (constructor input)
    
    // discrete_advection (outside input)
    // test_units (outside input)
    // test_flux (outside input)
    // require_convergence (outside input)

    bool print_all=false;
    
    parse_args.Add("-ch_linear",&linearize,"make some portions of the CH equation linear for testing purposes");
    parse_args.Add("-ch_sigma",&sigma,"sigma","surface tension coefficient");
    parse_args.Add("-ch_eta",&eta,"eta","fixed phase interface width");
    parse_args.Add("-ch_M",&M,"mobility","mobility constant");
    parse_args.Add_Not("-ch_no_status",&print_status,"do not print CH status messages");
    parse_args.Add("-ch_debug",&debug_output,"print CH debug messages");
    parse_args.Add_Not("-ch_no_stats",&print_gamma_stats,"print gamma stats");
    parse_args.Add("-ch_solve_residual",&print_solve_residual,"print residuals for the CH solves");
    parse_args.Add_Not("-ch_no_solve_results",&print_solve_results,"do not print results for the CH solves");
    parse_args.Add("-ch_print_all",&print_all,"print all stats and debug outputs");
    parse_args.Add("-ch_rel_error",&relative_error,"compute analytic error as relative");
    parse_args.Parse(true);

    if(print_all)
    {
        print_gamma_stats=true;
        print_status=true;
        print_solve_results=true;
        print_solve_residual=true;
        debug_output=true;
    }

    T volume_scale=pow(m,TV::m);
    T st_scale=kg*pow(m,(T)3)/(volume_scale*s*s);
    T mobility_scale=s*volume_scale/kg;
    if(!eta) eta=0.01; 
    eta*=distance_scale;
    if(!sigma) sigma=7.28e-2;
    sigma*=st_scale;
    lambda=(3/(2*sqrt(2)))*(sigma*eta); // always calculate lambda from eta and sigma
    if(!M) M=(1e-7/lambda)*st_scale*distance_scale;
    M*=mobility_scale;
    
    periodic=VECTOR<bool,TV::m>();
    periodic_boundary_scalar.is_periodic=periodic;
    side_contact_angle.Fill(pi/2);
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> CAHN_HILLIARD<TV>::
~CAHN_HILLIARD()
{
    if(symbolic) delete symbolic;
}
//#####################################################################
// Function Fill_Ghost (scalar cells)
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Fill_Ghost(ARRAYS_ND_BASE<T,IV>& output,const ARRAYS_ND_BASE<T,IV>& input,int ghost,std::function<T(const IV&,int)> boundary_N) const 
{
     // TODO: add support for more boundary types?
    ARRAYS_ND_BASE<T,IV>::Put(input,output); // interior
    Reflection_Cell_Neumann<TV,T>(grid,output,ghost,boundary_N); // Do this first, then periodic based on periodic_v
    periodic_boundary_scalar.Fill_Ghost_Cells(grid,input,output,ghost);
}
//#####################################################################
// Function Discrete_Advection
// Z needs 3 ghost cells
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Discrete_Advection(const GRID<TV>& grid,ARRAY<T,IV>& out,const ARRAY<T,IV>& Z,const ARRAY<T,FACE_INDEX<TV::m> >& u,T dt,T time,ARRAY<T,FACE_INDEX<TV::m> >* flux)
{
    WENO_INTERPOLATION_CELL<TV,T> interp;
    ARRAY<T,FACE_INDEX<TV::m> > weno_Z(grid.Cell_Indices());
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        weno_Z(it.face)=interp.Interpolate_Cell(grid,Z,grid.Face_X(it.face));
        if(flux) (*flux)(it.face)=weno_Z(it.face)*u(it.face);
    }
    
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        T div=0;
        for(int axis=0;axis<TV::m;axis++)
        {
            FACE_INDEX<TV::m> first=it.Full_First_Face_Index(axis); // same index as it.index
            FACE_INDEX<TV::m> second=it.Full_Second_Face_Index(axis); // larger index
            div+=(u(second)*weno_Z(second)-u(first)*weno_Z(first))/grid.dX(axis);
        }
        out(it.index)-=dt*div;        
    }
}
//#####################################################################
// Function Advect
// g_ghost, u_ghost need ghost of 3 for conservative advection
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Advect(ARRAY<T,IV>& output,const ARRAY<T,IV>& g_ghost,const ARRAY<T,FACE_INDEX<TV::m> >& u_ghost,T dt,T time)
{
    ARRAY<T,IV>* gamma_preadvect=nullptr;
    if(test_flux) gamma_preadvect=new ARRAY<T,IV>(output);
    advect_flux.Resize(u_ghost.domain);
    if(discrete_advection) Discrete_Advection(grid,output,g_ghost,u_ghost,dt,time,&advect_flux);
    else
    {
        advection_scalar.Set_Epsilon((1e-6)*sqr(distance_scale/time_scale)); // Unit Testing
        advection_scalar.Update_Advection_Equation_Cell(grid,output,g_ghost,u_ghost,dt,time,0,0,0,0,&advect_flux);
    }
    if(test_flux)
    {
        Test_Flux(grid,*gamma_preadvect,output,advect_flux,-dt,"gamma advection flux");
        delete gamma_preadvect;
    }
    Test_Analytic(output,"gamma_star",time);
    Print_Gamma_Stats(output,"after advection");
}
//#####################################################################
// Function Initialize_Gamma
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Initialize_Gamma(const GRID<TV>& grid,ARRAYS_ND_BASE<T,IV>& gamma,int ghost,T dt,T time) const
{
    if(symbolic) symbolic->Set_Step(time,dt);
    for(CELL_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
    {
        if(symbolic) gamma(it.index)=symbolic->Get_T("gamma",grid.Cell_X(it.index));
        else gamma(it.index)=0;
    }
}
//#####################################################################
// Function Gradient
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Gradient(ARRAY<T,FACE_INDEX<TV::m> >& out,const ARRAYS_ND_BASE<T,IV>& in_ghost) const
{
    for(FACE_RANGE_ITERATOR<TV::m> it(out.domain);it.Valid();it.Next())
        out(it.face)=(in_ghost(it.face.Second_Cell_Index())-in_ghost(it.face.First_Cell_Index()))*grid.one_over_dX(it.face.axis);
}
//#####################################################################
// Function Laplacian
// Requires ghost data for input to at least 1+ghost cells
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Laplacian(ARRAY<T,IV>& out,const ARRAYS_ND_BASE<T,IV>& in_ghost,int ghost) const
{
    for(CELL_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
    {
        T lap=0;
        for(int a=0;a<TV::m;a++)
        {
            T idx2=sqr(grid.one_over_dX(a));
            IV left=it.index.Add_Axis(a,-1);
            IV right=it.index.Add_Axis(a,1);
            
            lap+=idx2*(in_ghost(right)-2*in_ghost(it.index)+in_ghost(left));
        }
        out(it.index)=lap;
    }
}
//#####################################################################
// Function Set_Periodic_Boundaries
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Set_Periodic_Boundaries(const VECTOR<bool,TV::m>& p)
{
    periodic=p;
    periodic_boundary_scalar.is_periodic=periodic;
}
//#####################################################################
// Function Set_Neumann_Boundaries
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Set_Neumann_Boundaries(const ARRAY<bool,FACE_INDEX<TV::m> >& N, const ARRAY<T,FACE_INDEX<TV::m> >* N_values)
{
    psi_N=&N;
    psi_N_values=N_values;
}
//#####################################################################
// Function Set_Dirichlet_Boundaries
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Set_Dirichlet_Boundaries(const ARRAYS_ND_BASE<bool,IV>& D, const ARRAYS_ND_BASE<T,IV>* D_values)
{
    psi_D=&D;
    psi_D_values=D_values;
}
//#####################################################################
// Function Print_Gamma_Stats
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Print_Gamma_Stats(const ARRAYS_ND_BASE<T,IV>& gamma,std::string title) const
{
    if(!print_gamma_stats) return;
    T min=1e20,max=0;
    long double sum=0;
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        T g=gamma(it.index);
        if(g<min)min=g;
        if(g>max)max=g;
        sum+=g;
    }
    LOG::printf("gamma (%P): max=%.16P min=%.16P sum=%.20P\n",title,max,min,sum);
}
//#####################################################################
// Function Test_Flux
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Test_Flux(const GRID<TV>& flux_grid,const ARRAYS_ND_BASE<T,IV>& base,const ARRAYS_ND_BASE<T,IV>& control,const ARRAY<T,FACE_INDEX<TV::m> >& flux,T scale,const std::string title) const
{
    if(!test_flux) return;
    T total_error=0;
    T max_error=0;
    int count=0;
    for(CELL_ITERATOR<TV> it(flux_grid);it.Valid();it.Next())
    {
        T test_value=base(it.index);

        for(int axis=0;axis<TV::m;axis++)
            test_value+=scale*grid.one_over_dX(axis)*(flux(it.Full_Second_Face_Index(axis))-flux(it.Full_First_Face_Index(axis)));
        
        T error=test_value-control(it.index);
        if(abs(error)>max_error) max_error=abs(error);
        total_error+=sqr(error);
        count++;
    }
    LOG::printf("Flux (%P) \t L2: %P \t Linf: %P\n",title,sqrt(total_error/count),max_error);

}
//#####################################################################
// Function Test_Analytic
//#####################################################################
template<class TV>  void CAHN_HILLIARD<TV>::
Test_Analytic(const ARRAYS_ND_BASE<T,IV>& data,const char* name,T t,int g,bool relative)
{
    if(!error_particles||!viewer_output||!debug_particles||!symbolic) return;
    auto [ptr,ins]=error_particles->Insert(name,nullptr);
    if(ins)
    {
        *ptr=new DEBUG_PARTICLES<TV>;
        viewer_output->Add(std::string("error_")+name,**ptr,false,{.draw=false});
    }
    DEBUG_PARTICLES<TV>::Get(*ptr);
    auto eq=[this,name](TV X){return symbolic->Get_T(name,X);};
    
    // from Test_With_Analytic (scalar Cell-based)
    bool draw=!suppress_output&&error_output;
    T L2_error=0;
    T total_error=0;
    T Linf_error=0;
    T L2_analytic=0;
    T Linf_analytic=0;    
    int n=0;
    for(CELL_ITERATOR<TV> it(grid,g);it.Valid();it.Next())
    {
        TV X=grid.Cell_X(it.index);
        T analytic=eq(X);
        T error=data(it.index)-analytic;
        Linf_analytic=max(Linf_analytic,abs(analytic));
        Linf_error=max(Linf_error,abs(error));
        if(draw) Add_Ball(X,error,((error>0)?VECTOR<T,3>(1,0,0):VECTOR<T,3>(1,0,1)));
        total_error+=error;
        L2_error+=sqr(error);
        L2_analytic+=sqr(analytic);
        n++;
    }
    L2_error=sqrt(L2_error/n);
    L2_analytic=sqrt(L2_analytic/n);
    if(!relative||!L2_analytic) L2_analytic=1;
    if(!relative||!Linf_analytic) Linf_analytic=1;
    LOG::cout<<"L2 "<<grid.numbers_of_cells(0)<<" ("<<name<<((relative&&(L2_analytic!=1))?" [rel]":"")<<") at time "<<std::left<<std::setw(6)<<t<<" : "<<std::setw(16)<<L2_error/L2_analytic;
    LOG::cout<<"   max abs error: "<<Linf_error/Linf_analytic<<std::endl;
    
    DEBUG_PARTICLES<TV>::Get(debug_particles);
}
//#####################################################################
// Function Set_Analytic_Output
//#####################################################################
template<class TV> void CAHN_HILLIARD<TV>::
Set_Analytic_Output(bool print_output,bool draw_error,VIEWER_OUTPUT* vo,DEBUG_PARTICLES<TV>* p,HASHTABLE<std::string,DEBUG_PARTICLES<TV>*>* error_p)
{
    suppress_output=!print_output;
    error_output=draw_error;
    viewer_output=vo;
    debug_particles=p;
    error_particles=error_p;
}
namespace PhysBAM{
template class CAHN_HILLIARD<VECTOR<float,2> >;
template class CAHN_HILLIARD<VECTOR<float,3> >;
template class CAHN_HILLIARD<VECTOR<double,2> >;
template class CAHN_HILLIARD<VECTOR<double,3> >;
}
