//#####################################################################
// Copyright 2022
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAYS_ND_BASE.h>
#include <Core/Arrays/ARRAYS_ND_VIEW.h>
#include <Core/Boundaries/REFLECTION_HELPER.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Interpolation/INTERPOLATED_COLOR_MAP.h>
#include <Core/Interpolation/WENO_INTERPOLATION_CELL.h>
#include <Core/Level_Sets/LEVELSET.h>
#include <Core/Log/DEBUG_SUBSTEPS.h>
#include <Core/Log/LOG.h>
#include <Core/Log/SCOPE.h>
//#include <Core/Log/TEST_UNITS.h>
#include <Core/Ordinary_Differential_Equations/RUNGEKUTTA.h>
#include <Core/Poisson/HELMHOLTZ.h>
#include "CAHN_HILLIARD_HELMHOLTZ.h"
#include "CHH_SYMBOLIC.h"

using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> CAHN_HILLIARD_HELMHOLTZ<TV>::
CAHN_HILLIARD_HELMHOLTZ(const GRID<TV>& g,int ghost,PARSE_ARGS& parse_args,T kg,T s,T m)
    : CAHN_HILLIARD<TV>(g,ghost,parse_args,kg,s,m)
{

    parse_args.Add("-ch_h_tol",&ch_rel_tolerance,"error tolerance","Relative tolerance for Cahn-Hilliard solves");
    parse_args.Add("-ch_h_abs_tol",&ch_abs_tolerance,"error tolerance","Absolute tolerance for Cahn-Hilliard solves");
    parse_args.Add("-ch_b_tol",&b_rel_tolerance,"error tolerance","Relative tolerance for bounded gamma flux solve");
    parse_args.Add("-ch_b_abs_tol",&b_abs_tolerance,"error tolerance","Absolute tolerance for bounded gamma flux solve");
    parse_args.Add_Not("-ch_unbounded",&bounded,"do not bound gamma between 0 and 1");
    parse_args.Add("-ch_bound_limit",&bound_limit,"double","threshold at which to apply gamma bounding");
    parse_args.Add("-ch_scale_diagonal",&scale_diagonal,"diagonally scale redistribution solve before preconditioner");
    
    parse_args.Add("-ch_write_system",&write_ch_system,"write ch system matrices in octave format");
    parse_args.Add("-ch_bound_write_system",&write_bound_system,"write bound system matrices in octave format");
    parse_args.Add("-ch_output_intermediates",&output_intermediates,"output intermediates to viewer");
    parse_args.Parse(true);

    compute_bound_flux=bounded;
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> CAHN_HILLIARD_HELMHOLTZ<TV>::
~CAHN_HILLIARD_HELMHOLTZ()
{
    // if(symbolic) delete symbolic;
}
//#####################################################################
// Function Initialize_Symbolics
//#####################################################################
template<class TV> void CAHN_HILLIARD_HELMHOLTZ<TV>::
Initialize_Symbolics(std::string code)
{
    symbolic=new CHH_SYMBOLIC<TV>(code,linearize);
    symbolic->lambda=lambda;
    symbolic->eta=eta;
    symbolic->M=M;
    // symbolic->second_order=1; // How to determine?
    symbolic->sigma=sigma;
    symbolic->mass_scale=BASE::mass_scale;
    symbolic->time_scale=BASE::time_scale;
    symbolic->distance_scale=BASE::distance_scale;
    symbolic->speed_scale=BASE::distance_scale/BASE::time_scale;
}
//#####################################################################
// Function Cahn_Hilliard
//#####################################################################
template<class TV> void CAHN_HILLIARD_HELMHOLTZ<TV>::
Cahn_Hilliard(T dt,T time,ARRAYS_ND_BASE<T,IV>& gamma,ARRAYS_ND_BASE<T,IV>* gamma_prev,ARRAY<T,FACE_INDEX<TV::m> >* face_velocities_ab)
{
    RANGE<IV> gamma_size=gamma.domain;
    ARRAY<T,IV> gamma_bd(gamma_size);
    ARRAY<T,IV> gamma_ab(gamma_size);

    Print_Gamma_Stats(gamma,"before prepare");
    if(gamma_prev)
    {
        order=1.5;
        gamma_bd.array=(2*gamma.array-gamma_prev->array/2)/order; // 1.5*g2-2*g1+0.5*g0
        gamma_ab.array=2*gamma.array-gamma_prev->array; // 2*g1+g0
        if(symbolic) symbolic->second_order=1;
    }
    else
    {
        order=1;
        gamma_bd.array=gamma.array;
        gamma_ab.array=gamma.array;
        if(symbolic) symbolic->second_order=0;
    }

    if(symbolic) symbolic->Set_Step(time,dt*order);
    
    RANGE<IV> valid_range(gamma_bd.domain.Intersect(grid.Cell_Indices()));
    flux.Resize(valid_range);
    Test_Analytic(gamma_bd,"gamma_bd",time);
    Test_Analytic(gamma_ab,"gamma_ab",time);

    if(gamma_prev) gamma_prev->array=gamma.array;

    if(face_velocities_ab) Advect(gamma_bd,gamma_ab,*face_velocities_ab,dt,time);
    
    auto& gamma_star=gamma_bd; // Post advection
    ARRAY<T,IV>* gamma_presolve=nullptr;
    if(test_flux) gamma_presolve=new ARRAY<T,IV>(gamma_star);
    
    T eta2=eta*eta;
    T ML=M*lambda;
    T s=sqrt(ML/dt);
    T a=ML/(s*eta2);
    if(debug_output) LOG::cout<<"a="<<a<<" s="<<s<<" eta2="<<eta2<<" ML="<<ML<<std::endl;

    ARRAY<T,IV> A(gamma_size);
    if(linearize) A.array=(a-2)*gamma_ab.array;
    else
    {
        A.array=a*(4*gamma_ab.array*gamma_ab.array-6*gamma_ab.array+2)*gamma_ab.array-2*gamma_ab.array;
    }
    Test_Analytic(A,"gamma_A",time);
    ARRAY<T,IV> B(gamma_size),&D=B,C(gamma_size),gamma2(gamma_size); // B and D can share space
    B.array=gamma_star.array+A.array;
    if(volume) B.array*=volume->array;
   
    if(debug_output) LOG::cout<<"gamma_B sum="<<B.Sum()<<std::endl;
    if(symbolic)
        for(RANGE_ITERATOR<TV::m> it(gamma_size,0,0);it.Valid();it.Next())
            B(it.index)+=dt*symbolic->Get_T("src_ch",grid.Cell_X(it.index));
    Test_Analytic(B,"gamma_B",time);
    if(debug_output) LOG::cout<<"gamma_B sum="<<B.Sum()<<std::endl;

    if(print_status) LOG::cout<<"Solving first CH equation..."<<std::endl;
    Solve_Separated_CH_Equation(C,B,dt*s,gamma_ab,[this,s](TV X,int side,T g_ab){
        TV ret;
        if(symbolic) ret=(side%2?1:-1)*((M/s)*symbolic->Get_TV("grad_xi",X)-symbolic->Get_TV("grad_gamma2",X));
        else ret=TV()-0.5*(sigma/lambda)*cos(side_contact_angle(side))*(pi/2)*cos(pi*g_ab-pi/2);
        return ret;},true);
    Test_Analytic(C,"gamma_C",time);
    if(debug_output) LOG::cout<<"gamma_C sum="<<C.Sum()<<std::endl;

    D.array=C.array-A.array;
    if(volume) D.array*=volume->array;
    Test_Analytic(D,"gamma_D",time);
    if(debug_output) LOG::cout<<"gamma_D sum="<<D.Sum()<<std::endl;
    
    if(print_status) LOG::cout<<"Solving second CH equation..."<<std::endl;
    Solve_Separated_CH_Equation(gamma2,D,dt*s,gamma_ab,[this,s](TV X,int side,T g_ab){
        if(!symbolic) return TV()+0.5*(sigma/lambda)*cos(side_contact_angle(side))*(pi/2)*cos(pi*g_ab-pi/2);
        else return (side%2?1:-1)*symbolic->Get_TV("grad_gamma2",X);},true);

    // Compute Flux
    if(compute_flux)
    {
        ARRAY<T,FACE_INDEX<TV::m> > grad_C,grad_gamma2;
        PHYSBAM_ASSERT(gamma_size.Contains(grid.Cell_Indices(ghost)));

        RANGE<IV> valid_range(gamma_size.Intersect(grid.Cell_Indices()));
        grad_C.Resize(valid_range);
        grad_gamma2.Resize(valid_range);
        Fill_Ghost(C,C,ghost,[this,&gamma_ab](const IV& index,int side){
            if(!symbolic)
            {
                int axis=side/2;
                int sign=(side%2*2)-1;
                IV a=Fine_Node_To_Cell(index.Add_Axis(axis,-sign));
                T g_ab=gamma_ab(a);
                return (T)(-0.5*(sigma/lambda)*cos(side_contact_angle(side))*(pi/2)*cos(pi*g_ab-pi/2));
            }
            else return (side%2?1:-1)*symbolic->Get_TV("grad_gamma_C",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
        });
        Gradient(grad_C,C);
        Fill_Ghost(gamma2,gamma2,ghost,[this,&gamma_ab](const IV& index,int side){
            if(!symbolic)
            {
                int axis=side/2;
                int sign=(side%2*2)-1;
                IV a=Fine_Node_To_Cell(index.Add_Axis(axis,-sign));
                T g_ab=gamma_ab(a);
                return (T)(0.5*(sigma/lambda)*cos(side_contact_angle(side))*(pi/2)*cos(pi*g_ab-pi/2));
            }
            else return (side%2?1:-1)*symbolic->Get_TV("grad_gamma2",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
        });
        Gradient(grad_gamma2,gamma2);

        for(FACE_RANGE_ITERATOR<TV::m> it(flux.domain);it.Valid();it.Next())
            flux(it.face)=s*(grad_C(it.face)+grad_gamma2(it.face));
    }
    if(test_flux)
    {
        if(symbolic)
            for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                (*gamma_presolve)(it.index)+=dt*symbolic->Get_T("src_ch",grid.Cell_X(it.index));
        Test_Flux(grid,*gamma_presolve,gamma2,flux,dt,"gamma solve flux");
        delete gamma_presolve;
    }
    
    if(gamma_prev) gamma_prev->array=gamma.array; // keep old data
    gamma.array=gamma2.array;

    // Bound_Gamma here
    if(bounded)
    {
        ARRAY<T,IV> gamma_prebound=gamma2;
        Bound_Gamma(gamma,gamma_ab,dt,bound_limit);
        Test_Flux(grid,gamma_prebound,gamma,bound_flux,dt,"bound flux");
    }
    Test_Analytic(gamma,"gamma2",time);
    Print_Gamma_Stats(gamma,"after ch solve");    
}
//#####################################################################
// Function Register_For_Viewing
//#####################################################################
template<class TV> void CAHN_HILLIARD_HELMHOLTZ<TV>::
Register_For_Viewing(VIEWER_OUTPUT &vo)
{
    if(!output_intermediates) return;
    return;
}
//#####################################################################
// Function Bound_Gamma
//#####################################################################
template<class TV> void CAHN_HILLIARD_HELMHOLTZ<TV>::
Bound_Gamma(ARRAYS_ND_BASE<T,IV>& gamma,const ARRAYS_ND_BASE<T,IV>& gamma_ab,T dt,T limit)
{
    // Bound gamma between 0 and 1
    T max_gamma=0;
    T min_gamma=1;
    T gamma_sum=0;
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        T g=gamma(it.index);
        gamma_sum+=g;
        max_gamma=max(max_gamma,g);
        min_gamma=min(min_gamma,g);
    }
    if(debug_output) LOG::printf("Max/min gamma before bounded fix: %P %P \n",max_gamma,min_gamma);
    if(debug_output) LOG::printf("gamma sum before bounded fix: %P \n",gamma_sum);
    bound_flux.Resize(grid.Cell_Indices(),init_new,(T)0);
    if(((max_gamma-1)<=limit)&&(min_gamma>=-limit))
    {
        bound_flux.Fill(0);
        return;
    }

    ARRAY<T,IV> gamma_b_star(gamma.domain);
    ARRAY<T,IV> gamma2(gamma.domain);
    gamma2.array=gamma.array;
    auto weight=[](T g){ return 4*g-4*g*g; };
    while(((max_gamma-1)>limit||min_gamma<-limit))
    {
        T cell_volume=grid.Cell_Size();
        ARRAY<T,IV> Wb(gamma.domain);
        T Wbv_sum=0;
        T over_sum=0;

        for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            T g=gamma(it.index);
            T gb=clamp(g,(T)0,(T)1);
            gamma_b_star(it.index)=gb;
            T w=weight(gb);
            Wb(it.index)=w;
            Wbv_sum+=w;
            over_sum+=(g-gb);
        }
        Wbv_sum*=cell_volume;
        over_sum*=cell_volume;
        
        if(Wbv_sum)
        {
            for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                gamma(it.index)=gamma_b_star(it.index)+Wb(it.index)*over_sum/Wbv_sum;
            // gamma.array=gamma_b_star.array+Wb.array*over_sum/Wbv_sum;
        }

        T new_max_gamma=0;
        T new_min_gamma=1;
        gamma_sum=0;
        for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            T g=gamma(it.index);
            gamma_sum+=g;
            new_max_gamma=max(new_max_gamma,g);
            new_min_gamma=min(new_min_gamma,g);
        }
        if(debug_output) LOG::printf("Max/min gamma after bounded fix: %P %P \n",new_max_gamma,new_min_gamma);
        if(debug_output) LOG::printf("gamma sum after bounded fix: %P \n",gamma_sum);
        if(new_max_gamma<max_gamma||new_min_gamma>min_gamma)
        {
            max_gamma=min(max_gamma,new_max_gamma);
            min_gamma=max(min_gamma,new_min_gamma);
        }
        else
        {
            LOG::printf("WARNING: bounded fix made no improvement. Stopping loop.\n");
            break;
        }
    }
    if(!compute_bound_flux) return;

    ARRAY<T,IV> Wbc(gamma.domain);
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        T g=gamma(it.index);
        Wbc(it.index)=max((T)1e-6,weight(g));
    }
    //Fill ghost of Wbc for interpolation
    Fill_Ghost(Wbc,Wbc,ghost,[this,&gamma_ab](const IV& index,int side){
        if(!symbolic) return (T)0;
        else return (side%2?1:-1)*symbolic->Get_TV("grad_wbc",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
    });
    
    ARRAY<T,FACE_INDEX<TV::m> > Wbf(gamma.domain);
    AVERAGING_UNIFORM<TV> avg;
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        T wb=avg.Cell_To_Face(grid,it.face,Wbc);
        Wbf(it.face)=dt*wb;
    }

    HELMHOLTZ<TV> helmholtz(grid,false);
    helmholtz.print_system=write_bound_system;
    helmholtz.octave_output="bound-system";
    helmholtz.diagonal_scaling=scale_diagonal;
    
    PCG_OPTIONS<T> options;
    options.max_iterations=10000;
    options.restart_iterations=200;
    options.show_results=print_solve_results;
    options.show_residual=print_solve_residual;
    options.ensure_convergence=true;
    options.absolute_tolerance=b_abs_tolerance;
    options.relative_tolerance=b_rel_tolerance;
    
    ARRAY<bool,FACE_INDEX<TV::m> > psi_N(grid.Cell_Indices(),init_new,false);
    
    // Setting outer boundaries to true if not periodic
    ARRAY<T,FACE_INDEX<TV::m> > psi_N_values(grid.Cell_Indices(),init_new,(T)0);
    for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION);it.Valid();it.Next())
    {
        if(periodic(it.face.axis)) continue; 
        psi_N(it.face)=true;
    }
    helmholtz.Set_Neumann_Boundaries(psi_N,&psi_N_values);
    
    ARRAY<T,IV> rhs(grid.Cell_Indices(ghost));
    ARRAY<T,IV> Q(grid.Cell_Indices(ghost));
    for(CELL_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
    {
        rhs(it.index)=-(gamma(it.index)-gamma2(it.index));
    }
    helmholtz.Set_Alpha(0);
    helmholtz.Set_Beta(Wbf);
    helmholtz.Set_Periodic_Boundaries(periodic);
    if(print_status) LOG::cout<<"Solving bound flux equation..."<<std::endl;    
    bool converged=helmholtz.Solve(Q,rhs,grid.Cell_Indices(),options);
    if(require_convergence) PHYSBAM_ASSERT(converged);

    //Fill ghost of Q for gradient
    Fill_Ghost(Q,Q,ghost,[this,&gamma_ab](const IV& index,int side){
        if(!symbolic) return (T)0;
        else return (side%2?1:-1)*symbolic->Get_TV("grad_Q",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
    });

    ARRAY<T,FACE_INDEX<TV::m> > Q_grad(grid.Cell_Indices(),init_new,(T)0);
    Gradient(Q_grad,Q); // Keep(Q_grad_keep,Q_grad);
    // LOG::cout<<"updating mass flux with bound flux"<<std::endl;
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        T wb=avg.Cell_To_Face(grid,it.face,Wbc);
        bound_flux(it.face)=wb*Q_grad(it.face);
    }
    if(debug_output) LOG::printf("max bound flux: %.16P\n",bound_flux.Max_Abs());    
}
//#####################################################################
// Function Momentum_Fix
//#####################################################################
template<class TV> void CAHN_HILLIARD_HELMHOLTZ<TV>::
Momentum_Fix(ARRAYS_ND_BASE<TV,IV>& u,const ARRAYS_ND_BASE<T,IV>& mass,const ARRAYS_ND_BASE<T,IV>& mass_prev,const ARRAYS_ND_BASE<bool,IV>& psi_D) const
{
    for(RANGE_ITERATOR<IV::m> it(psi_D.domain);it.Valid();it.Next()){
        if(!psi_D(it.index)) u(it.index)*=mass_prev(it.index)/mass(it.index);
    }
}
//#####################################################################
// Function Momentum_Fix_Flux
//#####################################################################
template<class TV> void CAHN_HILLIARD_HELMHOLTZ<TV>::
Momentum_Fix_Flux(ARRAY<T,FACE_INDEX<TV::m> >& u,ARRAY<T,FACE_INDEX<TV::m> >& density,T rho0, T rho1) const
{   
    ARRAY<T,FACE_INDEX<TV::m> > corrections(u.domain);
    VECTOR<ARRAY<T,FACE_INDEX<TV::m> >,TV::m> flux_avg;
    for(int axis=0;axis<TV::m;axis++)
        flux_avg(axis).Resize(grid.Face_Indices(axis));

    for(int axis=0;axis<TV::m;axis++)
    {
        for(FACE_RANGE_ITERATOR<TV::m> it(flux_avg(axis).domain);it.Valid();it.Next())
        {
            FACE_INDEX<TV::m> f_low(it.face.axis,it.face.index.Add_Axis(it.face.axis,-1));
            FACE_INDEX<TV::m> f_high(it.face.axis,it.face.index);
            flux_avg(axis)(it.face)=(flux(f_low)+flux(f_high))/2;
        }
    }

    T rho_diff=rho1-rho0;
    for(FACE_RANGE_ITERATOR<TV::m> it(u.domain);it.Valid();it.Next())
    {
        for(int axis=0;axis<TV::m;axis++) // The axis direction where j is?
        {
            // for(int side=0;side<2;side++) // The side of the axis where j is?
            // {
                // FACE_INDEX<TV::m> j(it.face.axis,it.face.index.Add_Axis(axis,2*side-1));
                // FACE_INDEX<TV::m> i(it.face.axis,f1.index.Add_Axis(it.face.axis,-1));

            FACE_INDEX<TV::m> j(it.face.axis,it.face.index.Add_Axis(axis,-1));
            T f=flux_avg(axis)(it.face);
            T rho_j=density(j);
            T rho_i=density(it.face);

            T momentum_flux=rho_diff*(u(j)-u(it.face))*f;
            corrections(it.face)+=momentum_flux/(2*rho_i);
            corrections(j)+=momentum_flux/(2*rho_j);
            // }
        }
     }

    // // TODO: correct corrections for boundaries 
    // for(int axis=0;axis<TV::m;axis++)
    // {
    //     if(periodic)
    //     {
    //         for(FACE_RANGE_ITERATOR<TV::m> it(flux.domain,0,0,RF::ghost,axis*2);it.Valid();it.Next())
    //         {
    //             FACE_INDEX<TV::m> opp(it.face);
    //             opp.index(it.side/2)=flux.domain.max_corner(it.side/2);
    //             corrections(it.face)+=corrections(opp);
    //             corrections(opp)=corrections(it.face);
    //         }
    //     }
    // }

    u.array+=corrections.array;    
}
//#####################################################################
// Function Solve_Separated_CH_Equation
//#####################################################################
// Solves u - beta L u = f
template<class TV> void CAHN_HILLIARD_HELMHOLTZ<TV>::
Solve_Separated_CH_Equation(ARRAYS_ND_BASE<T,IV>& u,const ARRAYS_ND_BASE<T,IV>& f,T beta,const ARRAYS_ND_BASE<T,IV>& gamma_ab,std::function<TV(TV,int,T)> n_func,bool use_V) const
{
    HELMHOLTZ<TV> helmholtz(grid,on_nodes);
    RANGE<IV> valid_range(grid.Cell_Indices());

    PCG_OPTIONS<T> options;
    options.max_iterations=10000;
    options.restart_iterations=200;
    options.show_results=print_solve_results;
    options.show_residual=print_solve_residual;
    options.ensure_convergence=true;
    options.absolute_tolerance=ch_abs_tolerance;
    options.relative_tolerance=ch_rel_tolerance;
    if(use_V&&volume) helmholtz.Set_Alpha(*volume);
    helmholtz.Set_Beta(beta);
    helmholtz.print_system=write_ch_system;
    helmholtz.octave_output="ch-system";
    
    ARRAY<bool,FACE_INDEX<TV::m> > internal_psi_N;
    ARRAY<T,FACE_INDEX<TV::m> > internal_psi_N_values;

    internal_psi_N.Resize(u.domain.Intersect(valid_range));
    internal_psi_N_values.Resize(internal_psi_N.domain);
    for(FACE_RANGE_ITERATOR<TV::m> it(internal_psi_N.domain,0,0,RF::ghost);it.Valid();it.Next())
    {
        if(periodic(it.side/2)) continue;
        if(psi_N) internal_psi_N(it.face)=(*psi_N)(it.face);
        else internal_psi_N(it.face)=true;
        if(n_func && internal_psi_N(it.face))
        {
            T g=0;
            if(gamma_ab.Valid_Index(it.face.index)) g=gamma_ab(it.face.index);
            internal_psi_N_values(it.face)=n_func(grid.Face_X(it.face),it.side,g)(it.face.axis);
        }
    }
    
    helmholtz.Set_Neumann_Boundaries(internal_psi_N,&internal_psi_N_values);
    
    if(psi_D)
    {
        helmholtz.Set_Dirichlet_Boundaries(*psi_D,psi_D_values);
    }
    helmholtz.Set_Periodic_Boundaries(periodic);
    bool converged=helmholtz.Solve(u,f,valid_range,options);
    if(require_convergence) PHYSBAM_ASSERT(converged);
}
namespace PhysBAM{
template class CAHN_HILLIARD_HELMHOLTZ<VECTOR<float,2> >;
template class CAHN_HILLIARD_HELMHOLTZ<VECTOR<float,3> >;
template class CAHN_HILLIARD_HELMHOLTZ<VECTOR<double,2> >;
template class CAHN_HILLIARD_HELMHOLTZ<VECTOR<double,3> >;
}
