//#####################################################################
// Copyright 2022, Jason Goulding, Mehrnaz Ayazi
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CAHN_HILLIARD_HELMHOLTZ__
#define __CAHN_HILLIARD_HELMHOLTZ__
#include <Core/Advection/ADVECTION_CONSERVATIVE_ENO.h>
#include <Core/Advection/ADVECTION_CONSERVATIVE_WENO.h>
#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Arrays/ARRAYS_ND_BASE.h>
#include <Core/Boundaries/BOUNDARY.h>
#include <Core/Boundaries/BOUNDARY_PERIODIC.h>
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Poisson/PROJECTION_UNIFORM.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include "CAHN_HILLIARD.h"
#include <functional>

namespace PhysBAM{

// template<class TV> class CH_SYMBOLIC;

template<class TV>
class CAHN_HILLIARD_HELMHOLTZ:public CAHN_HILLIARD<TV>
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    typedef CAHN_HILLIARD<TV> BASE;
    // using BASE::mass_scale; using BASE::time_scale; using BASE::distance_scale;
    using BASE::ghost;
    using BASE::order;
    using BASE::print_gamma_stats;
    using BASE::print_status;
    using BASE::print_solve_results;
    using BASE::print_solve_residual;
    using BASE::debug_output;
    // using BASE::suppress_output; using BASE::error_output;
    // using BASE::viewer_output; using BASE::debug_particles; using BASE::error_particles;
    using BASE::psi_N; using BASE::psi_N_values;
    using BASE::psi_D; using BASE::psi_D_values;
    using BASE::grid;
    using BASE::symbolic;
    // using BASE::periodic_boundary_scalar; using BASE::advection_scalar;
    using BASE::periodic;
    using BASE::side_contact_angle;

    using BASE::eta;
    using BASE::sigma;
    using BASE::M;
    using BASE::lambda;
    using BASE::test_flux;
    using BASE::flux;
    using BASE::linearize;
    // using BASE::test_units; using BASE::discrete_advection;
    using BASE::Gradient;
    // using BASE::Laplacian;
    using BASE::Fill_Ghost;
    // using BASE::Discrete_Advection;
    using BASE::Advect;
    using BASE::Print_Gamma_Stats;
    using BASE::Test_Analytic;
    using BASE::Test_Flux;
    // using BASE::Set_Periodic_Boundaries;
    // using BASE::Set_Neumann_Boundaries;
    // using BASE::Set_Dirichlet_Boundaries;
    // using BASE::Set_Analytic_Output;
    
    bool bounded=true;
    T bound_limit=1e-6;
    bool scale_diagonal=false;

    bool write_ch_system=false;
    bool write_bound_system=false;
    bool output_intermediates=false;
    
    void Bound_Gamma(ARRAYS_ND_BASE<T,IV>& gamma,const ARRAYS_ND_BASE<T,IV>& gamma_ab,T dt,T limit=1e-6);
    void Solve_Separated_CH_Equation(ARRAYS_ND_BASE<T,IV>& u,const ARRAYS_ND_BASE<T,IV>& f,T beta,const ARRAYS_ND_BASE<T,IV>& gamma_ab,std::function<TV(TV,int,T)> n_func=0,bool use_V=false) const;
    
public:

    T ch_abs_tolerance=1e-30;
    T ch_rel_tolerance=1e-12;
    T b_abs_tolerance=1e-30;
    T b_rel_tolerance=1e-12;
      
    bool on_nodes=false; // false means on cells
    bool require_convergence=true;
    bool compute_flux=true;
    bool compute_bound_flux=true;
    
    ARRAY<T,FACE_INDEX<TV::m> > bound_flux;
    ARRAYS_ND_BASE<T,IV>* volume=nullptr; // pointer allocated outside class
    
    CAHN_HILLIARD_HELMHOLTZ(const GRID<TV>& g,int ghost,PARSE_ARGS& parse_args,T kg=1,T s=1,T m=1);
    ~CAHN_HILLIARD_HELMHOLTZ();

    void Register_For_Viewing(VIEWER_OUTPUT &vo);
    
    void Initialize_Symbolics(std::string code);
    void Cahn_Hilliard(T dt,T time,ARRAYS_ND_BASE<T,IV>& gamma,ARRAYS_ND_BASE<T,IV>* gamma_prev=nullptr,ARRAY<T,FACE_INDEX<TV::m> >* face_velocities_ab=nullptr);
    
    void Momentum_Fix(ARRAYS_ND_BASE<TV,IV>& u,const ARRAYS_ND_BASE<T,IV>& mass,const ARRAYS_ND_BASE<T,IV>& mass_prev,const ARRAYS_ND_BASE<bool,IV>& psi_D) const;
    void Momentum_Fix_Flux(ARRAY<T,FACE_INDEX<TV::m> >& u,ARRAY<T,FACE_INDEX<TV::m> >& density,T rho0,T rho1) const;
   
//#####################################################################
};
}
#endif
