//#####################################################################
// Copyright 2022
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAYS_ND_BASE.h>
#include <Core/Arrays/ARRAYS_ND_VIEW.h>
#include <Core/Boundaries/REFLECTION_HELPER.h>
#include <Core/Diff_Tools/Diff_Test.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Interpolation/INTERPOLATED_COLOR_MAP.h>
#include <Core/Interpolation/WENO_INTERPOLATION_CELL.h>
#include <Core/Level_Sets/LEVELSET.h>
#include <Core/Log/DEBUG_SUBSTEPS.h>
#include <Core/Log/LOG.h>
#include <Core/Log/SCOPE.h>
//#include <Core/Log/TEST_UNITS.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Ordinary_Differential_Equations/RUNGEKUTTA.h>
#include <Core/Poisson/HELMHOLTZ.h>
#include "CAHN_HILLIARD_OPTIMIZER.h"
#include "CHN_SOLVER.h"
#include "CHO_SYMBOLIC.h"

using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> CAHN_HILLIARD_OPTIMIZER<TV>::
CAHN_HILLIARD_OPTIMIZER(const GRID<TV>& g,int ghost,PARSE_ARGS& parse_args,T kg,T s,T m)
    : CAHN_HILLIARD<TV>(g,ghost,parse_args,kg,s,m)
{

    parse_args.Add("-ch_h_tol",&ch_rel_tolerance,"error tolerance","Relative tolerance for Cahn-Hilliard solves");
    parse_args.Add("-ch_h_abs_tol",&ch_abs_tolerance,"error tolerance","Absolute tolerance for Cahn-Hilliard solves");
    parse_args.Add("-ch_n_tol",&ch_n_rel_tolerance,"error tolerance","Relative tolerance for newton's method solve");
    parse_args.Add("-ch_n_abs_tol",&ch_n_abs_tolerance,"error tolerance","Absolute tolerance for newton's method solve");
    parse_args.Add("-ch_n_k_tol",&ch_n_k_tolerance,"error tolerance","Relative tolerance for newton's method's krylov solves");
    parse_args.Add("-ch_min",&r,"scalar","gamma minimum offset r for potential function");
    parse_args.Add("-ch_n_obj_err",&ch_n_obj_error,"error tolerance","Objective error for newton's method solve");
    parse_args.Add("-ch_n_ada_tol",&ch_n_ada_tolerance,"error tolerance","Adaptive tolerance for newton's method solve");
    parse_args.Add("-ch_n_max_itrs",&ch_n_max_iterations,"int","Number of Newton method iterations to allow");
    parse_args.Add_Not("-ch_no_precond",&use_preconditioner,"do not precondition objective function hessian");
    parse_args.Add_Not("-ch_cold_newton",&newton_warm_start,"do not warm-start the newton solver");
    parse_args.Add("-ch_B_newton",&newton_B_start,"warm-start the newton solver with B");
    parse_args.Add("-ch_b",&b,"scalar","arbitrary constant b");
    parse_args.Add("-ch_q",&target_q,&force_q,"scalar","compute b to get given q");
    parse_args.Add_Not("-ch_b_s",&arbitrary_constant,"use s for b");
    parse_args.Add("-ch_F_scale",&F_scale,"scalar","scale the F function");
    parse_args.Add("-ch_write_system",&write_ch_system,"write ch system matrices in octave format");
    parse_args.Add("-ch_diff_test",&diff_test,"test objective derivatives");
    parse_args.Add_Not("-ch_no_intermediates",&output_intermediates,"do not output intermediates to viewer");
    parse_args.Add_Not("-ch_C2_F",&alt_F,"use original C2 F function");
    parse_args.Parse(true);

    // output dt restriction
    if(!arbitrary_constant) LOG::cout<<"Maximum dt based on q=6k(1-2r): "<<pow<4>(eta)/(36*M*lambda*sqr(1-2*r)*sqr(F_scale))<<std::endl; // using k=sqrt(M*lambda*dt)/eta^2
    else LOG::cout<<"Maximum dt based on q=6b*dt(1-2r): "<<sqr(eta)/(6*b*(1-2*r)*F_scale)<<std::endl;
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> CAHN_HILLIARD_OPTIMIZER<TV>::
~CAHN_HILLIARD_OPTIMIZER()
{
    // if(symbolic) delete symbolic;
    if(C_curr) delete C_curr;
    if(C_prev) delete C_prev;
}
//#####################################################################
// Function Initialize_Symbolics
//#####################################################################
template<class TV> void CAHN_HILLIARD_OPTIMIZER<TV>::
Initialize_Symbolics(std::string code)
{
    cho_symbolic=new CHO_SYMBOLIC<TV>(code,r,b,F_scale,alt_F,arbitrary_constant,linearize);
    symbolic=cho_symbolic;
    symbolic->lambda=lambda;
    symbolic->eta=eta;
    symbolic->M=M;
    // symbolic->second_order=1; // How to determine?
    symbolic->sigma=sigma;
    symbolic->mass_scale=BASE::mass_scale;
    symbolic->time_scale=BASE::time_scale;
    symbolic->distance_scale=BASE::distance_scale;
    symbolic->speed_scale=BASE::distance_scale/BASE::time_scale;
}
//#####################################################################
// Function f
//#####################################################################
template<typename T> T df0(T x,T r,bool alt_F)
{
    PHYSBAM_ASSERT(x!=0);
    if(alt_F) return 6*(x-r)*(2*r*r-4*r*x+x)/x;
    else return 6*r*(1-2*r)*(1-r/x);
}
template<typename T> T df1(T x,T r)
{
    return -2*(r-x)*(3-2*r-4*x)-4*sqr(r-x);
}
template<typename T> T df(T x,T r,T eta,T F_scale,bool alt_F,bool linear)
{
    T eta2=eta*eta;
    if(linear) return F_scale*x/eta2;
    if(x<=r) // assumes x>=0
        return F_scale*df0(x,r,alt_F)/eta2;
    else if(x<(T)0.5)
        return F_scale*df1(x,r)/eta2;
    else if(x<=(T)(1-r))
        return -F_scale*df1(1-x,r)/eta2;
    else // assumes x<=1
        return -F_scale*df0(1-x,r,alt_F)/eta2;
}
//#####################################################################
// Function Cahn_Hilliard
//#####################################################################
template<class TV> void CAHN_HILLIARD_OPTIMIZER<TV>::
Cahn_Hilliard(T dt,T time,ARRAYS_ND_BASE<T,IV>& gamma,ARRAYS_ND_BASE<T,IV>* gamma_prev,ARRAY<T,FACE_INDEX<TV::m> >* face_velocities_ab)
{
    if(first_solve)
    {
        if(force_q)
        {
            b=target_q*sqr(eta)/(6*dt*(1-2*r)*F_scale);
            if(symbolic) cho_symbolic->b=b;
            LOG::printf("force q=%P, setting b=%P\n",target_q,b);
        }
        else
        {
            T temp_q=6*b*dt/sqr(eta)*(1-2*r)*F_scale;
            LOG::printf("q=%.16P",temp_q);
            if(temp_q>0.5)
            {
                b=sqr(eta)/(12*dt*(1-2*r)*F_scale);
                if(symbolic) cho_symbolic->b=b;
                LOG::printf(" - failsafe q>0.5, setting b=%P\n",b);
            }
            LOG::cout<<std::endl;
        }
    }
    if(diff_test)
    {
        auto f0=[this](T x){
            if(alt_F) return 3*(x-r)*(8*r*r-4*r*x-r+x)-12*pow<3>(r)*log(x/r);
            else return 6*r*(1-2*r)*(x-r-r*log(x/r));
        };
        auto f1=[this](T x){ return sqr(r-x)*(3-2*r-4*x); };
        auto f=[this,f0,f1](T x){
            T eta2=eta*eta;
            if(x<=r) // assumes x>=0
                return F_scale*f0(x)/eta2;
            else if(x<(T)0.5)
                return F_scale*f1(x)/eta2;
            else if(x<=(T)(1-r))
                return F_scale*f1(1-x)/eta2;
            else // assumes x<=1
                return F_scale*f0(1-x)/eta2;
        };
    
        RANDOM_NUMBERS<T> rand(1234);
        T eps=1e-6;
        for(int i=0;i<100;i++)
        {
            T x0=rand.Get_Uniform_Number(0+eps,1-eps); // random in range 0 to r
            T x1=x0+eps;
            Diff_Test(0,"f",f(x0),x0,df(x0,r,eta,F_scale,alt_F,linearize));
            Diff_Test(1,"f",f(x1),x1,df(x1,r,eta,F_scale,alt_F,linearize));
        }
    }
    
    RANGE<IV> gamma_size=gamma.domain;
    gamma_bd.Resize(gamma_size);
    gamma_ab.Resize(gamma_size);
    
    Print_Gamma_Stats(gamma,"before prepare");
    if(gamma_prev)
    {
        order=1.5;
        gamma_bd.array=(2*gamma.array-gamma_prev->array/2)/order; // 1.5*g2-2*g1+0.5*g0
        gamma_ab.array=2*gamma.array-gamma_prev->array; // 2*g1+g0
        if(symbolic) symbolic->second_order=1;
    }
    else
    {
        order=1;
        gamma_bd.array=gamma.array;
        gamma_ab.array=gamma.array;
        if(symbolic) symbolic->second_order=0;
    }

    if(symbolic) symbolic->Set_Step(time,dt*order);
    if(gamma_prev) Test_Analytic(*gamma_prev,"gamma0",time,0,relative_error);
    Test_Analytic(gamma,"gamma",time,0,relative_error);

    RANGE<IV> valid_range(gamma_bd.domain.Intersect(grid.Cell_Indices()));
    flux.Resize(valid_range);
    Test_Analytic(gamma_bd,"gamma_bd",time,0,relative_error);
    Test_Analytic(gamma_ab,"gamma_ab",time,0,relative_error);

    if(gamma_prev) gamma_prev->array=gamma.array;

    if(face_velocities_ab) Advect(gamma_bd,gamma_ab,*face_velocities_ab,dt,time);
    
    gamma_star=gamma_bd; // Post advection
    Test_Analytic(gamma_star,"gamma_star",time,0,relative_error);
    
    ARRAY<T,IV>* gamma_presolve=nullptr;
    if(test_flux) gamma_presolve=new ARRAY<T,IV>(gamma_star);
    
    T eta2=eta*eta;
    T ML=M*lambda;
    T s=sqrt(ML/dt);
    if(!arbitrary_constant) b=s;
    T a=b*dt;
    if(debug_output)
    {
        LOG::cout<<"a="<<a;
        LOG::cout<<" s="<<s;
        LOG::cout<<" eta2="<<eta2;
        LOG::cout<<" ML="<<ML<<std::endl;
    }

    ARRAY<T,IV> df_gamma_ab(gamma_size);
    for(RANGE_ITERATOR<TV::m> it(gamma_size,0,0);it.Valid();it.Next()) // Should include the ghost region for the laplacian
    {
        df_gamma_ab(it.index)=df(gamma(it.index),r,eta,F_scale,alt_F,linearize);
        if(gamma_prev) df_gamma_ab(it.index)+=df(gamma(it.index),r,eta,F_scale,alt_F,linearize)-df((*gamma_prev)(it.index),r,eta,F_scale,alt_F,linearize);
    }
    Test_Analytic(df_gamma_ab,"df_ab_gamma",time,0,relative_error);
    ARRAY<T,IV> lap_df(valid_range);
    Laplacian(lap_df,df_gamma_ab);
    Test_Analytic(lap_df,"lap_df_ab",time,0,relative_error);
    
    ARRAY<T,IV> D(gamma_size),Arhs(gamma_size),gamma2(gamma_size);
    A.Resize(gamma_size);
    B.Resize(gamma_size);
    C.Resize(gamma_size);
    if(newton_warm_start)
    {
        if(C_curr&&C_prev) gamma2.array=2*C_curr->array-C_prev->array;
        else if(C_curr) gamma2.array=C_curr->array;
        else if(C_prev) PHYSBAM_ASSERT("C_prev defined but not C_curr");
    }
    
    for(RANGE_ITERATOR<TV::m> it(valid_range,0,0);it.Valid();it.Next())
        D(it.index)=2*gamma_ab(it.index)-ML*dt*lap_df(it.index);
    Arhs.array=gamma_star.array-D.array;
   
    if(debug_output) LOG::cout<<"A rhs sum="<<Arhs.Sum()<<std::endl;
    if(symbolic)
        for(RANGE_ITERATOR<TV::m> it(valid_range,0,0);it.Valid();it.Next())
            Arhs(it.index)+=dt*symbolic->Get_T("src_ch",grid.Cell_X(it.index));
    Test_Analytic(Arhs,"gamma_D",time,0,relative_error);
    if(debug_output) LOG::cout<<"A rhs (w/source) sum="<<Arhs.Sum()<<std::endl;
    PHYSBAM_DEBUG_WRITE_SUBSTEP("CH: compute rhs for A solve",3);

    if(print_status) LOG::cout<<"Solving first CH equation..."<<std::endl;
    Solve_Helmholtz_CH_Equation(A,Arhs,dt*s,gamma_ab,[this,a,s,dt](TV X,int side,T g_ab){
        if(!symbolic) return TV();
        else
        {
            return -symbolic->Get_TV("grad_gamma_C",X)+(a-s*dt)*symbolic->Get_TV("grad_df_ab_gamma",X)+s*dt*symbolic->Get_TV("grad_bc_xi",X)/lambda;
        }
    });
    Test_Analytic(A,"gamma_A",time,0,relative_error);
    if(debug_output) LOG::cout<<"gamma_A sum="<<A.Sum()<<std::endl;
    PHYSBAM_DEBUG_WRITE_SUBSTEP("CH: A solve",3);
    for(RANGE_ITERATOR<TV::m> it(valid_range,0,0);it.Valid();it.Next())
        B(it.index)=2*gamma_ab(it.index)-a*s*dt*lap_df(it.index);
    B.array+=A.array;
    Test_Analytic(B,"gamma_B",time,0,relative_error);
    if(debug_output) LOG::cout<<"gamma_B sum="<<B.Sum()<<std::endl;
    PHYSBAM_DEBUG_WRITE_SUBSTEP("CH: compute rhs for C solve",3);

    if(newton_B_start) gamma2.array=B.array;
    if(print_status) LOG::cout<<"Solving second CH equation using Newton's Method..."<<std::endl;
    Solve_Newton_CH_Equation(gamma2,C,B,dt*s,a*F_scale,dt,gamma_ab,[this,s](TV X,int side,T g_ab){
        if(!symbolic) return TV();
        else return symbolic->Get_TV("grad_gamma_C",X);}); // TODO: update this
    if(newton_warm_start)
    {
        if(!C_curr) C_curr=new ARRAY<T,IV>(gamma_size);
        else
        {
            if(!C_prev) C_prev=new ARRAY<T,IV>(gamma_size);
            C_prev->array=C_curr->array;
        }
        C_curr->array=C.array;
    }
    Test_Analytic(C,"gamma_C",time,0,relative_error);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("CH: C solve",3);

    // I want to see if gamma2=C-s*dth*df(gamma2)
    // or C-s*dth*df(gamma2)-gamma2=0?
    if(debug_output)
    {
        T L2_error=0;
        T Linf_error=0;
        int n=0;
        ARRAY<T,IV> error(valid_range);
        for(RANGE_ITERATOR<TV::m> it(valid_range,0,0);it.Valid();it.Next())
        {
            T e=C(it.index)-a*df(gamma2(it.index),r,eta,F_scale,alt_F,linearize)-gamma2(it.index);
            error(it.index)=e;
            L2_error+=e*e;
            Linf_error=abs(e)>Linf_error?abs(e):Linf_error;
            n++;
        }
        LOG::printf("alt gamma2 test L2: %P Linf: %P\n",sqrt(L2_error/n),Linf_error);
    }
    if(debug_output&&symbolic)
    {
        T L2_error=0;
        T Linf_error=0;
        int n=0;
        ARRAY<T,IV> error(valid_range);
        for(RANGE_ITERATOR<TV::m> it(valid_range,0,0);it.Valid();it.Next())
        {
            TV x=grid.Cell_X(it.index);
            T e=symbolic->Get_T("gamma_bd",x)-symbolic->Get_T("gamma_bd_alt",x);
            error(it.index)=e;
            L2_error+=e*e;
            Linf_error=abs(e)>Linf_error?abs(e):Linf_error;
            n++;
        }
        LOG::printf("backwards gamma_bd test L2: %P Linf: %P\n",sqrt(L2_error/n),Linf_error);
    }

    // Compute Flux
    if(compute_flux)
    {
        ARRAY<T,FACE_INDEX<TV::m> > grad_C,grad_A,grad_dF_ab;
        PHYSBAM_ASSERT(gamma_size.Contains(grid.Cell_Indices(ghost)));

        grad_C.Resize(valid_range);
        grad_A.Resize(valid_range);
        grad_dF_ab.Resize(valid_range);
        Fill_Ghost(C,C,ghost,[this](const IV& index,int side){
            if(!symbolic) return (T)0;
            else return (side%2?1:-1)*symbolic->Get_TV("grad_gamma_C",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
        });
        Gradient(grad_C,C);
        Fill_Ghost(A,A,ghost,[this](const IV& index,int side){
            if(!symbolic) return (T)0;
            else return (side%2?1:-1)*symbolic->Get_TV("grad_gamma_A",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
        });
        Gradient(grad_A,A);
        Gradient(grad_dF_ab,df_gamma_ab);

        for(FACE_RANGE_ITERATOR<TV::m> it(flux.domain);it.Valid();it.Next())
            flux(it.face)=s*grad_C(it.face)+s*grad_A(it.face)+(ML-a*s)*grad_dF_ab(it.face);
    }
    if(test_flux)
    {
        if(symbolic)
            for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                (*gamma_presolve)(it.index)+=dt*symbolic->Get_T("src_ch",grid.Cell_X(it.index));
        Test_Flux(grid,*gamma_presolve,gamma2,flux,dt,"gamma solve flux");
        delete gamma_presolve;
    }

    for(CELL_ITERATOR<TV> it(grid,ghost,GHOST_REGION);it.Valid();it.Next())
    {
        // I don't want to see the ghost region data
        A(it.index)=0;
        B(it.index)=0;
        C(it.index)=0;
        gamma_star(it.index)=0;
        gamma_ab(it.index)=0;
        gamma_bd(it.index)=0;
    }
    
    if(gamma_prev) gamma_prev->array=gamma.array; // keep old data
    gamma.array=gamma2.array;

    Test_Analytic(gamma,"gamma2",time,0,relative_error);
    Print_Gamma_Stats(gamma,"after ch solve");

    if(first_solve) first_solve=false;
}
//#####################################################################
// Function Register_For_Viewing
//#####################################################################
template<class TV> void CAHN_HILLIARD_OPTIMIZER<TV>::
Register_For_Viewing(VIEWER_OUTPUT &vo)
{
    if(!output_intermediates) return;
    SCALAR_FIELD_OPTIONS<VIEWER_OUTPUT::F<T>> opt{.draw=false,.cell_centered=true};
    opt.color_map.Set({});
    opt.color_map.Value().colors.Add_Control_Point(0,{0,0,0});
    opt.color_map.Value().colors.Add_Control_Point(1,{1,1,1});
    vo.Add("ch_A","grid",A,false,opt);
    vo.Add("ch_B","grid",B,false,opt);
    vo.Add("ch_C","grid",C,false,opt);
    vo.Add("ch_gamma_star","grid",gamma_star,false,opt);
    vo.Add("ch_gamma_ab","grid",gamma_ab,false,opt);
    vo.Add("ch_gamma_bd","grid",gamma_bd,false,opt);
}
//#####################################################################
// Function Solve_Helmholtz_CH_Equation
//#####################################################################
// Solves u - beta L u = f
template<class TV> void CAHN_HILLIARD_OPTIMIZER<TV>::
Solve_Helmholtz_CH_Equation(ARRAYS_ND_BASE<T,IV>& u,const ARRAYS_ND_BASE<T,IV>& f,T beta,const ARRAYS_ND_BASE<T,IV>& gamma_ab,std::function<TV(TV,int,T)> n_func) const
{
    HELMHOLTZ<TV> helmholtz(grid,false);
    RANGE<IV> valid_range(grid.Cell_Indices());

    PCG_OPTIONS<T> options;
    options.max_iterations=1000;
    options.restart_iterations=200;
    options.show_results=print_solve_results;
    options.show_residual=print_solve_residual;
    options.ensure_convergence=true;
    options.absolute_tolerance=ch_abs_tolerance;
    options.relative_tolerance=ch_rel_tolerance;
    helmholtz.Set_Beta(beta);
    helmholtz.print_system=write_ch_system;
    helmholtz.octave_output="ch-h-system";
    
    ARRAY<bool,FACE_INDEX<TV::m> > internal_psi_N;
    ARRAY<T,FACE_INDEX<TV::m> > internal_psi_N_values;

    internal_psi_N.Resize(u.domain.Intersect(valid_range));
    internal_psi_N_values.Resize(internal_psi_N.domain);
    for(FACE_RANGE_ITERATOR<TV::m> it(internal_psi_N.domain,0,0,RF::ghost);it.Valid();it.Next())
    {
        if(periodic(it.side/2)) continue;
        if(psi_N) internal_psi_N(it.face)=(*psi_N)(it.face);
        else internal_psi_N(it.face)=true;
        if(n_func && internal_psi_N(it.face))
        {
            T g=0;
            if(gamma_ab.Valid_Index(it.face.index)) g=gamma_ab(it.face.index);
            internal_psi_N_values(it.face)=(it.side%2?1:-1)*n_func(grid.Face_X(it.face),it.side,g)(it.face.axis); // TODO: interpolate gamma_ab from ghost region?
        }
    }
    
    helmholtz.Set_Neumann_Boundaries(internal_psi_N,&internal_psi_N_values);
    
    if(psi_D)
    {
        helmholtz.Set_Dirichlet_Boundaries(*psi_D,psi_D_values);
    }
    helmholtz.Set_Periodic_Boundaries(periodic);
    bool converged=helmholtz.Solve(u,f,valid_range,options);
    if(require_convergence) PHYSBAM_ASSERT(converged);
}
//#####################################################################
// Function Solve_Newton_CH_Equation
//#####################################################################
template<class TV> void CAHN_HILLIARD_OPTIMIZER<TV>::
Solve_Newton_CH_Equation(ARRAYS_ND_BASE<T,IV>& u,ARRAYS_ND_BASE<T,IV>& v,const ARRAYS_ND_BASE<T,IV>& f,T beta,T k,T dt,const ARRAYS_ND_BASE<T,IV>& gamma_ab,std::function<TV(TV,int,T)> n_func) const
{
    CHN_SOLVER<TV> solver(grid,eta,r,k,dt,linearize,alt_F);
    RANGE<IV> valid_range(grid.Cell_Indices());

    NEWTON_OPTIONS<T> options;
    options.max_iterations=ch_n_max_iterations;
    options.restart_iterations=1000;
    options.show_results=print_solve_results;
    options.show_residual=print_solve_residual;
    options.ensure_convergence=true;
    options.absolute_tolerance=ch_n_abs_tolerance;
    options.relative_tolerance=ch_n_rel_tolerance;
    options.adaptive_tolerance=ch_n_ada_tolerance;
    options.krylov_tolerance=ch_n_k_tolerance;
    options.objective_error=ch_n_obj_error;
    solver.Set_Beta(beta);
    solver.print_system=write_ch_system;
    solver.octave_output="ch-n-system";
    solver.use_preconditioner=use_preconditioner;
    solver.debug=debug_output;
    solver.diff_test=diff_test;
    
    ARRAY<bool,FACE_INDEX<TV::m> > internal_psi_N;
    ARRAY<T,FACE_INDEX<TV::m> > internal_psi_N_values;

    internal_psi_N.Resize(u.domain.Intersect(valid_range));
    internal_psi_N_values.Resize(internal_psi_N.domain);
    for(FACE_RANGE_ITERATOR<TV::m> it(internal_psi_N.domain,0,0,RF::ghost);it.Valid();it.Next())
    {
        if(periodic(it.side/2)) continue;
        if(psi_N) internal_psi_N(it.face)=(*psi_N)(it.face);
        else internal_psi_N(it.face)=true;
        if(n_func && internal_psi_N(it.face))
        {
            T g=0;
            if(gamma_ab.Valid_Index(it.face.index)) g=gamma_ab(it.face.index);
            internal_psi_N_values(it.face)=(it.side%2?1:-1)*n_func(grid.Face_X(it.face),it.side,g)(it.face.axis); // TODO: interpolate gamma_ab from ghost region?
        }
    }
    
    solver.Set_Neumann_Boundaries(internal_psi_N,&internal_psi_N_values);
    
    if(psi_D)
    {
        solver.Set_Dirichlet_Boundaries(*psi_D,psi_D_values);
    }
    solver.Set_Periodic_Boundaries(periodic);
    bool converged=solver.Solve(u,v,f,valid_range,options);
    if(require_convergence) PHYSBAM_ASSERT(converged);
}
namespace PhysBAM{
template class CAHN_HILLIARD_OPTIMIZER<VECTOR<float,2> >;
template class CAHN_HILLIARD_OPTIMIZER<VECTOR<float,3> >;
template class CAHN_HILLIARD_OPTIMIZER<VECTOR<double,2> >;
template class CAHN_HILLIARD_OPTIMIZER<VECTOR<double,3> >;
}
