//#####################################################################
// Copyright 2022-23, Jason Goulding
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAYS_ND_BASE.h>
#include <Core/Basic_Geometry/SPHERE.h>
#include <Core/Boundaries/REFLECTION_HELPER.h>
#include <Core/Interpolation/INTERPOLATED_COLOR_MAP.h>
#include <Core/Interpolation/WENO_INTERPOLATION_CELL.h>
#include <Core/Level_Sets/LEVELSET.h>
#include <Core/Log/DEBUG_SUBSTEPS.h>
#include <Core/Log/LOG.h>
#include <Core/Log/SCOPE.h>
#include <Core/Ordinary_Differential_Equations/RUNGEKUTTA.h>
#include <Core/Poisson/HELMHOLTZ.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include "ANALYTIC_TESTS.h"
#include "CHHNS_SYMBOLIC.h"
#include "CHNS_DRIVER.h"
#include "CHONS_SYMBOLIC.h"

using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> CHNS_DRIVER<TV>::
CHNS_DRIVER(const STREAM_TYPE stream_type_input,PARSE_ARGS& parse_args)
    :stream_type(stream_type_input),debug_particles(DEBUG_PARTICLES<TV>::Get()),
    grid(IV(),RANGE<TV>::Unit_Box()),projection(grid,true,true)
{
    T m=1;
    T s=1;
    T kg=1;
    
    TV g;
    bool use_gravity=false;
    bool use_g=false;
    bool print_all=false;
    bool all_rel_tol=false;
    bool all_abs_tol=false;
    bool use_cho=false;

    parse_args.Add("-refine",&refine,"scalar","refine resolution and time step size");
    parse_args.Add("-restart",&restart,"frame","restart frame");
    parse_args.Add("-old_write",&old_write_frequency,"frame","write previous state values every # frames");
    parse_args.Add("-substeps",&write_substeps_level,"level","output-substep level");
    parse_args.Add("-e",&last_frame,"frame","last frame");
    parse_args.Add("-frame_dt",&frame_dt,"scalar","frame_dt");
    parse_args.Add("-spf",&steps_per_frame,"int","minimum number of time steps per frame");
    parse_args.Add("-mu_tol",&mu_solver_tolerance,&use_mu_tol,"error tolerance","Relative tolerance for viscosity solves");
    parse_args.Add("-mu_abs_tol",&mu_abs_tolerance,&use_mu_abs_tol,"error tolerance","Absolute tolerance for viscosity solves");
    parse_args.Add("-p_tol",&p_solver_tolerance,&use_p_tol,"error tolerance","Relative tolerance for pressure solve");
    parse_args.Add("-p_abs_tol",&p_abs_tolerance,&use_p_abs_tol,"error tolerance","Absolute tolerance for pressure solve");
    parse_args.Add("-rho0",&rho0,"rho0","phase 0 density parameter");
    parse_args.Add("-rho1",&rho1,"rho1","phase 1 density parameter");
    parse_args.Add("-mu0",&mu0,"mu0","phase 0 viscosity parameter");
    parse_args.Add("-mu1",&mu1,"mu1","phase 1 viscosity parameter");
    parse_args.Add("-gravity",&use_gravity,"Apply gravity force");
    parse_args.Add("-gravity_v",&g,&use_g,"vector","Apply gravity force according to vector");
    parse_args.Add("-second_order",&second_order_time,"second order time estimates");
    parse_args.Add("-code",&code,"string","symbolic equations to test (use with test 1)");
    parse_args.Add_Not("-no_st_force",&st_force,"disable surface tension forces");
    parse_args.Add_Not("-no_viscosity",&use_viscosity,"disable surface tension forces");
    parse_args.Add("-ghost_out",&ghost_output,"output ghost values");
    parse_args.Add("-flux",&show_flux,"output flux values to viewer");
    parse_args.Add("-test_flux",&test_flux,"output flux test values to standard out");
    parse_args.Add("-o",&viewer_dir.output_directory,"dir","output directory");
    parse_args.Add("-suppress_output",&suppress_output,"do not print opengl output");
    parse_args.Add("-error_output",&error_output,"print error vs analytic to opengl output");
    parse_args.Add_Not("-no_require_convergence",&require_convergence,"Do not assert on solver convergence failure");
    parse_args.Add("-balanced",&balanced,"use the balanced force algorithm");
    parse_args.Add("-test_units",&test_units,"Call TEST_UNITS for test_units.py");
    parse_args.Add("-display_curvature",&display_curvature,"output debug curvature particles");
    parse_args.Add("-curvature",&curvature,"int","curvature type for surface tension types 2+");
    parse_args.Add("-ideal_curv",&use_ideal_curv,"calculate ideal curvature with radius from the center of the domain");
    parse_args.Add_Not("-no_adv",&use_advection,"Do not apply advection");
    parse_args.Add("-old_adv",&old_advection,"Use weno advection for momentum advection");
    parse_args.Add_Not("-no_pressure",&solve_pressure,"Do not solve pressure");
    parse_args.Add_Not("-no_ch",&solve_cahn_hilliard,"disable cahn-hilliard solve");
    parse_args.Add("-test_sym",&test_vs_symbolic,"Compute L2 and Linf error using symbolic code");
    parse_args.Add("-source_sym",&symbolic_source,"Use symbolic code as source terms");
    parse_args.Add("-force_u",&force_u,"Force velocity to match symbolic u/v at step start");
    parse_args.Add("-show_st",&show_st_force,"output surface tension forces to viewer");
    parse_args.Add("-d_div",&debug_divergence,"output the max velocity divergence after projection");
    parse_args.Add("-momentum",&fix_momentum,"Use the CH momentum fix");
    parse_args.Add("-st_type",&st_type,"int","specify surface tension type");
    parse_args.Add("-st_offset",&st_offset,"scalar","specify surface tension function offset");
    parse_args.Add("-st_offset2",&st_offset2,"scalar","specify second surface tension function offset");
    parse_args.Add("-keep_all",&keep_all,"keep all intermediates");
    parse_args.Add("-reflection_test",&reflection_test,"test non-periodic reflection by simulating larger periodic grid");
    parse_args.Add("-invalidate_ghost",&invalidate_ghost,"invalid ghost data set to 1e20");
    parse_args.Add("-div_free",&div_free,"project the initial velocity field");
    parse_args.Add("-dump_st",&dump_st_data,"write surface tension force data to disk");
    parse_args.Add("-discrete_adv",&discrete_advection,"use discrete advection with WENO interpolation");
    parse_args.Add("-weno_u",&use_weno_u,"use WENO interpolation on u for momentum advection");
    parse_args.Add_Not("-no_status",&print_status,"do not print non-CH status messages");
    parse_args.Add_Not("-no_parameters",&print_parameters,"do not print various parameters");
    parse_args.Add("-density_stats",&print_density_stats,"print density stats");
    parse_args.Add("-momentum_stats",&print_momentum_stats,"print momentum stats");
    parse_args.Add("-energy_stats",&print_energy_stats,"print energy stats");
    parse_args.Add("-velocity_stats",&print_velocity_stats,"print velocity stats");
    parse_args.Add("-velocity_error",&print_velocity_error,"print L2 and L-inf velocity error (compared to 0)");
    parse_args.Add("-ns_solve_residual",&print_solve_residual,"print residuals for the NS viscosity and pressure solves");
    parse_args.Add_Not("-no_ns_solve_results",&print_solve_results,"do not print results for the NS viscosity and pressure solves");
    parse_args.Add("-print_all",&print_all,"print all stats and debug outputs");
    parse_args.Add("-write_viscosity_system",&write_viscosity,"write viscosity system matrices in octave format");
    parse_args.Add("-write_pressure_system",&write_pressure,"write pressure system matrices in octave format");
    parse_args.Add("-m",&m,"scale","length scale");
    parse_args.Add("-s",&s,"scale","time scale");
    parse_args.Add("-kg",&kg,"scale","mass scale");
    parse_args.Add_Not("-no_ch_adv_flux",&use_ch_adv_flux,"do not use cahn-hilliard advection flux");
    parse_args.Add_Not("-no_ch_solve_flux",&use_ch_solve_flux,"do not use cahn-hilliard advection flux");           
    parse_args.Add_Not("-no_bound_flux",&use_bound_flux,"apply the solved bound flux to mass flux");
    parse_args.Add("-ch_optimizer",&use_cho,"Use the optimzer class Cahn-Hilliard");
    
    // TODO: remove?
    parse_args.Add("-tol",&solver_tolerance,&all_rel_tol,"error tolerance","Set relative tolerance for all solves");
    parse_args.Add("-abs_tol",&abs_tolerance,&all_abs_tol,"error tolerance","Set absolute tolerance for all solves");    
    parse_args.Parse(true);

    if(print_all)
    {
        print_status=true;
        print_parameters=true;
        print_density_stats=true;
        print_momentum_stats=true;
        print_energy_stats=true;
        print_velocity_stats=true;
        print_velocity_error=true;
        print_solve_results=true;
        print_solve_residual=true;
    }
    
    if(use_gravity)
    {
        gravity.y=-(T)9.8;
    }
    if(use_g)
    {
        LOG::cout<<"gravity="<<g<<std::endl;
        gravity=g;
    }

    // Unit scaling
    speed_scale*=m/s;
    volume_scale*=pow(m,TV::m);
    distance_scale*=m;
    mass_scale*=kg;
    time_scale*=s;
    density_scale*=kg/volume_scale;
    visc_scale*=kg*m*m/(volume_scale*s);
    pressure_scale*=kg*m*m/(volume_scale*s*s);
    ns_source_scale*=kg*m/(volume_scale*s*s);
    
    frame_dt*=time_scale;
    rho0*=density_scale;
    rho1*=density_scale;
    mu0*=visc_scale;
    mu1*=visc_scale;
    gravity*=speed_scale/time_scale;

    // Set up Cahn Hilliard
    if(use_cho)
    {
        cho=new CAHN_HILLIARD_OPTIMIZER<TV>(grid,ghost,parse_args,mass_scale,time_scale,distance_scale);
        ch=cho;
    }
    else
    {
        chh=new CAHN_HILLIARD_HELMHOLTZ<TV>(grid,ghost,parse_args,mass_scale,time_scale,distance_scale);
        ch=chh;
    }
    if(print_parameters) LOG::cout<<"M="<<ch->M<<std::endl;
    ch->Set_Periodic_Boundaries(periodic); // Parse argument, technically? Also might be based on test type? Is this done twice?

    // Use PARSE_ARGS instead
    if(all_rel_tol)
    {
        if(chh)
        {
            chh->ch_rel_tolerance=solver_tolerance;
            chh->b_rel_tolerance=solver_tolerance;
        }
        else if(cho)
        {
            cho->ch_rel_tolerance=solver_tolerance;
            cho->ch_n_rel_tolerance=solver_tolerance;
        }
        p_solver_tolerance=solver_tolerance;
        mu_solver_tolerance=solver_tolerance;
    }
    if(all_abs_tol)
    {
        if(chh)
        {
            chh->ch_abs_tolerance=abs_tolerance;
            chh->b_abs_tolerance=abs_tolerance;
        }
        else if(cho)
        {
            cho->ch_abs_tolerance=abs_tolerance;
            cho->ch_n_abs_tolerance=abs_tolerance;
        }
        p_abs_tolerance=solver_tolerance;
        mu_abs_tolerance=solver_tolerance;
    }
    ch->discrete_advection=discrete_advection; // Parse argument for NS?
    ch->test_units=test_units; // Parse argument for NS
    ch->test_flux=test_flux; // Parse argument for NS
    if(chh)
    {
        chh->require_convergence=require_convergence; // Parse argument for NS
        chh->compute_bound_flux=chh->compute_bound_flux&&use_bound_flux; // Parse argument for NS
        use_bound_flux=chh->compute_bound_flux;
    }
    else if(cho)
    {
        cho->require_convergence=require_convergence; // Parse argument for NS
        use_bound_flux=false;
    }


    // setup time
    current_frame=restart?(restart):0;
    time=current_frame*frame_dt;
    
    // Set up Symbolic
    if(code!="")
    {
        if(print_status) LOG::cout<<"setting up symbolic"<<std::endl;
        if(use_cho) symbolic = new CHONS_SYMBOLIC<TV>(code,ch->linearize,st_force,(st_type>=2)?2:st_type,use_advection,fix_momentum,cho->r,cho->b,cho->F_scale,cho->alt_F,cho->arbitrary_constant);
        else symbolic = new CHHNS_SYMBOLIC<TV>(code,ch->linearize,st_force,(st_type>=2)?2:st_type,use_advection,fix_momentum);

        //TODO: chons_symbolic currently does not handle forced q values properly (changing b is not reflected in analytics)
        
        symbolic->lambda=ch->lambda;
        symbolic->eta=ch->eta;
        symbolic->M=ch->M;
        symbolic->rho0=rho0;
        symbolic->rho1=rho1;
        symbolic->mu0=mu0;
        symbolic->mu1=mu1;
        symbolic->gravity=gravity;
        symbolic->second_order=0;
        symbolic->sigma=ch->sigma;
        symbolic->mass_scale=mass_scale;
        symbolic->time_scale=time_scale;
        symbolic->distance_scale=distance_scale;
        symbolic->speed_scale=speed_scale;
        symbolic->pressure_scale=pressure_scale;

        if(second_order_time&&fill_old&&symbolic_source)
        {
            symbolic->second_order=1;
            first_step=false;
        }

        T dt=(frame_dt/steps_per_frame)/refine;

        symbolic->Set_Step(time,dt);
    }
    if(symbolic)
    {
        ch->Initialize_Symbolics(code); // Parse argument, technically
        // Set up test function
        if(test_vs_symbolic)
            ch->Set_Analytic_Output(!suppress_output,error_output,&viewer_output,&debug_particles,&error_particles);
    }
    
    // Viewer output setup
    if(TV::m==2) viewer_output.Add_Global({.is_2d=true});
    if(TV::m==3) viewer_output.template Add_Axes<TV::m>("axes",{});
    viewer_output.Add("debug_particles",debug_particles,false,{});
    viewer_output.Add("grid",grid,true,{.draw=(TV::m==2)});
    std::string xyz[]={"x","y","z"};
    for(int a=0;a<TV::m;a++)
        viewer_output.Add("grid_"+xyz[a],face_grids(a),true,{.draw=false});

    viewer_output.Add("face_velocities","grid",face_velocities,false,{.draw=false,.cell_centered=true,.color={{1,0,1}}});
    viewer_output.Add("velocities","grid",display_velocities,false,{.cell_centered=true,.color={{0,1,0}}});
    viewer_output.Add_Scalar("face_density","grid",density,false,{.draw=false});

    SCALAR_FIELD_OPTIONS<VIEWER_OUTPUT::F<T>> gamma_opt{.draw=false,.cell_centered=true};
    gamma_opt.color_map.Set({});
    gamma_opt.color_map.Value().colors.Add_Control_Point(0,{0,0,0});
    gamma_opt.color_map.Value().colors.Add_Control_Point(1,{1,1,1});
    viewer_output.Add("gamma","grid",gamma,false,gamma_opt);

    SCALAR_FIELD_OPTIONS<VIEWER_OUTPUT::F<T>> display_gamma_opt{.cell_centered=true};
    display_gamma_opt.color_map.Set({});
    display_gamma_opt.color_map.Value().colors.Add_Control_Point(0,{0,0,0});
    display_gamma_opt.color_map.Value().colors.Add_Control_Point(1,{1,1,1});
    viewer_output.Add("display_gamma","grid",display_gamma,false,display_gamma_opt);

    SCALAR_FIELD_OPTIONS<VIEWER_OUTPUT::F<T>> mass_opt{.draw=false,.cell_centered=true};
    mass_opt.color_map.Set({});
    mass_opt.color_map.Value().colors.Add_Control_Point(0,{0,0,0});
    mass_opt.color_map.Value().colors.Add_Control_Point(10,{1,1,1});
    viewer_output.Add("mass","grid",mass,false,mass_opt);

    SCALAR_FIELD_OPTIONS<VIEWER_OUTPUT::F<T>> mass0_opt{.draw=false,.cell_centered=true};
    mass0_opt.color_map.Set({});
    mass0_opt.color_map.Value().colors.Add_Control_Point(0,{0,0,0});
    mass0_opt.color_map.Value().colors.Add_Control_Point(10,{1,1,1});
    viewer_output.Add("mass0","grid",mass0,false,mass0_opt);

    SCALAR_FIELD_OPTIONS<VIEWER_OUTPUT::F<T>> mass1_opt{.draw=false,.cell_centered=true};
    mass1_opt.color_map.Set({});
    mass1_opt.color_map.Value().colors.Add_Control_Point(0,{0,0,0});
    mass1_opt.color_map.Value().colors.Add_Control_Point(10,{1,1,1});
    viewer_output.Add("mass1","grid",mass1,false,mass1_opt);

    SCALAR_FIELD_OPTIONS<VIEWER_OUTPUT::F<T>> volume_opt{.draw=false,.cell_centered=true};
    volume_opt.color_map.Set({});
    volume_opt.color_map.Value().colors.Add_Control_Point(0,{0,0,0});
    volume_opt.color_map.Value().colors.Add_Control_Point(10,{0,1,1});
    viewer_output.Add("volume","grid",volume,false,volume_opt);

    SCALAR_FIELD_OPTIONS<VIEWER_OUTPUT::F<T>> pressure_opt{.draw=false,.cell_centered=true};
    pressure_opt.color_map.Set({});
    pressure_opt.color_map.Value().Initialize_Colors(0,1,false,false,false);
    viewer_output.Add("pressure","grid",pressure,false,pressure_opt);

    SCALAR_FIELD_OPTIONS<VIEWER_OUTPUT::F<T>> curvature_opt{.draw=false,.cell_centered=true};
    curvature_opt.color_map.Set({});
    curvature_opt.color_map.Value().Initialize_Colors(0,1,false,false,false);
    viewer_output.Add("curvature","grid",curvature_display,false,curvature_opt);
    
    if(write_debug_data)
    {
        viewer_output.Add("pressure_change","grid",projection.p,false,{.draw=false,.cell_centered=true});
        viewer_output.Add("psi_N","grid",projection.elliptic_solver->psi_N,false,{.draw=false,.color={{0,1,1}}});
        viewer_output.Add("psi_D","grid",projection.elliptic_solver->psi_D,false,{.draw=false,.color={{1,0,1}},.cell_centered=true});
    }

    if(keep_all)
    {
        viewer_output.Add_Scalar("rho_ab","grid",rho_ab_keep,false,{.draw=false});
        viewer_output.Add_Scalar("rho_bd","grid",rho_bd_keep,false,{.draw=false});
        viewer_output.Add_Scalar("rho_star","grid",rho_star_keep,false,{.draw=false});
        viewer_output.Add("rho_u_ab","grid",rho_u_ab_keep,false,{.draw=false});
        viewer_output.Add("rho_u_bd","grid",rho_u_bd_keep,false,{.draw=false});
        viewer_output.Add("u_bd","grid",u_bd_keep,false,{.draw=false});
        viewer_output.Add("rho_u_star","grid",rho_u_star_keep,false,{.draw=false});
        viewer_output.Add("u_ab","grid",u_ab_keep,false,{.draw=false});
        viewer_output.Add("u_star_hat","grid",u_star_hat_keep,false,{.draw=false});
        viewer_output.Add("u_star","grid",u_star_keep,false,{.draw=false});
        viewer_output.Add("u_star_star","grid",u_star_star_keep,false,{.draw=false});
        viewer_output.Add("u_star_star_visc","grid",u_star_star_visc_keep,false,{.draw=false});
        viewer_output.Add("u_visc","grid",u_visc_keep,false,{.draw=false});
        viewer_output.Add("u_visc_prev","grid",u_visc_prev_keep,false,{.draw=false});
        viewer_output.Add("u_corrections","grid",corrections_keep,false,{.draw=false});

        viewer_output.Add("gamma_ab","grid",gamma_ab_keep,false,{.draw=false,.cell_centered=true});
        viewer_output.Add("gamma_bd","grid",gamma_bd_keep,false,{.draw=false,.cell_centered=true});
        viewer_output.Add("gamma_star","grid",gamma_star_keep,false,{.draw=false,.cell_centered=true});
        viewer_output.Add("p_diff","grid",p_diff_keep,false,{.draw=false,.cell_centered=true});

        for(int a=0;a<TV::m;a++)
            viewer_output.Add("momentum_flux_"+xyz[a],"grid_"+xyz[a],momentum_flux(a),false,{.draw=false});

        for(int a=0;a<TV::m;a++)
            viewer_output.Add("density_flux_"+xyz[a],"grid_"+xyz[a],density_flux_keep(a),false,{.draw=false});

        for(int a=0;a<TV::m;a++)
            viewer_output.Add("u_avg_"+xyz[a],"grid_"+xyz[a],u_avg_keep(a),false,{.draw=false});

        viewer_output.Add("density_cell_flux","grid",density_cell_flux_keep,false,{.draw=false});
        viewer_output.Add("gamma_flux","grid",gamma_flux_keep,false,{.draw=false});
        viewer_output.Add("gamma_advect_flux","grid",gamma_advect_flux_keep,false,{.draw=false});
        viewer_output.Add("gamma_solve_flux","grid",gamma_solve_flux_keep,false,{.draw=false});
    }
    if(dump_st_data)
    {
        viewer_output.Add("st_force","grid",st_force_keep,false,{.draw=false});
        viewer_output.Add("grad_gamma","grid",grad_gamma_keep,false,{.draw=false});
    }

    ch->Register_For_Viewing(viewer_output);
    
    DEBUG_SUBSTEPS::write_substeps_level=write_substeps_level;
    DEBUG_SUBSTEPS::writer=[this](const std::string& title,int level){viewer_output.Write_Substep(title,level);};
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> CHNS_DRIVER<TV>::
~CHNS_DRIVER()
{
    DEBUG_SUBSTEPS::writer=0;
    // TODO: delete hashtable (iterate)
    for(auto&p:error_particles)
        error_particles.Delete(p.first);
    if(symbolic) delete symbolic;
    if(chh)
    {
        ch=nullptr;
        delete chh;
    }
    else if(cho)
    {
        ch=nullptr;
        delete cho;
    }
}
//#####################################################################
// Function Initialize_Grid
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Initialize_Grid(IV counts,RANGE<TV> domain) 
{
    grid.Initialize(counts*refine,domain);
    for(int axis=0;axis<TV::m;axis++)
        face_grids(axis)=grid.Get_Face_MAC_Grid(axis);

    projection.Initialize_Grid(grid);
    face_velocities.Resize(grid.Cell_Indices(face_ghost),init_new,1e20);
    face_velocities_old.Resize(grid.Cell_Indices(face_ghost),init_new,1e20);
    gamma.Resize(grid.Cell_Indices(ghost),init_new,1e20);
    curvature_display.Resize(grid.Cell_Indices(ghost),init_new,0);
    gamma_old.Resize(grid.Cell_Indices(ghost),init_new,1e20);
    pressure.Resize(grid.Cell_Indices(ghost),init_new,1e20);
    pressure_old.Resize(grid.Cell_Indices(ghost),init_new,1e20);
    density.Resize(grid.Cell_Indices(face_ghost),init_new,0);
    density_old.Resize(grid.Cell_Indices(face_ghost),init_new,0);
    mass.Resize(grid.Cell_Indices(ghost),init_new,0);
    mass0.Resize(grid.Cell_Indices(ghost),init_new,0);
    mass1.Resize(grid.Cell_Indices(ghost),init_new,0);
    volume.Resize(grid.Cell_Indices(ghost),init_new,0);
    viscosity.Resize(grid.Cell_Indices(face_ghost),init_new,1e20);
    for(int axis=0;axis<TV::m;axis++)
        momentum_flux(axis).Resize(grid.Face_Indices(axis));

    display_velocities.Resize(grid.Cell_Indices(),init_new,1e20);
    display_gamma.Resize(grid.Cell_Indices(),init_new,1e20);

    psi_N.Resize(grid.Cell_Indices(),init_new,false);
    ch_psi_N_values.Resize(grid.Cell_Indices(),init_new,1e20);
    p_psi_N_values.Resize(grid.Cell_Indices(),init_new,1e20);
    psi_D.Resize(grid.Cell_Indices(1),init_new,false);
    ch_psi_D_values.Resize(grid.Cell_Indices(1),init_new,1e20);
    p_psi_D_values.Resize(grid.Cell_Indices(1),init_new,1e20);
}
//#####################################################################
// Function Initialize
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Fill_Analytics()
{
    T dt=(frame_dt/steps_per_frame)/refine;
    ch->Initialize_Gamma(grid,gamma,ghost,dt,0);
                    
    for(CELL_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
        pressure(it.index)=symbolic->Get_T("p",grid.Cell_X(it.index));
    for(FACE_ITERATOR<TV> it(grid,face_ghost);it.Valid();it.Next())
    {
        face_velocities(it.face)=symbolic->Get_TV("u",grid.Face_X(it.face))(it.Axis());
        density(it.face)=symbolic->Get_T("density",grid.Face_X(it.face));
        viscosity(it.face)=symbolic->Get_T("viscosity",grid.Face_X(it.face));
    }

    if(fill_old)
    {
        ch->Initialize_Gamma(grid,gamma_old,ghost,dt,-dt);
        for(CELL_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
            pressure_old(it.index)=symbolic->Get_T("p0",grid.Cell_X(it.index));
        for(FACE_ITERATOR<TV> it(grid,face_ghost);it.Valid();it.Next())
        {
            face_velocities_old(it.face)=symbolic->Get_TV("u0",grid.Face_X(it.face))(it.Axis());
            density_old(it.face)=symbolic->Get_T("density0",grid.Face_X(it.face));
        }
    }
}
//#####################################################################
// Function Initialize
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Initialize()
{
    // Finish Initialization using grid and initial data
    if(reflection_test)
    {
        // Resize Grid
        RANGE<TV> new_domain(grid.domain.min_corner,grid.domain.max_corner+grid.domain.Edge_Lengths());
        GRID<TV> new_grid(grid.numbers_of_cells*2,new_domain);
        // // Resize CH Grid
        // if(ch) ch->grid=new_grid;
        // Resize all Arrays
        face_velocities.Resize(grid.Cell_Indices(grid.numbers_of_cells(0)),init_new,1e20);
        face_velocities_old.Resize(grid.Cell_Indices(grid.numbers_of_cells(0)),init_new,1e20);
        gamma.Resize(grid.Cell_Indices(grid.numbers_of_cells(0)),init_new,1e20);
        gamma_old.Resize(grid.Cell_Indices(grid.numbers_of_cells(0)),init_new,1e20);
        pressure.Resize(grid.Cell_Indices(grid.numbers_of_cells(0)),init_new,1e20);
        pressure_old.Resize(grid.Cell_Indices(grid.numbers_of_cells(0)),init_new,1e20);
        density.Resize(grid.Cell_Indices(grid.numbers_of_cells(0)),init_new,0);
        density_old.Resize(grid.Cell_Indices(grid.numbers_of_cells(0)),init_new,0);
        viscosity.Resize(grid.Cell_Indices(grid.numbers_of_cells(0)),init_new,1e20);
        
        // Fill in new space via reflection
        for(int axis=0;axis<TV::m;axis++)
        {          
            //Call Reflection function for "right" side of axis, with num_ghost=num_cells(axis)
            Reflection_Face_Slip_Side<TV,T>(new_grid,face_velocities,grid.numbers_of_cells(axis),2*axis+1,0,0,false);
            Reflection_Face_Slip_Side<TV,T>(new_grid,face_velocities_old,grid.numbers_of_cells(axis),2*axis+1,0,0,false);
            
            Reflection_Face_Neumann_Side<TV,T>(new_grid,density,grid.numbers_of_cells(axis),2*axis+1,0);
            Reflection_Face_Neumann_Side<TV,T>(new_grid,density_old,grid.numbers_of_cells(axis),2*axis+1,0);
            Reflection_Face_Neumann_Side<TV,T>(new_grid,viscosity,grid.numbers_of_cells(axis),2*axis+1,0);
            Reflection_Cell_Neumann_Side<TV,T>(new_grid,gamma,grid.numbers_of_cells(axis),2*axis+1,0);
            Reflection_Cell_Neumann_Side<TV,T>(new_grid,gamma_old,grid.numbers_of_cells(axis),2*axis+1,0);
            Reflection_Cell_Neumann_Side<TV,T>(new_grid,pressure,grid.numbers_of_cells(axis),2*axis+1,0);
            Reflection_Cell_Neumann_Side<TV,T>(new_grid,pressure_old,grid.numbers_of_cells(axis),2*axis+1,0);
        }
        grid=new_grid;
        for(int axis=0;axis<TV::m;axis++)
            face_grids(axis)=grid.Get_Face_MAC_Grid(axis);
        periodic.Fill(true);
        slip_boundary=false;
        noslip_boundary=false;
        free_boundary=false;

        projection.Initialize_Grid(grid);
        face_velocities.Resize(grid.Cell_Indices(face_ghost),init_new,1e20);
        face_velocities_old.Resize(grid.Cell_Indices(face_ghost),init_new,1e20);
        gamma.Resize(grid.Cell_Indices(ghost),init_new,1e20);
        curvature_display.Resize(grid.Cell_Indices(ghost),init_new,0);
        gamma_old.Resize(grid.Cell_Indices(ghost),init_new,1e20);
        pressure.Resize(grid.Cell_Indices(ghost),init_new,1e20);
        pressure_old.Resize(grid.Cell_Indices(ghost),init_new,1e20);
        density.Resize(grid.Cell_Indices(face_ghost),init_new,0);
        density_old.Resize(grid.Cell_Indices(face_ghost),init_new,0);
        viscosity.Resize(grid.Cell_Indices(face_ghost),init_new,1e20);
        for(int axis=0;axis<TV::m;axis++)
            momentum_flux(axis).Resize(grid.Face_Indices(axis));

        display_velocities.Resize(grid.Cell_Indices(),init_new,1e20);
        display_gamma.Resize(grid.Cell_Indices(),init_new,1e20);

        psi_N.Resize(grid.Cell_Indices(),init_new,false);
        ch_psi_N_values.Resize(grid.Cell_Indices(),init_new,1e20);
        p_psi_N_values.Resize(grid.Cell_Indices(),init_new,1e20);
        psi_D.Resize(grid.Cell_Indices(1),init_new,false);
        ch_psi_D_values.Resize(grid.Cell_Indices(1),init_new,1e20);
        p_psi_D_values.Resize(grid.Cell_Indices(1),init_new,1e20);
    }
    
    if(!symbolic)
    {
        face_velocities.array*=speed_scale;
        pressure.array*=pressure_scale;
    }
    Update_Display_Gamma(gamma);
    Update_Display_Velocity(face_velocities);

    // Boundary Condition Setup
    for(int side=0;side<TV::m*2;side++)
    {
        if(periodic(side/2)) side_bc_type(side)=BC_PERIODIC;
        else if(slip_boundary) side_bc_type(side)=BC_SLIP;
        else if(noslip_boundary) side_bc_type(side)=BC_NOSLIP;
        else if(free_boundary) side_bc_type(side)=BC_FREE;
    }
    if(print_parameters) LOG::cout<<"side settings: "<<side_bc_type<<std::endl;
    Enforce_Boundary_Conditions(face_velocities);  // fix boundary u

    // setup domain boundaries
    for(int axis=0;axis<TV::m;axis++)
        boundary_scalar_p.is_periodic(axis)=(side_bc_type(axis*2)==BC_PERIODIC);

    // setup laplace
    projection.elliptic_solver->Set_Relative_Tolerance(use_p_tol?p_solver_tolerance:solver_tolerance);
    projection.elliptic_solver->Set_Absolute_Tolerance(use_p_abs_tol?p_abs_tolerance:abs_tolerance);
    projection.elliptic_solver->pcg.Set_Maximum_Iterations(1000);
    projection.elliptic_solver->pcg.evolution_solver_type=krylov_solver_cg;
    projection.elliptic_solver->pcg.cg_restart_iterations=200;
    projection.elliptic_solver->pcg.show_results=print_solve_results;
    projection.elliptic_solver->pcg.show_residual=print_solve_residual;
    projection.poisson->Set_Variable_beta(true);
    projection.periodic=boundary_scalar_p.is_periodic;
    projection.elliptic_solver->periodic_boundary=projection.periodic;
    ch->Set_Periodic_Boundaries(boundary_scalar_p.is_periodic);
    Set_Boundary_Conditions();
    
    if(reflection_test||!symbolic_source) // symbolic test fills ghost regions
    {
        Ghost_Fill(gamma,gamma,ghost,"grad_gamma");
        Fill_Ghost(pressure,pressure,ghost,[this](const IV& index,int side) // Fill the ghost region using linear extrapolation
        {
            if(symbolic_source) return (side%2?1:-1)*symbolic->Get_TV("grad_p",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
            else
            {
                int axis=side/2;
                int sign=(side%2*2)-1;
                IV a=Fine_Node_To_Cell(index.Add_Axis(axis,-sign));
                IV b=a.Add_Axis(axis,-sign);
                return (pressure(a)-pressure(b))/grid.dX(axis);
            }
        });
        Ghost_Fill(face_velocities,face_velocities,face_ghost,"u","grad_u","grad_v","grad_w");
        Compute_Density(density,gamma,face_ghost);
        Compute_Viscosity(viscosity,gamma,face_ghost); // Compute viscosity at time+dt

        if(symbolic_source)
        {
            Ghost_Fill(gamma_old,gamma_old,ghost,"grad_gamma0");
            Ghost_Fill(face_velocities_old,face_velocities_old,face_ghost,"u0","grad_u0","grad_v0","grad_w0");
            Compute_Density(density_old,gamma_old,face_ghost);

            Fill_Ghost(pressure_old,pressure_old,ghost,[this](const IV& index,int side) // Fill the ghost region using linear extrapolation
            {
                if(symbolic_source) return (side%2?1:-1)*symbolic->Get_TV("grad_p0",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
                else
                {
                    int axis=side/2;
                    int sign=(side%2*2)-1;
                    IV a=Fine_Node_To_Cell(index.Add_Axis(axis,-sign));
                    IV b=a.Add_Axis(axis,-sign);
                    return (pressure(a)-pressure(b))/grid.dX(axis);
                }
            });
        }
    }
    Compute_Curvature();
    Print_Density_Stats(density,"initialization");
    Compute_Divergence(face_velocities,"initialization");

    if(!restart)
    {
        output_number=0;
        Write_Output_Files("Initialization");
    }
    
    if(div_free&&solve_pressure)
    {
        T dt=(frame_dt/steps_per_frame)/refine;
        if(second_order_time&&fill_old) Project(face_velocities_old,density_old,dt,time);
        Project(face_velocities,density,dt,time);
        Compute_Divergence(face_velocities,"after initial projection");
    }

    if(restart)
    {
        if(restart>2&&second_order_time) first_step=false;
        Read_Output_Files();
    }
    
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
        total_gamma+=gamma(it.index);
    gamma_sum=total_gamma;
    init_gamma_sum=gamma_sum;

    // calculate initial momentum
    TV momentum;
    T ke=0;
    T volume=grid.Cell_Size();
    
    for(FACE_ITERATOR<TV> it(grid,0,INTERIOR_REGION);it.Valid();it.Next())
    {
        momentum(it.face.axis)+=density(it.face)*face_velocities(it.face)*volume;
        ke+=density(it.face)*sqr(face_velocities(it.face))*volume/2;
    }

    for(int axis=0;axis<TV::m;axis++)
    {
        if(side_bc_type(axis*2)==BC_PERIODIC)
        {
            for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis);it.Valid();it.Next())
            {
                momentum(it.face.axis)+=density(it.face)*face_velocities(it.face)*volume;
                ke+=density(it.face)*sqr(face_velocities(it.face))*volume/2;
            }
        }
        else
        {
            for(int side=0;side<2;side++)
            {
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis+side);it.Valid();it.Next())
                {
                    momentum(it.face.axis)+=density(it.face)*face_velocities(it.face)*volume/2;
                    ke+=density(it.face)*sqr(face_velocities(it.face))*volume/4;
                }
            }
        }
    }

    if(print_momentum_stats) LOG::cout<<"Initial total momentum: "<<momentum<<std::endl;
    if(print_energy_stats) LOG::cout<<"Initial total kinetic energy: "<<ke<<std::endl;
    initial_momentum=momentum;
    total_momentum=momentum;
    initial_ke=ke;
    total_ke=ke;
}
//#####################################################################
// Function Set_Boundary_Conditions
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Set_Boundary_Conditions()
{
    psi_D.Fill(false);
    psi_N.Fill(false);
    p_psi_D_values.Fill(0);
    
    for(int axis=0;axis<TV::m;axis++)
        for(int axis_side=0;axis_side<2;axis_side++)
        {
            int side=2*axis+axis_side;
            if(side_bc_type(side)!=BC_PERIODIC)
            {
                IV exterior_cell_offset=axis_side==0?-IV::Axis_Vector(axis):IV();
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,side);it.Valid();it.Next())
                {
                    IV out_cell=it.Face_Index()+exterior_cell_offset;
                    if(side_bc_type(side)==BC_SLIP)
                    {
                        psi_N(it.face)=true;
                    }
                    else if(side_bc_type(side)==BC_NOSLIP)
                    {
                        psi_N(it.face)=true;
                    }
                    else //if(side_bc_type(side)==BC_FREE)
                    {
                        psi_D(out_cell)=true;
                        if(!symbolic_source) p_psi_D_values(out_cell)=0; // dirchlet value for pressure???
                    }
                }
            }
        }

    ARRAY<bool,FACE_INDEX<TV::m> >::Put(psi_N,projection.elliptic_solver->psi_N);
    ch->Set_Neumann_Boundaries(psi_N);
    ARRAY<bool,IV>::Put(psi_D,projection.elliptic_solver->psi_D);
    ARRAY<T,IV>::Put(p_psi_D_values,projection.p);
    // ch->Set_Dirchlet_Boundaries(psi_D);
}
//#####################################################################
// Function Enforce_Boundary_Conditions
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Enforce_Boundary_Conditions(ARRAY<T,FACE_INDEX<TV::m> >& u)
{
    for(int side=0;side<TV::m*2;side++)
        for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,side);it.Valid();it.Next())
            if(side_bc_type(side)==BC_SLIP || side_bc_type(side)==BC_NOSLIP)
                if(!symbolic_source&&!force_u) u(it.face)=0;
}
//#####################################################################
// Function Execute_Main_Program
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Execute_Main_Program()
{
    Initialize();
    Simulate_To_Frame(last_frame);
}
//#####################################################################
// Function Simulate_To_Frame
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Simulate_To_Frame(int frame)
{
    for(;current_frame<frame;current_frame++)
    {
        LOG::SCOPE scope("FRAME","Frame %d",current_frame+1);

        T dt=(frame_dt/steps_per_frame)/refine;
        for(int substep=0;substep<refine*steps_per_frame;substep++)
            Step(substep,dt);
        
        Write_Output_Files(LOG::sprintf("Frame %P",current_frame+1));
    }
}
//#####################################################################
// Function Step
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Step(int substep,T dt)
{
    speed_limit=grid.dX(0)/dt;
    
    if(print_parameters) LOG::printf("L2 stats dx %d dt %d \n",grid.dX(0),dt);

    if(second_order_time&&(!first_step))
    {
        order=1.5;
        if(symbolic) symbolic->second_order=1;
    }
    else
    {
        order=1;
        if(symbolic) symbolic->second_order=0;
    }
    if(symbolic) symbolic->Set_Step(time,dt);
    
    if(symbolic&&force_u)
    {
        for(FACE_ITERATOR<TV> it(grid,face_ghost);it.Valid();it.Next())
        {
            TV X=grid.Face_X(it.face);
            face_velocities_old(it.face)=symbolic->Get_TV("u0",X)(it.face.axis);
            face_velocities(it.face)=symbolic->Get_TV("u",X)(it.face.axis);
        }
    }
    
#define Rename(n,o) decltype(o) n; n.Exchange(o);
    // At this point, face_velocities and density should be filled in the ghost region from either the previous substep or the Initialize function
    ARRAY<T,FACE_INDEX<TV::m> > face_velocities_ab(grid.Cell_Indices(face_ghost),init_new,1e20);
    ARRAY<T,FACE_INDEX<TV::m> > face_velocities_bd(grid.Cell_Indices(face_ghost),init_new,1e20);
    ARRAY<T,FACE_INDEX<TV::m> > rho_u_bd(grid.Cell_Indices(face_ghost),init_new,1e20);
    ARRAY<T,FACE_INDEX<TV::m> > u_bd(grid.Cell_Indices(face_ghost),init_new,1e20);
    ARRAY<T,FACE_INDEX<TV::m> > rho_u_ab(grid.Cell_Indices(face_ghost),init_new,1e20);
    ARRAY<T,FACE_INDEX<TV::m> > density_bd(grid.Cell_Indices(face_ghost),init_new,1e20);
    ARRAY<T,IV> density_cell_bd(grid.Cell_Indices(ghost),init_new,1e20);
    ARRAY<T,IV> density_cell_ab(grid.Cell_Indices(ghost),init_new,1e20);
    ARRAY<T,FACE_INDEX<TV::m> > density_ab(grid.Cell_Indices(face_ghost),init_new,1e20);
    
    if(second_order_time&&(!first_step))
    {
        face_velocities_ab.array=2*face_velocities.array-face_velocities_old.array;
        face_velocities_bd.array=2*face_velocities.array-face_velocities_old.array/2;
        rho_u_bd.array=(2*face_velocities.array*density.array-face_velocities_old.array*density_old.array/2)/order;
        rho_u_ab.array=2*face_velocities.array*density.array-face_velocities_old.array*density_old.array;
        density_bd.array=(2*density.array-density_old.array/2)/order;
        density_ab.array=2*density.array-density_old.array;

        density_cell_bd.array=(2*(rho0+(rho1-rho0)*gamma.array)-(rho0+(rho1-rho0)*gamma_old.array)/2)/order;
        density_cell_ab.array=2*(rho0+(rho1-rho0)*gamma.array)-(rho0+(rho1-rho0)*gamma_old.array);
    }
    else
    {
        face_velocities_ab.array=face_velocities.array;
        face_velocities_bd.array=face_velocities.array;
        rho_u_bd.array=face_velocities.array*density.array;
        rho_u_ab.array=face_velocities.array*density.array;
        density_bd.array=density.array;
        density_ab.array=density.array;

        density_cell_bd.array=rho0+(rho1-rho0)*gamma.array;
        density_cell_ab.array=rho0+(rho1-rho0)*gamma.array;
    }
    Test_Analytic_T(gamma_old,"gamma0",time);
    Test_Analytic_T(gamma,"gamma",time);
    Compute_Density(density,gamma,face_ghost);
    Test_Analytic_TV(face_velocities_old,"u0",time);
    Test_Analytic_TV(face_velocities,"u",time);

    Print_Density_Stats(density,"step start");
    Total_Momentum(face_velocities,density,"step start",time);
    Print_Velocity_Stats(face_velocities,"step start");
    T dt_hat=dt/order;

    u_bd.array=rho_u_bd.array/density_bd.array;
    Keep(u_bd_keep,u_bd);
        
    Test_Analytic_TV(face_velocities_ab,"u_ab",time,face_ghost); Keep(u_ab_keep,face_velocities_ab);
    Test_Analytic_T(density_old,"density0",time,face_ghost);
    Test_Analytic_T(density,"density",time,face_ghost);
    Test_Analytic_T(viscosity,"viscosity",time,face_ghost);
    Test_Analytic_TV(rho_u_bd,"rho_u_bd",time,face_ghost); Keep(rho_u_bd_keep,rho_u_bd);
    Test_Analytic_TV(rho_u_ab,"rho_u_ab",time,face_ghost); Keep(rho_u_ab_keep,rho_u_ab);
    Test_Analytic_T(density_bd,"rho_bd",time,face_ghost); Keep(rho_bd_keep,density_bd);
    Test_Analytic_T(density_ab,"rho_ab",time,face_ghost); Keep(rho_ab_keep,density_ab);
    Test_Analytic_T(pressure,"p",time,ghost,"p",true);

    // Total_Momentum(face_velocities_ab,gamma,"u_ab",time);
    Print_Velocity_Stats(face_velocities_ab,"u_ab");
    Total_Momentum(rho_u_bd,"rho_u_bd",time);
    // Total_Momentum(rho_u_ab,"rho_u_ab",time);
       
    density_old=density;

    Print_Density_Stats(density,"before advection");

    ARRAY<T,FACE_INDEX<TV::m> > density_cell_flux(grid.Cell_Indices(face_ghost));
    ARRAY<T,IV> density_preadvect=density_cell_bd;
    VECTOR<ARRAY<T,FACE_INDEX<TV::m> >,TV::m> density_flux;
    ARRAY<T,FACE_INDEX<TV::m> > density_test=density_bd;
    ARRAY<T,FACE_INDEX<TV::m> > momentum_test=rho_u_bd;

    /////////////////////////////////////////////////
    // Cahn-Hilliard
    //Prepare CH
    auto gamma_save=gamma;
    // Keep(gamma_bd_keep,ch->gamma_bd);
    // Keep(gamma_ab_keep,ch->gamma_ab);
    // Keep(gamma_star_keep,ch->gamma_bd);
    // Update_Display_Gamma(ch->gamma_bd);
    // PHYSBAM_DEBUG_WRITE_SUBSTEP("after gamma advection",2);
    
    // Cahn-Hilliard Solve
    if(solve_cahn_hilliard)
    {
        ARRAYS_ND_BASE<T,IV>* g_prev=(first_step||!second_order_time)?nullptr:&gamma_old;
        ARRAY<T,FACE_INDEX<TV::m> >* u_ab=use_advection?&face_velocities_ab:nullptr;
        ch->Cahn_Hilliard(dt_hat,time,gamma,g_prev,u_ab);
        if(use_ch_adv_flux) Keep(gamma_advect_flux_keep,ch->advect_flux);
        if(use_ch_solve_flux) Keep(gamma_solve_flux_keep,ch->flux);

    }
    else if(print_status) LOG::cout<<"Not solving Cahn-Hilliard!"<<std::endl;
    Fill_Ghost(gamma,gamma,ghost,[this](const IV& index,int side){
        if(symbolic_source) return (side%2?1:-1)*symbolic->Get_TV("grad_gamma2",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
        else
        {
            int axis=side/2;
            int sign=(side%2*2)-1;
            IV a=Fine_Node_To_Cell(index.Add_Axis(axis,-sign));
            // IV b=a.Add_Axis(axis,-sign);
            T gamma_ab=2*gamma(a)-gamma_old(a);
            return 0.5*(ch->sigma/ch->lambda)*cos(ch->side_contact_angle(side))*(pi/2)*cos(pi*gamma_ab-pi/2);
        }
    });
    Update_Display_Gamma(gamma);
    // if(symbolic && linearize) Test_Analytic_T(gamma,"test_gamma",time);
    if(first_step) gamma_old=gamma_save;
    //////////////////////////////////////////////////
    
    Compute_Density(density,gamma,face_ghost); // Compute density at time+dt
    Test_Analytic_T(density,"density2",time,face_ghost);
    Compute_Viscosity(viscosity,gamma,face_ghost); // Compute viscosity at time+dt
    Compute_Curvature();
    
    Print_Density_Stats(density,"after ch solve");
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after ch solve",2);
    
    if(use_advection&&old_advection)
    {
        // Compute mass flux
        ARRAY<T,FACE_INDEX<TV::m> > mass_flux(grid.Cell_Indices(face_ghost));
        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            mass_flux(it.face)=rho_u_ab(it.face)-(rho1-rho0)*ch->flux(it.face); // rho*u_ab?
            if(use_bound_flux&&chh) mass_flux(it.face)-=(rho1-rho0)*chh->bound_flux(it.face);
        }
        for(int side=0;side<TV::m*2;side++)
        {
            if(side_bc_type(side)==BC_PERIODIC) continue;
            else if(side_bc_type(side)==BC_SLIP)
            {
                Reflection_Face_Slip_Side<TV,T>(grid,mass_flux,face_ghost,side,0,0,true);
            }
            else if(side_bc_type(side)==BC_NOSLIP)
            {
                Reflection_Face_NoSlip_Side<TV,T>(grid,mass_flux,ghost,side,0,true);
            }
            else PHYSBAM_FATAL_ERROR("Invalid Boundary Type");
        }
        boundary_scalar_p.Fill_Ghost_Faces(grid,mass_flux,mass_flux,face_ghost);

        if(discrete_advection) Discrete_Advection(grid,rho_u_bd,face_velocities_ab,mass_flux,dt_hat,time);
        else advection_scalar.Update_Advection_Equation_Face(grid,rho_u_bd,face_velocities_ab,mass_flux,dt_hat,time);
    }
    else if(use_advection&&(use_ch_adv_flux||use_ch_solve_flux)) // Momentum advection
    {
        // Flux setup and testing
        T rho_diff=rho1-rho0;
        ARRAY<T,FACE_INDEX<TV::m> > full_rho_flux(grid.Cell_Indices(face_ghost));
        ARRAY<T,FACE_INDEX<TV::m> > full_gamma_flux(grid.Cell_Indices(face_ghost));
        
        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            full_rho_flux(it.face)=0;
            if(use_ch_adv_flux) full_rho_flux(it.face)+=rho0*face_velocities_ab(it.face)+rho_diff*ch->advect_flux(it.face);
            if(use_ch_solve_flux) full_rho_flux(it.face)+=-rho_diff*ch->flux(it.face);
            full_gamma_flux(it.face)=ch->advect_flux(it.face)-ch->flux(it.face);
            if(use_bound_flux&&chh)
            {
                full_gamma_flux(it.face)-=chh->bound_flux(it.face);
                full_rho_flux(it.face)-=rho_diff*chh->bound_flux(it.face);
            }
        }
        Test_Analytic_TV(full_rho_flux,"mass_flux",time);
        Keep(gamma_flux_keep,full_gamma_flux);

        ARRAY<T,IV> density_cell(gamma.domain);
        density_cell.array=rho0+rho_diff*gamma.array;
        if(symbolic)
            for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                density_preadvect(it.index)+=dt_hat*rho_diff*symbolic->Get_T("src_ch",grid.Cell_X(it.index));
        Test_Flux(grid,density_preadvect,density_cell,full_rho_flux,-dt_hat,"full density cell");
        
        // Fill the ghost regions for fluxes (needed for averaging)
        for(int side=0;side<TV::m*2;side++)
        {
            if(side_bc_type(side)==BC_PERIODIC) continue;
            Reflection_Face_Slip_Side<TV,T>(grid,full_rho_flux,face_ghost,side,0);
        }
        boundary_scalar_p.Fill_Ghost_Faces(grid,full_rho_flux,full_rho_flux,face_ghost);

        // Flux averaging (full density flux (n+1) on faces)
        VECTOR<ARRAY<T,FACE_INDEX<TV::m> >,TV::m> density_flux_full;
        for(int axis=0;axis<TV::m;axis++)
            density_flux_full(axis).Resize(grid.Face_Indices(axis));
            
        for(int axis=0;axis<TV::m;axis++)
        {    
            for(FACE_ITERATOR<TV> it(face_grids(axis));it.Valid();it.Next())
            {
                IV second_cell=it.Second_Cell_Index();
                IV first_cell=second_cell.Add_Axis(axis,-1);
                density_flux_full(axis)(it.face)=(full_rho_flux.Component(it.face.axis)(first_cell)+full_rho_flux.Component(it.face.axis)(second_cell))/2;
            }
        }
        if(symbolic)
            for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                density_test(it.face)+=dt_hat*rho_diff*symbolic->Get_T("src_ch",grid.Face_X(it.face));
        for(int axis=0;axis<TV::m;axis++)
            Test_Flux(grid.Get_Face_MAC_Grid(axis),density_test.Component(axis),density.Component(axis),density_flux_full(axis),-dt_hat,LOG::sprintf("face density flux full %P",axis));

        // Creating momentum flux from velocity and density flux
        for(int axis=0;axis<TV::m;axis++)
            u_avg_keep(axis).Resize(grid.Face_Indices(axis));

        for(int axis=0;axis<TV::m;axis++)
            Keep(density_flux_keep(axis),density_flux_full(axis));

        
        WENO_INTERPOLATION_CELL<TV,T> interp;
        for(int axis=0;axis<TV::m;axis++)
        {    
            for(FACE_ITERATOR<TV> it(face_grids(axis));it.Valid();it.Next())
            {
                IV second_cell_u=it.Second_Cell_Index();
                IV first_cell_u=second_cell_u.Add_Axis(it.face.axis,-1);

                T u_avg=0;
                if(use_weno_u) u_avg=interp.Interpolate_Cell(face_grids(axis),face_velocities_ab.Component(axis),face_grids(axis).Face_X(it.face));
                else u_avg=(face_velocities_ab.Component(axis)(first_cell_u)+face_velocities_ab.Component(axis)(second_cell_u))/2;
                u_avg_keep(axis)(it.face)=u_avg;
                momentum_flux(axis)(it.face)=density_flux_full(axis)(it.face)*u_avg;
            }
        }

        // Apply momentum flux
        Total_Momentum(rho_u_bd,"before advection (rho_u_bd)",time);
        ARRAY<T,FACE_INDEX<TV::m> > flux_div(grid.Cell_Indices());
        for(int axis=0;axis<TV::m;axis++)
        {
            for(CELL_ITERATOR<TV> it(face_grids(axis));it.Valid();it.Next())
            {
                T div=0;
                for(int a=0;a<TV::m;a++)
                {
                    T dx=face_grids(axis).dX(a);
                    div+=(momentum_flux(axis)(it.Full_Second_Face_Index(a))-momentum_flux(axis)(it.Full_First_Face_Index(a)))/dx;
                }
                rho_u_bd.Component(axis)(it.index)-=dt_hat*div;
                flux_div.Component(axis)(it.index)=div;
                
            }
        }
        
        Total_Momentum(rho_u_bd,"after advection (rho_u_bd)",time);
    }

    Rename(rho_u_star,rho_u_bd);
    Invalidate_Ghost(rho_u_star,face_ghost); // No need for ghost data past here
    Invalidate_Ghost(rho_u_ab,face_ghost); // No need for ghost data past here (rho_u_ab no longer used?)
    Test_Analytic_TV(rho_u_star,"rho_u_star",time); Keep(rho_u_star_keep,rho_u_star);

    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        rho_u_star(it.face)=rho_u_star(it.face)/density(it.face);
    Rename(u_star,rho_u_star);
    Invalidate_Ghost(u_star,face_ghost); // No need for ghost data past here ?? Do I actually need this for momentum fix?
    Update_Display_Velocity(u_star);
    Print_Velocity_Stats(u_star,"after momentum advection");
    Test_Analytic_TV(u_star,"u_star",time); Keep(u_star_keep,u_star);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after momentum advection",2);
        
    // Add Forces
    // NEW: Add Forces -> Add Prediction Forces (time n)
    Compute_Divergence(u_star,"before forces");
    Total_Momentum(u_star,density,"before forces",time);
    ARRAY<T,FACE_INDEX<TV::m> > p_forces(grid.Cell_Indices(),init_new,(T)0);
    if(balanced&&0) Add_Forces(p_forces,gamma_old,density_old,dt_hat);
    else Add_Forces(p_forces,gamma,density,dt_hat);
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        u_star(it.face)+=p_forces(it.face); // from u_star to u_star_star
    Rename(u_star_star,u_star);

    Compute_Divergence(u_star_star,"after forces");
    Invalidate_Ghost(u_star_star,face_ghost); // No need for ghost data here
    Test_Analytic_TV(u_star_star,"u_star_star",time); Keep(u_star_star_keep,u_star_star);
    Total_Momentum(u_star_star,density,"after forces",time);
    Update_Display_Velocity(u_star_star);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after add forces",2);
    
    if(use_viscosity)
    {
        Add_Viscosity(u_star_star,face_velocities_ab,dt_hat);
        Test_Analytic_TV(u_star_star,"u_star_star_visc",time); Keep(u_star_star_visc_keep,u_star_star);
    }
    Update_Display_Velocity(u_star_star);
    Total_Momentum(u_star_star,density,"after add viscosity",time);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after add viscosity",2);
        
    //face_velocities_ab not used past here
    Invalidate_Ghost(face_velocities_ab,face_ghost); // No need for ghost data here
        
    // Subtract pressure gradient from u_star_star
    ARRAY<T,FACE_INDEX<TV::m> > grad_p(grid.Cell_Indices(face_ghost),init_new,1e20);
    Gradient(grad_p,pressure,face_ghost);
    Test_Analytic_TV(grad_p,"grad_p",time);

    // NEW: Add Pressure Gradient (time n) over Density (time n)
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        if(balanced&&0) u_star_star(it.face)+=-(dt_hat/density_old(it.face))*grad_p(it.face);
        else u_star_star(it.face)+=-(dt_hat/density(it.face))*grad_p(it.face);
    }
    Test_Analytic_TV(u_star_star,"u_visc_prev",time); Keep(u_visc_prev_keep,u_star_star);

    Update_Display_Velocity(u_star_star);
    Total_Momentum(u_star_star,density,"after subtract pressure gradient",time);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after subtract pressure gradient",2);

    if(use_viscosity)
    {
        Solve_Viscosity(u_star_star,dt_hat);
        Test_Analytic_TV(u_star_star,"u_visc",time); Keep(u_visc_keep,u_star_star);
    }
    Rename(u_visc,u_star_star);

    if(balanced&&0)
    {
        ARRAY<T,FACE_INDEX<TV::m> > new_forces(grid.Cell_Indices(),init_new,(T)0);
        Add_Forces(new_forces,gamma,density,dt_hat);
        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            // NEW: Remove Prediction Forces (time n)
            // NEW: Remove Pressure Gradient (time n) over Density (time n)
            u_visc(it.face)-=p_forces(it.face)-(dt_hat/density_old(it.face))*grad_p(it.face); 
            // NEW: Add Forces (time n+1)           
            // NEW: Add Pressure Gradient (time n) over Density (time n+1)
            u_visc(it.face)+=new_forces(it.face)-(dt_hat/density(it.face))*grad_p(it.face);            
        }
    }
    Rename(u_visc_star,u_visc);

    // Pressure Projection
    Update_Display_Velocity(u_visc_star);
    Total_Momentum(u_visc_star,density,"after viscosity solve",time);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after viscosity solve",2);
    // Total_Momentum(u_visc_star,density,"before project",time);
    Compute_Divergence(u_visc_star,"before projection");
    pressure_old=pressure;        
        
    if(solve_pressure) Project(u_visc_star,density,dt_hat,time); // from u_star to u at n+1
    else projection.p.Fill(0);
    Compute_Divergence(u_visc_star,"after projection");
        
    ARRAY<T,IV> p_diff(grid.Cell_Indices(),init_new,1e20);
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        pressure(it.index)=pressure_old(it.index)+projection.p(it.index);
        p_diff(it.index)=projection.p(it.index);
    }
    Fill_Ghost(pressure,pressure,ghost,[this](const IV& index,int side){
        // Fill ghost region using linear extrapolation
        if(symbolic_source) return (side%2?1:-1)*symbolic->Get_TV("grad_p2",grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
        else
        {
            int axis=side/2;
            int sign=(side%2*2)-1;
            IV a=Fine_Node_To_Cell(index.Add_Axis(axis,-sign));
            IV b=a.Add_Axis(axis,-sign);
            return (pressure(a)-pressure(b))/grid.dX(axis);
        }
    });
    Test_Analytic_T(p_diff,"p_diff",time+dt); Keep(p_diff_keep,p_diff);
    Test_Analytic_T(pressure,"p2",time+dt,0,"p2",true);

    face_velocities_old.array=face_velocities.array; // saving u at n
    face_velocities.array=u_visc_star.array; // updating u at n to n+1
    Ghost_Fill(face_velocities,face_velocities,face_ghost,"u2","grad_u2","grad_v2","grad_w2");
    Test_Analytic_TV(face_velocities,"u2",time+dt);
    Total_Momentum(face_velocities,density,"step end",time+dt);
    Print_Total_Energy(gamma,face_velocities,time+dt);
    Update_Display_Velocity(face_velocities);

    Print_Velocity_Stats(face_velocities,"end step");
    
    TV max_u;
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        if(abs(face_velocities(it.face))>abs(max_u(it.face.axis))) max_u(it.face.axis)=face_velocities(it.face);
    LOG::cout<<"max velocities: "<<max_u<<std::endl;

    ARRAY<T,FACE_INDEX<TV::m> > u_diff(grid.Cell_Indices(face_ghost));
    u_diff.array=face_velocities.array-face_velocities_old.array;
    Test_Analytic_TV(u_diff,"u_diff",time+dt);
    
#undef Rename

    time+=dt;
    if(first_step) first_step=false;
    LOG::cout<<std::endl;

    PHYSBAM_DEBUG_WRITE_SUBSTEP("end time step %i",1,substep+1);
    PHYSBAM_ASSERT(max_u.Magnitude()<speed_limit);
}
//#####################################################################
// Function Discrete_Advection
// Z needs 3 ghost cells
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Discrete_Advection(const GRID<TV>& grid,ARRAY<T,FACE_INDEX<TV::m> >& out,const ARRAY<T,FACE_INDEX<TV::m> >& Z_face,const ARRAY<T,FACE_INDEX<TV::m> >& u_face,T dt,T time)
{
    ARRAY<TV,FACE_INDEX<TV::m> > V_face(grid.Cell_Indices(3));

    for(FACE_ITERATOR<TV> it(grid,3);it.Valid();it.Next())
        V_face(it.Full_Index())=avg.Face_To_Face_Vector(grid,it.Full_Index(),u_face);

    for(int i=0;i<TV::m;i++)
    {
        GRID<TV> node_grid(grid.Get_Face_Grid(i));
        // UPDATE_ADVECTION_EQUATION_HELPER<TV::m>::Apply(*this,node_grid,Z.Component(i),Z_ghost.Component(i),V_face.Component(i),dt,time,flux?&(*flux)(i):0);

        WENO_INTERPOLATION_CELL<TV,T> interp;
        ARRAY<T,FACE_INDEX<TV::m> > weno_Z(node_grid.Cell_Indices());
        for(FACE_ITERATOR<TV> it(node_grid);it.Valid();it.Next())
            weno_Z(it.face)=interp.Interpolate_Cell(grid,Z_face.Component(i),node_grid.Face_X(it.face));
    
        for(CELL_ITERATOR<TV> it(node_grid);it.Valid();it.Next())
        {
            T div=0;
            for(int axis=0;axis<TV::m;axis++)
            {
                FACE_INDEX<TV::m> first=it.Full_First_Face_Index(axis); // same index as it.index
                FACE_INDEX<TV::m> second=it.Full_Second_Face_Index(axis); // larger index
                div+=(V_face(second)(i)*weno_Z(second)-V_face(first)(i)*weno_Z(first))/grid.dX(axis);
            }
            out.Component(i)(it.index)-=dt*div;
        }
    }
}
//#####################################################################
// Function Add_Forces
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Add_Forces(ARRAY<T,FACE_INDEX<TV::m> >& u,ARRAY<T,IV>& gamma_ghost,const ARRAY<T,FACE_INDEX<TV::m> >& rho,T dt)
{
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        u(it.face)+=dt*gravity(it.face.axis);

    if(symbolic&&symbolic_source)
    {
        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
            u(it.face)+=(dt/rho(it.face))*symbolic->Get_TV("src_ns",grid.Face_X(it.face))(it.face.axis);
    }

    if(st_force && ch)
    {
        ARRAY<T,FACE_INDEX<TV::m> > st_force(grid.Cell_Indices());
        std::function<T(T)> g;
        switch(st_type)
        {
            case 2:
                g=[this](T gamma){
                    T a=st_offset;
                    if(gamma<a) return (T)0;
                    if(gamma>1-a) return (T)1;
                    return sqr(-gamma+a)*(2*gamma+4*a-3)/pow<3>(-1+2*a);
                };
                break;
            case 3:
                g=[](T gamma){return (gamma>0.5)?1:0;};
                break;
            case 4:
                g=[](T gamma){return (1+tanh(2*(2*gamma-1)))/2;};
                break;
            case 6:
                g=[this](T gamma){
                    T a=st_offset;
                    T b=st_offset2;
                    if(gamma<a) return (T)0;
                    if(gamma>b) return (T)1;
                    return sqr(a-gamma)*(a-3*b+2*gamma)/pow<3>(a-b);
                };
                break;
            default:
                g=[](T gamma){return gamma;};
        }
        Compute_Surface_Tension(st_force,gamma_ghost,st_type,g,curvature,display_curvature);
        if(display_curvature)
        {
            Flush_Frame("curvature");
            for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                Add_Arrow(grid.Face_X(it.face),TV().Add_Axis(it.face.axis,st_force(it.face)),VECTOR<T,3>(1,0.2,0.2));
            Flush_Frame("surface tension face");
            for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                Add_Arrow(grid.Cell_X(it.index),avg.Face_To_Cell_Vector(grid,it.index,st_force),VECTOR<T,3>(1,0.2,0.2));
            Flush_Frame("surface tension cell vector");
        }
        if(dump_st_data)
        {
            ARRAY<T,FACE_INDEX<TV::m> > grad_gamma(grid.Cell_Indices());
            Gradient(grad_gamma,gamma_ghost,0);

            grad_gamma_keep=grad_gamma;
            st_force_keep=st_force;
        }

        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
            u(it.face)+=(dt/rho(it.face))*st_force(it.face);
    }
}
//#####################################################################
// Function Compute_Density
// Uses ghost+1 ghost cells for gamma
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Compute_Density(ARRAY<T,FACE_INDEX<TV::m> >& density,const ARRAYS_ND_BASE<T,IV>& gamma_ghost,int ghost)
{
    ARRAY<T,IV> density_cell(gamma_ghost.domain);
    Compute_Density(density_cell,gamma_ghost);
    for(FACE_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
        density(it.face)=(density_cell(it.First_Cell_Index())+density_cell(it.Second_Cell_Index()))/2;
    Compute_Mass(density_cell,gamma_ghost);
}
//#####################################################################
// Function Compute_Density
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Compute_Density(ARRAY<T,IV>& density,const ARRAYS_ND_BASE<T,IV>& gamma)
{
    // Linear interpolation from rho0 to rho1 based on gamma
    for(RANGE_ITERATOR<TV::m> it(gamma.domain);it.Valid();it.Next())
        density(it.index)=rho0+(rho1-rho0)*gamma(it.index);
}
//#####################################################################
// Function Compute_Mass
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Compute_Mass(const ARRAY<T,IV>& density_cell,const ARRAYS_ND_BASE<T,IV>& gamma)
{
    if(!chh||!chh->volume){mass=density_cell; return;}
    for(RANGE_ITERATOR<TV::m> it(gamma.domain);it.Valid();it.Next())
    {
        mass(it.index)=density_cell(it.index)*((*(chh->volume))(it.index));
        mass0(it.index)=rho0*(1-gamma(it.index))*((*(chh->volume))(it.index));
        mass1(it.index)=rho1*gamma(it.index)*((*(chh->volume))(it.index));
    }
}
//#####################################################################
// Function Compute_Viscosity
// Uses ghost+1 ghost cells for gamma
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Compute_Viscosity(ARRAY<T,FACE_INDEX<TV::m> >& viscosity,const ARRAYS_ND_BASE<T,IV>& gamma,int input_ghost)
{
    // Linear interpolation from mu0 to mu1 based on gamma
    for(FACE_ITERATOR<TV> it(grid,input_ghost);it.Valid();it.Next())
    {
        T g=(gamma(it.First_Cell_Index())+gamma(it.Second_Cell_Index()))/2;
        viscosity(it.face)=mu0*(1-g)+mu1*g;
    }
}
//#####################################################################
// Function Add_Viscosity
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Add_Viscosity(ARRAY<T,FACE_INDEX<TV::m> >& u_out,const ARRAY<T,FACE_INDEX<TV::m> >& u_in,T dt)
{
    //face_to_cell(x or y, cell_index)
    // average face neighbors to cell in either x or y axis
    auto face_to_cell=[](const ARRAY<T,FACE_INDEX<TV::m> >& face_array,int axis,IV index)
    {
        FACE_INDEX<TV::m> left_neighbor(axis,index);
        FACE_INDEX<TV::m> right_neighbor(axis,index.Add_Axis(axis,1));
        return (face_array(left_neighbor)+face_array(right_neighbor))/2;
    };
    //face_to_node(x or y, node_index? corner?)
    // average faces to node position in either x or y axis
    auto face_to_node=[](const ARRAY<T,FACE_INDEX<TV::m> >& face_array,int axis,IV index)
    {
        FACE_INDEX<TV::m> left_neighbor(axis,index.Add_Axis(1-axis,-1));
        FACE_INDEX<TV::m> right_neighbor(axis,index);
        return (face_array(left_neighbor)+face_array(right_neighbor))/2;
    };
    
    // This only works in 2D
    VECTOR<ARRAY<T,IV>,TV::m> out;
    for(int axis=0;axis<TV::m;axis++)
        out(axis).Resize(grid.Cell_Indices(1),init_new,1e20);
    for(CELL_ITERATOR<TV> it(grid,1);it.Valid();it.Next())
    {
        // for(int axis=0;axis<TV::m;axis++)
        for(int axis=0;axis<2;axis++)
        {
            T idx=grid.one_over_dX(axis);
            T idy=grid.one_over_dX(1-axis);
            T out_x=idx*(viscosity(it.Full_Second_Face_Index(axis))*(face_to_cell(u_in,axis,it.index.Add_Axis(axis,1))-face_to_cell(u_in,axis,it.index))*idx-viscosity(it.Full_First_Face_Index(axis))*(face_to_cell(u_in,axis,it.index)-face_to_cell(u_in,axis,it.index.Add_Axis(axis,-1)))*idx);
            out_x+=idy*(viscosity(it.Full_Second_Face_Index(1-axis))*(face_to_node(u_in,1-axis,it.index.Add_Axis(axis,1).Add_Axis(1-axis,1))-face_to_node(u_in,1-axis,it.index.Add_Axis(1-axis,1)))*idx-viscosity(it.Full_First_Face_Index(1-axis))*(face_to_node(u_in,1-axis,it.index.Add_Axis(axis,1))-face_to_node(u_in,1-axis,it.index))*idx);

            out(axis)(it.index)=out_x;
        }
    }
    VECTOR<ARRAY<T,IV>,TV::m> test_out;
    for(int axis=0;axis<TV::m;axis++)
        test_out(axis).Resize(grid.Cell_Indices(),init_new,1e20);
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        test_out(0)(it.index)=out(0)(it.index);
        test_out(1)(it.index)=out(1)(it.index);
    }
    Test_Analytic_TV(out,"visc",time);

    {
        T div_u=0;
        T max_abs_div_u=0;
        for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            T d=0;
            for(int a=0;a<TV::m;a++)
            {
                FACE_INDEX<TV::m> f0(a,it.index);
                FACE_INDEX<TV::m> f1(f0);
                f1.index(a)++;
                d+=(u_out(f1)-u_out(f0))*grid.one_over_dX(a);
            }
            div_u+=d;
            max_abs_div_u=max(abs(d),max_abs_div_u);
        }
        if(print_velocity_stats) LOG::cout<<"div_u (before visc) sum: "<<div_u<<"\t max abs div_u:"<<max_abs_div_u<<"\t div_grad_t_u sum: "<<test_out(0).Sum()<<" "<<test_out(1).Sum()<<std::endl;
    }

    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        u_out(it.face)+=dt*avg.Cell_To_Face(grid,it.face,out(it.face.axis))/density(it.face);
}
//#####################################################################
// Function Solve Viscosity
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Solve_Viscosity(ARRAY<T,FACE_INDEX<TV::m> >& u,T dt)
{
    PCG_OPTIONS<T> options;
    options.max_iterations=1000;
    options.restart_iterations=200;
    options.show_results=print_solve_results;
    options.show_residual=print_solve_residual;
    options.ensure_convergence=true;
    options.absolute_tolerance=use_mu_abs_tol?mu_abs_tolerance:abs_tolerance;
    options.relative_tolerance=use_mu_tol?mu_solver_tolerance:solver_tolerance;

    ARRAY<T,FACE_INDEX<TV::m> > test_range(grid.Cell_Indices(),init_new,1e20);
    
    VECTOR<ARRAY<bool,FACE_INDEX<TV::m> >,TV::m> psi_Ns;
    VECTOR<ARRAY<T,FACE_INDEX<TV::m> >,TV::m> psi_Ns_values;
    VECTOR<ARRAY<bool,IV>,TV::m> psi_Ds;
    VECTOR<ARRAY<T,IV>,TV::m> psi_Ds_values;
    for(int axis=0;axis<TV::m;axis++)
    {
        GRID<TV> solve_grid=periodic(axis)?grid:grid.Get_Face_MAC_Grid(axis);
        HELMHOLTZ<TV> helmholtz(solve_grid,false);
        helmholtz.Set_Periodic_Boundaries(periodic);
        helmholtz.print_system=write_viscosity;
        helmholtz.octave_output=LOG::sprintf("viscosity-system-%P",axis);
        
        RANGE<IV> axis_range=test_range.Component(axis).domain;
        if(periodic(axis)) axis_range.max_corner(axis)-=1; // TODO: I don't really like this fix. It appears to be working right, though.
        psi_Ns(axis).Resize(axis_range,init_new,false);
        psi_Ns_values(axis).Resize(axis_range,init_new,1e20);
        psi_Ds(axis).Resize(axis_range.Thickened(1),init_new,false);
        psi_Ds_values(axis).Resize(axis_range.Thickened(1),init_new,1e20);

        for(int side=0;side<TV::m*2;side++)
        {
            // if(periodic(side/2)) continue;
            if(side_bc_type(side)==BC_PERIODIC) continue;
            // for(FACE_RANGE_ITERATOR<TV::m> it(axis_range,0,0,RF::ghost,side);it.Valid();it.Next())
            // {
                    
            //     psi_Ns(axis)(it.face)=true;
                    
            //     if(symbolic_source)
            //     {
            //         if(axis==0) psi_Ns_values(axis)(it.face)=(it.side%2?1:-1)*symbolic->Get_TV("grad_u_visc",solve_grid.Face_X(it.face))(it.face.axis);
            //         else if(axis==1) psi_Ns_values(axis)(it.face)=(it.side%2?1:-1)*symbolic->Get_TV("grad_v_visc",solve_grid.Face_X(it.face))(it.face.axis);
            //         else psi_Ns_values(axis)(it.face)=(it.side%2?1:-1)*symbolic->Get_TV("grad_w_visc",solve_grid.Face_X(it.face))(it.face.axis);
            //     }
            //     else psi_Ns_values(axis)(it.face)=0;
            // }
            else if(side/2==axis)
            {
                for(RANGE_ITERATOR<TV::m> it(axis_range,0,-1,RI::ghost|RI::duplicate_corners,side);it.Valid();it.Next())
                {
                    psi_Ds(axis)(it.index)=true;
                    if(symbolic_source)
                        psi_Ds_values(axis)(it.index)=symbolic->Get_TV("u_visc",solve_grid.Cell_X(it.index))(axis);
                    else psi_Ds_values(axis)(it.index)=0;
                }
            }
            else
            {
                if(side_bc_type(side)==BC_SLIP)
                {
                    for(FACE_RANGE_ITERATOR<TV::m> it(axis_range,0,0,RF::ghost,side);it.Valid();it.Next())
                    {
                    
                        psi_Ns(axis)(it.face)=true;
                    
                        if(symbolic_source)
                        {
                            if(axis==0) psi_Ns_values(axis)(it.face)=(it.side%2?1:-1)*symbolic->Get_TV("grad_u_visc",solve_grid.Face_X(it.face))(it.face.axis);
                            else if(axis==1) psi_Ns_values(axis)(it.face)=(it.side%2?1:-1)*symbolic->Get_TV("grad_v_visc",solve_grid.Face_X(it.face))(it.face.axis);
                            else psi_Ns_values(axis)(it.face)=(it.side%2?1:-1)*symbolic->Get_TV("grad_w_visc",solve_grid.Face_X(it.face))(it.face.axis);
                        }
                        else psi_Ns_values(axis)(it.face)=0;
                    }
                }
                else if(side_bc_type(side)==BC_NOSLIP)
                {
                    for(RANGE_ITERATOR<TV::m> it(axis_range,1,0,RI::ghost|RI::duplicate_corners,side);it.Valid();it.Next())
                    {
                        psi_Ds(axis)(it.index)=true;
                        if(symbolic_source)
                            psi_Ds_values(axis)(it.index)=symbolic->Get_TV("u_visc",solve_grid.Cell_X(it.index))(axis);
                        else psi_Ds_values(axis)(it.index)=0;
                    }
                }
            }
        }

        helmholtz.Set_Neumann_Boundaries(psi_Ns(axis),&psi_Ns_values(axis));
        helmholtz.Set_Dirichlet_Boundaries(psi_Ds(axis),&psi_Ds_values(axis));

        ARRAY<T,IV> a(axis_range);
        ARRAY<T,IV> rhs(axis_range);
        for(RANGE_ITERATOR<TV::m> it(axis_range);it.Valid();it.Next())
        {
            a(it.index)=density.Component(axis)(it.index);
            rhs(it.index)=u.Component(axis)(it.index)*density.Component(axis)(it.index);
        }
        
        ARRAY<T,FACE_INDEX<TV::m> > b(axis_range);
        for(FACE_RANGE_ITERATOR<TV::m> it(axis_range);it.Valid();it.Next())
        {
            IV up_cell(it.face.index);
            IV down_cell(it.face.index.Add_Axis(it.face.axis,-1));
            b(it.face)=0.5*dt*(viscosity.Component(axis)(it.face.index.Add_Axis(it.face.axis,1))+viscosity.Component(axis)(it.face.index));
        }

        helmholtz.Set_Alpha(a);
        helmholtz.Set_Beta(b);

        if(print_status) LOG::cout<<"Solving "<<(axis?"second":"first")<<" viscosity equation..."<<std::endl;
        bool converged=helmholtz.Solve(u.Component(axis),rhs,axis_range,options);
        if(require_convergence) PHYSBAM_ASSERT(converged);

        if(periodic(axis))
        {
            for(RANGE_ITERATOR<TV::m> it(axis_range,0,-1,RI::duplicate_corners|RI::ghost,2*axis);it.Valid();it.Next())
            {
                IV opp=it.index;
                opp(axis)=axis_range.max_corner(axis);
                u.Component(axis)(opp)=u.Component(axis)(it.index);
            }
        }
    }
}
//#####################################################################
// Function Project
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Project(ARRAY<T,FACE_INDEX<TV::m> >& u,const ARRAY<T,FACE_INDEX<TV::m> >& rho,T dt,T time)
{
    if(ch)
    {
        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
            projection.poisson->beta_face(it.face)=dt/rho(it.face);
    }
    else projection.poisson->beta_face.Fill(2/(rho0+rho1));

    Enforce_Boundary_Conditions(u);  // fix boundary u
    Update_Display_Velocity(u);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("before apply boundary in project",3);
    boundary_scalar_p.Apply_Boundary_Condition_Face(grid,u);
    Update_Display_Velocity(u);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after apply boundary in project",3);
    if(print_status) LOG::cout<<"Solving pressure equation..."<<std::endl;
    projection.Make_Divergence_Free(u,dt,time);
}
//#####################################################################
// Function Write_Output_Files
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Write_Output_Files(const std::string& title)
{
    if(!suppress_output)
    {   
        viewer_output.Flush_Frame(title.c_str());
        if((current_frame+1)%old_write_frequency==0)
        {
            // write _old files
            Write_To_File(stream_type,viewer_dir.Frame_File("gamma_old"),gamma_old);
            Write_To_File(stream_type,viewer_dir.Frame_File("face_velocities_old"),face_velocities_old);
            Write_To_File(stream_type,viewer_dir.Frame_File("pressure_old"),pressure_old);
        }
    }
}
//#####################################################################
// Function Update_Display_Gamma
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Update_Display_Gamma(const ARRAY<T,IV>& input)
{
    for(RANGE_ITERATOR<TV::m> it(display_gamma.domain);it.Valid();it.Next())
        display_gamma(it.index)=input(it.index);
}
//#####################################################################
// Function Update_Display_Velocity
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Update_Display_Velocity(const ARRAY<T,FACE_INDEX<TV::m> >& input)
{
    for(FACE_RANGE_ITERATOR<TV::m> it(display_velocities.domain);it.Valid();it.Next())
        display_velocities(it.face)=input(it.face);
}
//#####################################################################
// Function Read_Output_Files
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Read_Output_Files()
{
    viewer_dir.Set(current_frame);
    if(second_order_time)
    {    
        std::string filename;
        filename=viewer_dir.Frame_File("gamma_old");
        PHYSBAM_ASSERT(File_Exists(filename));
        if(print_status) LOG::cout<<"Reading gamma_old from \""<<filename<<"\""<<std::endl;
        Read_From_File(filename,gamma_old);
        Compute_Density(density_old,gamma_old,face_ghost);

        filename=viewer_dir.Frame_File("face_velocities_old");
        PHYSBAM_ASSERT(File_Exists(filename));
        if(print_status) LOG::cout<<"Reading face_velocities_old from \""<<filename<<"\""<<std::endl;
        Read_From_File(filename,face_velocities_old);
        
        filename=viewer_dir.Frame_File("pressure_old");
        PHYSBAM_ASSERT(File_Exists(filename));
        if(print_status) LOG::cout<<"Reading pressure_old from \""<<filename<<"\""<<std::endl;
        Read_From_File(filename,pressure_old);
    }
    
    std::string filename;
    filename=viewer_dir.Frame_File("gamma");
    PHYSBAM_ASSERT(File_Exists(filename));
    if(print_status) LOG::cout<<"Reading gamma from \""<<filename<<"\""<<std::endl;
    Read_From_File(filename,gamma);
    Compute_Density(density,gamma,face_ghost);
    Compute_Viscosity(viscosity,gamma,face_ghost);
    Update_Display_Gamma(gamma);
        
    filename=viewer_dir.Frame_File("curvature");
    PHYSBAM_ASSERT(File_Exists(filename));
    if(print_status) LOG::cout<<"Reading curvature from \""<<filename<<"\""<<std::endl;
    Read_From_File(filename,curvature_display);

    filename=viewer_dir.Frame_File("face_velocities");
    PHYSBAM_ASSERT(File_Exists(filename));
    if(print_status) LOG::cout<<"Reading face_velocities from \""<<filename<<"\""<<std::endl;
    Read_From_File(filename,face_velocities);
    Update_Display_Velocity(face_velocities);
        
    filename=viewer_dir.Frame_File("pressure");
    PHYSBAM_ASSERT(File_Exists(filename));
    if(print_status) LOG::cout<<"Reading pressure from \""<<filename<<"\""<<std::endl;
    Read_From_File(filename,pressure);
}
//#####################################################################
// Function Invalidate_Ghost (cells)
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Invalidate_Ghost(ARRAY<T,IV>& output,int ghost) const
{
    if(!invalidate_ghost) return;
    for(CELL_ITERATOR<TV> it(grid,ghost,GHOST_REGION);it.Valid();it.Next())
        output(it.index)=(T)1e20;
}
//#####################################################################
// Function Invalidate_Ghost (faces)
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Invalidate_Ghost(ARRAY<T,FACE_INDEX<TV::m> >& output,int ghost) const
{
    if(!invalidate_ghost) return;
    for(FACE_ITERATOR<TV> it(grid,ghost,GHOST_REGION);it.Valid();it.Next())
        output(it.face)=(T)1e20;
}
//#####################################################################
// Function Ghost_Fill (cells) (strings)
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Ghost_Fill(ARRAY<T,IV>& output,const ARRAY<T,IV>& input,int ghost,std::string n) const 
{
    if(symbolic_source) Fill_Ghost(output,input,ghost,
        [this,n](const IV& index,int side){
            return (side%2?1:-1)*symbolic->Get_TV(n,grid.Face_X(Fine_Node_To_Face(index,side/2)))(side/2);
        });
    else Fill_Ghost(output,input,ghost);
}
//#####################################################################
// Function Ghost_Fill (faces) (strings)
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Ghost_Fill(ARRAY<T,FACE_INDEX<TV::m> >& output,const ARRAY<T,FACE_INDEX<TV::m> >& input,int ghost,std::string d,std::string n0,std::string n1,std::string n2) const 
{
    if(symbolic_source) Fill_Ghost(output,input,ghost,
        [this,d](const IV& index,int side,int face_axis){
            return symbolic->Get_TV(d,grid.Face_X(Fine_Node_To_Face(index,face_axis)))(face_axis);
        },
        [this,n0,n1,n2](const IV& index,int side,int face_axis){
            int n=(side%2?1:-1);
            int axis=side/2;
            TV X=grid.Face_X(Fine_Node_To_Face(index,face_axis));
            if(face_axis==0) return n*symbolic->Get_TV(n0,X)(axis);
            else if(face_axis==1) return n*symbolic->Get_TV(n1,X)(axis);
            else return n*symbolic->Get_TV(n2,X)(axis);
        });
    else Fill_Ghost(output,input,ghost);
}
//#####################################################################
// Function Fill_Ghost (faces)
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Fill_Ghost(ARRAY<T,FACE_INDEX<TV::m> >& output,const ARRAY<T,FACE_INDEX<TV::m> >& input,int ghost,std::function<T(const IV&,int,int)> boundary_D,std::function<T(const IV&,int,int)> boundary_N) const 
{
    ARRAY<T,FACE_INDEX<TV::m> >::Put(input,output); // interior
    for(int side=0;side<TV::m*2;side++)
    {
        if(side_bc_type(side)==BC_PERIODIC) continue;
        else if(side_bc_type(side)==BC_SLIP)
        {
            if(boundary_N && boundary_D) Reflection_Face_Slip_Side<TV,T>(grid,output,ghost,side,boundary_D,boundary_N,true);
            else Reflection_Face_Slip_Side<TV,T>(grid,output,ghost,side,0,0,true);
        }
        else if(side_bc_type(side)==BC_NOSLIP)
        {
            if(boundary_D) Reflection_Face_NoSlip_Side<TV,T>(grid,output,ghost,side,boundary_D,true);
            else Reflection_Face_NoSlip_Side<TV,T>(grid,output,ghost,side,0,true);
        }
        else PHYSBAM_FATAL_ERROR("Invalid Boundary Type");
    }
    boundary_scalar_p.Fill_Ghost_Faces(grid,input,output,ghost);    
}
//#####################################################################
// Function Fill_Ghost (scalar cells)
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Fill_Ghost(ARRAY<T,IV>& output,const ARRAY<T,IV>& input,int ghost,std::function<T(const IV&,int)> boundary_N) const 
{
    ARRAY<T,IV>::Put(input,output); // interior
    for(int side=0;side<TV::m*2;side++)
    {
        if(side_bc_type(side)==BC_PERIODIC) continue;
        Reflection_Cell_Neumann_Side<TV,T>(grid,output,ghost,side,boundary_N);
    }
    boundary_scalar_p.Fill_Ghost_Cells(grid,input,output,ghost);
}
//#####################################################################
// Function Gradient
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Gradient(ARRAY<T,FACE_INDEX<TV::m> >& out,const ARRAYS_ND_BASE<T,IV>& in_ghost,int ghost) const
{
    for(FACE_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next())
        out(it.face)=(in_ghost(it.face.Second_Cell_Index())-in_ghost(it.face.First_Cell_Index()))*grid.one_over_dX(it.face.axis);
}
//#####################################################################
// Function Gradient (Ghostless)
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Gradient(ARRAY<T,FACE_INDEX<TV::m> >& out,const ARRAY<T,IV>& in) const
{
    for(FACE_RANGE_ITERATOR<TV::m> it(grid.Cell_Indices());it.Valid();it.Next())
    {
        IV c0=it.face.First_Cell_Index();
        IV c1=it.face.Second_Cell_Index();
        if(side_bc_type(it.side)==BC_PERIODIC)
        {
            c1=wrap(c1,IV(),grid.numbers_of_cells);
            c0=wrap(c0,IV(),grid.numbers_of_cells);
        }
        if(grid.Valid_Cell(c1) && grid.Valid_Cell(c0))
            out(it.face)=(in(c1)-in(c0))*grid.one_over_dX(it.face.axis);
        else out(it.face)=0;
    }
}
//#####################################################################
// Function Compute_Surface_Tension
// Requires ghost data for gamma to at least 2 cells
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Compute_Surface_Tension(ARRAY<T,FACE_INDEX<TV::m> >& st_force,ARRAY<T,IV>& gamma_ghost,int type,std::function<T(T)> g,int curvature,bool display_curvature) const
{
    if(type==0)
    {
        // Discretization of Surface Tension from divergence of outer product of gradients?

        // x face configuration
        // 0 G A 0
        // D FXH E
        // 0 C B 0
        T dx3=pow(grid.dX(0),(T)3);
        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            IV iH=it.face.index;
            IV iA=iH.Add_Axis(1-it.face.axis,1);
            IV iB=iH.Add_Axis(1-it.face.axis,-1);
            IV iF=iH.Add_Axis(it.face.axis,-1);
            IV iC=iF.Add_Axis(1-it.face.axis,-1);
            IV iD=iH.Add_Axis(it.face.axis,-2);
            IV iE=iH.Add_Axis(it.face.axis,1);
            IV iG=iF.Add_Axis(1-it.face.axis,1);
            T A=gamma_ghost(iA);
            T B=gamma_ghost(iB);
            T C=gamma_ghost(iC);
            T D=gamma_ghost(iD);
            T E=gamma_ghost(iE);
            T F=gamma_ghost(iF);
            T G=gamma_ghost(iG);
            T H=gamma_ghost(iH);
            
            T LF=D+H+G+C-4*F;
            T LH=E+F+A+B-4*H;
            T diff=(2*(H-F)*(LF+LH)-(B-F)*(A-F)+(G-H)*(C-H))/(2*dx3);
            st_force(it.face)+=-0.125*ch->lambda*diff;
        }
    }
    else if(type==1)
    {
        auto df0=[this](T x, T r){
            PHYSBAM_ASSERT(x!=0);
            if(cho->alt_F) return 6*(x-r)*(2*r*r-4*r*x+x)/x;
            return 6*r*(1-2*r)*(1-r/x);
        };
        auto df1=[this](T x,T r){
            return -2*(r-x)*(3-2*r-4*x)-4*sqr(r-x);
        };
        auto dF=[this,df0,df1](T g){
            T eta2=ch->eta*ch->eta;
            if(cho)
            {
                T fscale=cho->F_scale;
                if(g<=cho->r) // assumes x>=0
                    return fscale*df0(g,cho->r)/eta2;
                else if(g<(T)0.5)
                    return fscale*df1(g,cho->r)/eta2;
                else if(g<=(T)(1-cho->r))
                    return -fscale*df1(1-g,cho->r)/eta2;
                else // assumes x<=1
                    return -fscale*df0(1-g,cho->r)/eta2;
            }
            else return (4*g*g*g-6*g*g+2*g)/eta2;
        };
        
        // compute divergence of gradient of gamma
        ARRAY<T,IV> xi(grid.Cell_Indices(1));
        for(CELL_ITERATOR<TV> it(grid,1);it.Valid();it.Next())
        {
            T lap_g=0;
            for(int a=0;a<TV::m;a++)
            {
                T idx2=sqr(grid.one_over_dX(a));
                IV left=it.index.Add_Axis(a,-1);
                IV right=it.index.Add_Axis(a,1);
            
                lap_g+=idx2*((gamma_ghost(right)-gamma_ghost(it.index))-(gamma_ghost(it.index)-gamma_ghost(left)));
            }
            // compute dF of gamma
            // compute xi=lambda*(dF(gamma)-Laplacian(gamma)) 
            xi(it.index)=ch->lambda*(dF(gamma_ghost(it.index))-lap_g);
        }
        
        if(display_curvature)
        {
            ARRAY<T,IV> curv(grid.Cell_Indices(1));
            T max_curve=-1e20;
            T min_curve=1e20;
            for(CELL_ITERATOR<TV> it(grid,1);it.Valid();it.Next())
            {
                curv(it.index)=xi(it.index)/ch->sigma;
                max_curve=max(max_curve,curv(it.index));
                min_curve=min(min_curve,curv(it.index));
            }
        
            for(CELL_ITERATOR<TV> it(grid,1);it.Valid();it.Next())
            {
                if(curv(it.index)>=0) Add_Ball(grid.Cell_X(it.index),curv(it.index),VECTOR<T,3>(0,1,0));
                else Add_Ball(grid.Cell_X(it.index),curv(it.index),VECTOR<T,3>(1,0,0));
            }

            for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
            {
                T face_curve=(xi(it.First_Cell_Index())+xi(it.Second_Cell_Index()))/2/ch->sigma;
                if(face_curve>=0) Add_Ball(grid.Face_X(it.face),face_curve,VECTOR<T,3>(0,1,0));
                else Add_Ball(grid.Face_X(it.face),face_curve,VECTOR<T,3>(1,0,0));
            }
        }

        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            T face_xi=(xi(it.First_Cell_Index())+xi(it.Second_Cell_Index()))/2;
            T left_gamma=gamma_ghost(it.First_Cell_Index());
            T right_gamma=gamma_ghost(it.Second_Cell_Index());
            T grad_g=(right_gamma-left_gamma)/grid.dX(it.face.axis);
            st_force(it.face)=4*face_xi*grad_g;
        }
    }
    else
    {
        // compute gradient of gamma
        ARRAY<T,FACE_INDEX<TV::m> > grad_gamma(grid.Cell_Indices(1));
        Gradient(grad_gamma,gamma_ghost,1);

        if(!g) g=[](T gamma){return gamma;};

        AVERAGING_UNIFORM<TV> avg;

        // div(grad(g)/|grad(g)|)
        ARRAY<TV,FACE_INDEX<TV::m> > grad_V(grid.Cell_Indices(1));
        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            IV cell2=it.Second_Cell_Index();
            IV cell1=it.First_Cell_Index();
            TV grad2=avg.Face_To_Cell_Vector(grid,cell2,grad_gamma).Normalized();
            TV grad1=avg.Face_To_Cell_Vector(grid,cell1,grad_gamma).Normalized();
            grad_V(it.face)=(grad2+grad1)/2;
        }
        
        // compute divergence of gradient of gamma
        ARRAY<T,IV> curv(grid.Cell_Indices(1));

        LEVELSET<TV> level(grid,at_cells,gamma_ghost,5);
        
        for(CELL_ITERATOR<TV> it(grid,1);it.Valid();it.Next())
        {
            if(use_ideal_curv)
            {
                T r=(grid.Cell_X(it.index)-grid.domain.Center()).Magnitude();
                curv(it.index)=2/r;
            }
            else
            {
                T div_grad=0;
                for(int a=0;a<TV::m;a++)
                {
                    // Compute Divergence of Gradient Vectors on Cells
                    FACE_INDEX<TV::m> right_face=it.Full_Second_Face_Index(a);
                    FACE_INDEX<TV::m> left_face=it.Full_First_Face_Index(a);
                    
                    div_grad+=grid.one_over_dX(a)*(grad_V(right_face)(a)-grad_V(left_face)(a));
                }
                // compute dF of gamma
                // compute xi=lambda*(dF(gamma)-Laplacian(gamma))
                switch(curvature%2)
                {
                    case 0:
                        curv(it.index)=-div_grad;
                        break;
                    case 1:
                        curv(it.index)=level.Compute_Curvature(gamma_ghost,it.index);
                        break;
                    default:
                        break;
                }
            }
        }
        if(display_curvature)
        {
            T max_curve=-1e20;
            T min_curve=1e20;
            for(CELL_ITERATOR<TV> it(grid,1);it.Valid();it.Next())
            {
                max_curve=max(max_curve,curv(it.index));
                min_curve=min(min_curve,curv(it.index));
            }       
            for(CELL_ITERATOR<TV> it(grid,1);it.Valid();it.Next())
            {
                if(curv(it.index)>=0) Add_Ball(grid.Cell_X(it.index),curv(it.index),VECTOR<T,3>(0,1,0));
                else Add_Ball(grid.Cell_X(it.index),curv(it.index),VECTOR<T,3>(1,0,0));
            }

            for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
            {
                T face_curve=(curv(it.First_Cell_Index())+curv(it.Second_Cell_Index()))/2/ch->sigma;
                if(face_curve>=0) Add_Ball(grid.Face_X(it.face),face_curve,VECTOR<T,3>(0,1,0));
                else Add_Ball(grid.Face_X(it.face),face_curve,VECTOR<T,3>(1,0,0));
            }
        }

        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            T face_curv=(curv(it.First_Cell_Index())+curv(it.Second_Cell_Index()))/2;   
            T left_gamma=gamma_ghost(it.First_Cell_Index());
            T right_gamma=gamma_ghost(it.Second_Cell_Index());
            T grad_g=(g(right_gamma)-g(left_gamma))/grid.dX(it.face.axis);
            // st_force(it.face)=4*sigma*face_curv*grad_g;
            st_force(it.face)=ch->sigma*face_curv*grad_g;
        }
    }
}
//#####################################################################
// Function Compute_Curvature
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Compute_Curvature()
{
    // Compute Curvature, store it in curvature_display
    // check curvature for type of curvature to compute
    // compute gradient of gamma
    ARRAY<T,FACE_INDEX<TV::m> > grad_gamma(grid.Cell_Indices(face_ghost));
    Gradient(grad_gamma,gamma,face_ghost);

    if(display_curvature)
    {
        AVERAGING_UNIFORM<TV> avg;
        
        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
            Add_Arrow(grid.Face_X(it.face),10*TV().Add_Axis(it.face.axis,grad_gamma(it.face)),VECTOR<T,3>(0,0.7,1));
        Flush_Frame("Gamma Gradient");

        for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            TV grad_V=avg.Face_To_Cell_Vector(grid,it.index,grad_gamma);
            Add_Arrow(grid.Cell_X(it.index),10*grad_V,VECTOR<T,3>(0,0.7,1));
        }
        Flush_Frame("Gamma Gradient (Cells)");        
    }

    // div(grad(g)/|grad(g)|)
    ARRAY<TV,FACE_INDEX<TV::m> > grad_V(grid.Cell_Indices(1));
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        IV cell2=it.Second_Cell_Index();
        IV cell1=it.First_Cell_Index();
        TV grad2=avg.Face_To_Cell_Vector(grid,cell2,grad_gamma).Normalized();
        TV grad1=avg.Face_To_Cell_Vector(grid,cell1,grad_gamma).Normalized();
        grad_V(it.face)=(grad2+grad1)/2;
        if(display_curvature) Add_Arrow(grid.Face_X(it.face),10*grad_V(it.face),VECTOR<T,3>(1,1,0));
    }
    if(display_curvature) Flush_Frame("Gamma Gradient Norm");
        
    // compute divergence of gradient of gamma
    LEVELSET<TV> level(grid,at_cells,gamma,5);
        
    for(CELL_ITERATOR<TV> it(grid,1);it.Valid();it.Next())
    {
        T div_grad=0;
        for(int a=0;a<TV::m;a++)
        {
            // Compute Divergence of Gradient Vectors on Cells
            FACE_INDEX<TV::m> right_face=it.Full_Second_Face_Index(a);
            FACE_INDEX<TV::m> left_face=it.Full_First_Face_Index(a);
                
            div_grad+=grid.one_over_dX(a)*(grad_V(right_face)(a)-grad_V(left_face)(a));
        }
        switch(curvature%2)
        {
            case 0:
                curvature_display(it.index)=-div_grad;
                break;
            case 1:
                curvature_display(it.index)=level.Compute_Curvature(gamma,it.index);
                break;
            default:
                break;
        }
    }

    if(display_curvature && st_force && ch)
    {
        ARRAY<T,FACE_INDEX<TV::m> > st_force(grid.Cell_Indices());
        std::function<T(T)> g;
        switch(st_type)
        {
            case 2:
                g=[](T gamma){
                    T a=0.2;
                    if(gamma<a) return (T)0;
                    if(gamma>1-a) return (T)1;
                    return sqr(-gamma+a)*(2*gamma+4*a-3)/pow(-1+2*a,(T)3);
                };
                break;
            case 3:
                g=[](T gamma){return (gamma>0.5)?1:0;};
                break;
            case 4:
                g=[](T gamma){return (1+tanh(2*(2*gamma-1)))/2;};
                break;
            default:
                g=[](T gamma){return gamma;};
        }
        Compute_Surface_Tension(st_force,gamma,st_type,g,curvature,display_curvature);
        Flush_Frame("curvature");
        for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
            Add_Arrow(grid.Face_X(it.face),TV().Add_Axis(it.face.axis,st_force(it.face)),VECTOR<T,3>(1,0.2,0.2));
        Flush_Frame("surface tension face");
        for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
            Add_Arrow(grid.Cell_X(it.index),avg.Face_To_Cell_Vector(grid,it.index,st_force),VECTOR<T,3>(1,0.2,0.2));
        Flush_Frame("surface tension cell vector");
    }
}
//#####################################################################
// Function Compute_Divergence
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Compute_Divergence(const ARRAY<T,FACE_INDEX<TV::m> >& u,std::string title)
{
    if(!print_velocity_stats) return;
    T div_u=0;
    T max_abs_div_u=0;
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        T d=0;
        for(int a=0;a<TV::m;a++)
        {
            FACE_INDEX<TV::m> f0(a,it.index);
            FACE_INDEX<TV::m> f1(f0);
            f1.index(a)++;
            d+=(u(f1)-u(f0))*grid.one_over_dX(a);
        }
        div_u+=d;
        max_abs_div_u=max(abs(d),max_abs_div_u);
    }
    LOG::cout<<"div_u ("<<title<<") sum: "<<div_u<<"\t max abs div_u:"<<max_abs_div_u<<std::endl;
}
//#####################################################################
// Function Print_Density_Stats
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Print_Density_Stats(const ARRAY<T,FACE_INDEX<TV::m> >& rho,std::string title)
{
    if(!print_density_stats) return;
    T mmin=1e20,mmax=0,sum=0;
    for(FACE_ITERATOR<TV> it(grid,0,INTERIOR_REGION);it.Valid();it.Next())
        sum+=rho(it.face)*grid.Cell_Size();
    
    auto getvolume = [this](FACE_INDEX<TV::m> fi,CAHN_HILLIARD_HELMHOLTZ<TV>* chh) -> T{
        if(!chh||!chh->volume) return grid.Cell_Size();
        IV fi0=fi.First_Cell_Index();
        IV fi1=fi.Second_Cell_Index();
        T a=0,b=0;
        if((*(chh->volume)).domain.Lazy_Inside_Half_Open(fi0))
            a=(*(chh->volume))(fi0);
        if((*(chh->volume)).domain.Lazy_Inside_Half_Open(fi1))
            b=(*(chh->volume))(fi1);
        return (a+b)/2.0;
    };
    for(int axis=0;axis<TV::m;axis++)
    {
        if(side_bc_type(axis*2)==BC_PERIODIC)
            for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis);it.Valid();it.Next())
            {
                T vol=getvolume(it.face,chh);
                sum+=rho(it.face)*vol;
            }
        else
            for(int side=0;side<2;side++)
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis+side);it.Valid();it.Next())
                {
                    T vol=getvolume(it.face,chh);
                    sum+=rho(it.face)*vol/2;
                }
    }

    T dmin=1e20,dmax=0;
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        T vol=getvolume(it.face,chh);
        T r=rho(it.face)*vol; // volume
        dmin=min(rho(it.face),dmin);
        dmax=max(rho(it.face),dmax);
        if(r<mmin) mmin=r;
        if(r>mmax) mmax=r;
    }
    LOG::printf("mass (%P): max=%.16P min=%.16P sum=%.16P diff=%.16P\n",title,mmax,mmin,sum,sum-total_rho);

    LOG::printf("density (%P): max=%.16P min=%.16P\n",title,dmax,dmin);
    total_rho=sum;
}
//#####################################################################
// Function Print_Velocity_Stats
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Print_Velocity_Stats(const ARRAY<T,FACE_INDEX<TV::m> >& u,std::string title)
{
    if(!print_velocity_stats) return;
    T L2_sum=0;
    TV max_abs;
    T Linf=0;
    for(FACE_ITERATOR<TV> it(grid,0,INTERIOR_REGION);it.Valid();it.Next())
    {
        T value=u(it.face);
        max_abs(it.face.axis)=(abs(value)>max_abs(it.face.axis))?abs(value):max_abs(it.face.axis);
        T e=value-1;
        L2_sum+=sqr(e)*grid.Cell_Size();
        Linf=(abs(e)>Linf)?abs(e):Linf;
    }
    for(int axis=0;axis<TV::m;axis++)
    {
        if(side_bc_type(axis*2)==BC_PERIODIC)
            for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis);it.Valid();it.Next())
            {
                T value=u(it.face);
                max_abs(it.face.axis)=(abs(value)>max_abs(it.face.axis))?abs(value):max_abs(it.face.axis);
                T e=value-1;
                L2_sum+=sqr(e)*grid.Cell_Size();
                Linf=(abs(e)>Linf)?abs(e):Linf;
            }
        else
            for(int side=0;side<2;side++)
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis+side);it.Valid();it.Next())
                {
                    T value=u(it.face);
                    max_abs(it.face.axis)=(abs(value)>max_abs(it.face.axis))?abs(value):max_abs(it.face.axis);
                    T e=value-1;
                    L2_sum+=sqr(e)*grid.Cell_Size()/2;
                    Linf=(abs(e)>Linf)?abs(e):Linf;
                }
    }
        
    LOG::printf("velocities (%P) [L2] [Linf]: %P [%P] [%P]\n",title,max_abs,sqrt(L2_sum),Linf);
    Compute_Divergence(u,title);
}
//#####################################################################
// Function Total_Momentum
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Total_Momentum(const ARRAY<T,FACE_INDEX<TV::m> >& rho_u,std::string title,T time)
{
    if(!print_momentum_stats) return;
    TV momentum;
    for(FACE_ITERATOR<TV> it(grid,0,INTERIOR_REGION);it.Valid();it.Next())
        momentum(it.face.axis)+=rho_u(it.face)*grid.Cell_Size();
    
    for(int axis=0;axis<TV::m;axis++)
    {
        if(side_bc_type(axis*2)==BC_PERIODIC)
            for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis);it.Valid();it.Next())
                momentum(it.face.axis)+=rho_u(it.face)*grid.Cell_Size();
        else
            for(int side=0;side<2;side++)
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis+side);it.Valid();it.Next())
                    momentum(it.face.axis)+=rho_u(it.face)*grid.Cell_Size()/2;
    }
    LOG::cout<<"Total mo("<<title<<",time="<<time<<"): "<<momentum<<" step: "<<momentum-total_momentum
             <<" init: "<<momentum-initial_momentum<<std::endl;
    total_momentum=momentum;
}
//#####################################################################
// Function Total_Momentum
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Total_Momentum(const ARRAY<T,FACE_INDEX<TV::m> >& u,const ARRAY<T,IV>& gamma_ghost,std::string title,T time)
{
    if(!print_momentum_stats) return;
    ARRAY<T,FACE_INDEX<TV::m> > local_density(grid.Cell_Indices(),init_new,1e20);
    Compute_Density(local_density,gamma_ghost);

    TV momentum;
    T ke=0;

    for(FACE_ITERATOR<TV> it(grid,0,INTERIOR_REGION);it.Valid();it.Next())
    {
        momentum(it.face.axis)+=local_density(it.face)*u(it.face)*grid.Cell_Size();
        ke+=local_density(it.face)*sqr(u(it.face))*grid.Cell_Size()/2;
    }
    
    for(int axis=0;axis<TV::m;axis++)
    {
        if(side_bc_type(axis*2)==BC_PERIODIC)
        {
            for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis);it.Valid();it.Next())
            {
                momentum(it.face.axis)+=local_density(it.face)*u(it.face)*grid.Cell_Size();
                ke+=local_density(it.face)*sqr(u(it.face))*grid.Cell_Size()/2;
            }
        }
        else
        {
            for(int side=0;side<2;side++)
            {
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis+side);it.Valid();it.Next())
                {
                    momentum(it.face.axis)+=local_density(it.face)*u(it.face)*grid.Cell_Size()/2;
                    ke+=local_density(it.face)*sqr(u(it.face))*grid.Cell_Size()/4;
                }
            }
        }
    }
    LOG::cout<<"Total mo("<<title<<",time="<<time<<"): "<<momentum<<" step: "<<momentum-total_momentum
             <<" init: "<<momentum-initial_momentum<<std::endl;
    LOG::cout<<"Total ke("<<title<<",time="<<time<<"): "<<ke<<" step: "<<ke-total_ke
             <<" init: "<<ke-initial_ke<<std::endl;
    total_momentum=momentum;
    total_ke=ke;
}
//#####################################################################
// Function Total_Momentum
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Total_Momentum(const ARRAY<T,FACE_INDEX<TV::m> >& u,const ARRAY<T,FACE_INDEX<TV::m> >& density_ghost,std::string title,T time)
{
    if(!print_momentum_stats) return;
    TV momentum;
    T ke=0;

    for(FACE_ITERATOR<TV> it(grid,0,INTERIOR_REGION);it.Valid();it.Next())
    {
        momentum(it.face.axis)+=density_ghost(it.face)*u(it.face)*grid.Cell_Size();
        ke+=density_ghost(it.face)*sqr(u(it.face))*grid.Cell_Size()/2;
    }
    
    for(int axis=0;axis<TV::m;axis++)
    {
        if(side_bc_type(axis*2)==BC_PERIODIC)
        {
            for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis);it.Valid();it.Next())
            {
                momentum(it.face.axis)+=density_ghost(it.face)*u(it.face)*grid.Cell_Size();
                ke+=density_ghost(it.face)*sqr(u(it.face))*grid.Cell_Size()/2;
            }
        }
        else
        {
            for(int side=0;side<2;side++)
            {
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis+side);it.Valid();it.Next())
                {
                    momentum(it.face.axis)+=density_ghost(it.face)*u(it.face)*grid.Cell_Size()/2;
                    ke+=density_ghost(it.face)*sqr(u(it.face))*grid.Cell_Size()/4;
                }
            }
        }
    }
    LOG::cout<<"Total mo("<<title<<",time="<<time<<"): "<<momentum<<" step: "<<momentum-total_momentum
             <<" init: "<<momentum-initial_momentum<<std::endl;
    LOG::cout<<"Total ke("<<title<<",time="<<time<<"): "<<ke<<" step: "<<ke-total_ke
             <<" init: "<<ke-initial_ke<<std::endl;
    total_momentum=momentum;
    total_ke=ke;
}
//#####################################################################
// Function Total_Energy
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Print_Total_Energy(const ARRAY<T,IV>& gamma,const ARRAY<T,FACE_INDEX<TV::m> >& u,T time) const
{
    if(!print_energy_stats) return;
    T free_energy=0;
    T V=grid.Cell_Size();
    auto F=[this](T g){return sqr(g)*sqr(1-g)/sqr(ch->eta);};
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        IV right=it.index.Add_Axis(0,1);
        IV up=it.index.Add_Axis(1,1);
        free_energy+=ch->lambda*(F(gamma(it.index))+0.5*(sqr((gamma(right)-gamma(it.index))*grid.one_over_dX(0))+sqr((gamma(up)-gamma(it.index))*grid.one_over_dX(1))))*V;
    }

    T kinetic_energy=0;
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        kinetic_energy+=0.5*sqr(u(it.face))*V;
    }
    LOG::printf("Energy (Free): %P \t (Kinetic): %P \n",free_energy,kinetic_energy);
    LOG::printf("Energy at time %P (Total): %P \n",time,free_energy+kinetic_energy);
}
//#####################################################################
// Function Test_Analytic_T
//#####################################################################
template<class TV> template<class A> void CHNS_DRIVER<TV>::
Test_Analytic_T(const A& data,const char* name,T t,int g,const char* alias,bool sub_avg)
{
    if(!test_vs_symbolic||!symbolic) return;
    auto [ptr,ins]=error_particles.Insert(name,nullptr);
    if(ins)
    {
        *ptr=new DEBUG_PARTICLES<TV>;
        viewer_output.Add(std::string("error_")+name,**ptr,false,{.draw=false});
    }
    DEBUG_PARTICLES<TV>::Get(*ptr);
    auto f=[this,name](TV X){return symbolic->Get_T(name,X);};
    Test_With_Analytic<TV,T>(grid,data,alias?alias:name,f,t,!suppress_output&&error_output,g,sub_avg);
    DEBUG_PARTICLES<TV>::Get(&debug_particles);
}
//#####################################################################
// Function Test_Analytic_TV
//#####################################################################
template<class TV> template<class A> void CHNS_DRIVER<TV>::
Test_Analytic_TV(const A& data,const char* name,T t,int g,const char* alias)
{
    if(!test_vs_symbolic||!symbolic) return;
    auto [ptr,ins]=error_particles.Insert(name,nullptr);
    if(ins)
    {
        *ptr=new DEBUG_PARTICLES<TV>;
        viewer_output.Add(std::string("error_")+name,**ptr,false,{.draw=false});
    }
    DEBUG_PARTICLES<TV>::Get(*ptr);
    auto f=[this,name](TV X){return symbolic->Get_TV(name,X);};
    Test_With_Analytic<TV,T>(grid,data,alias?alias:name,f,t,!suppress_output&&error_output,g);
    DEBUG_PARTICLES<TV>::Get(&debug_particles);
}
//#####################################################################
// Function Test_Periodic
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Test_Periodic(const ARRAY<T,FACE_INDEX<TV::m> >& u,const char* step)
{
    LOG::cout<<"Testing periodicity at step \""<<step<<"\":"<<std::endl;
    for(int side=0;side<TV::m*2;side+=2)
    {
        for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,side);it.Valid();it.Next())
        {
            FACE_INDEX<TV::m> opp(it.face);
            opp.index(it.face.axis)=grid.Cell_Indices().max_corner(it.face.axis);
            // Add_Ball(grid.Face_X(it.face),grid.dX(it.face.axis)/4,VECTOR<T,3>(1,0,0));
            // Add_Ball(grid.Face_X(opp),grid.dX(it.face.axis)/4,VECTOR<T,3>(1,1,0));
            T abs_diff=abs(u(it.face)-u(opp));
            if(abs_diff>1e-13)
                LOG::printf("\t %P vs %P: %P\n",it.face,opp,abs_diff);
        }
        // Flush_Frame("periodics");
    }        
}
//#####################################################################
// Function Test_Flux
//#####################################################################
template<class TV> void CHNS_DRIVER<TV>::
Test_Flux(const GRID<TV>& flux_grid,const ARRAYS_ND_BASE<T,IV>& base,const ARRAYS_ND_BASE<T,IV>& control,const ARRAY<T,FACE_INDEX<TV::m> >& flux,T scale,const std::string title)
{
    if(!test_flux) return;
    T total_error=0;
    T max_error=0;
    int count=0;
    for(CELL_ITERATOR<TV> it(flux_grid);it.Valid();it.Next())
    {
        T test_value=base(it.index);

        for(int axis=0;axis<TV::m;axis++)
            test_value+=scale*grid.one_over_dX(axis)*(flux(it.Full_Second_Face_Index(axis))-flux(it.Full_First_Face_Index(axis)));
        
        T error=test_value-control(it.index);
        if(abs(error)>max_error) max_error=abs(error);
        total_error+=sqr(error);
        count++;
    }
    LOG::printf("Flux (%P) \t L2: %P \t Linf: %P\n",title,sqrt(total_error/count),max_error);
}
namespace PhysBAM{
// template class CHNS_DRIVER<VECTOR<float,2> >;
// template class CHNS_DRIVER<VECTOR<float,3> >;
template class CHNS_DRIVER<VECTOR<double,2> >;
template class CHNS_DRIVER<VECTOR<double,3> >;
}
