//#####################################################################
// Copyright 2022-23, Jason Goulding
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CHNS_DRIVER__
#define __CHNS_DRIVER__
#include <Core/Advection/ADVECTION_CONSERVATIVE_ENO.h>
#include <Core/Advection/ADVECTION_CONSERVATIVE_WENO.h>
#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Boundaries/BOUNDARY.h>
#include <Core/Boundaries/BOUNDARY_PERIODIC.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Poisson/PROJECTION_UNIFORM.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include "CAHN_HILLIARD.h"
#include "CAHN_HILLIARD_HELMHOLTZ.h"
#include "CAHN_HILLIARD_OPTIMIZER.h"
namespace PhysBAM{

template<class TV> class CHNS_SYMBOLIC;
// template<class TV> class CAHN_HILLIARD;

template<class TV>
class CHNS_DRIVER
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
//    typedef MESH_OBJECT<TV,TV::m-1> T_SURFACE;


public:
    enum BC_TYPE {BC_FREE, BC_SLIP, BC_NOSLIP, BC_PERIODIC};
    
    int current_frame=0;
    T time=0;
    int output_number=0;
    int ghost=5;
    int face_ghost=4;
    T gamma_sum=0;
    T init_gamma_sum=0;

    STREAM_TYPE stream_type;
    DEBUG_PARTICLES<TV>& debug_particles;
    HASHTABLE<std::string,DEBUG_PARTICLES<TV>*> error_particles;

    T distance_scale=1;
    T time_scale=1;
    T mass_scale=1;
    T speed_scale=1;
    T volume_scale=1;
    T density_scale=1;
    T st_scale=1;
    T visc_scale=1;
    T pressure_scale=1;
    T ns_source_scale=1;
    
    int last_frame=100;
    T frame_dt=(T)1/24;
    T speed_limit=0;
    int steps_per_frame=1;
    int restart=0;
    int old_write_frequency=10;
    int write_substeps_level=0;
    bool write_debug_data=false;
    VIEWER_DIR viewer_dir{"output"};
    VIEWER_OUTPUT viewer_output{stream_type,viewer_dir};
    bool N_boundary=true;
    bool debug_divergence=false;
    T rho0=(T)250;
    T rho1=1000;
    T mu0=0;
    T mu1=0;

    int refine=1;
    T total_gamma=0;
    
    T order=1; //time order coefficient?
    bool fix_momentum=false;
    TV total_momentum;
    TV initial_momentum;
    T total_ke=0;
    T initial_ke=0;
    T total_rho=0;

    T solver_tolerance=1e-10;
    T abs_tolerance=1e-30;
    bool use_mu_tol=false;
    T mu_solver_tolerance=1e-10;
    bool use_mu_abs_tol=false;
    T mu_abs_tolerance=1e-30;
    bool use_p_tol=false;
    T p_solver_tolerance=1e-10;
    bool use_p_abs_tol=false;
    T p_abs_tolerance=1e-30;

    CHNS_SYMBOLIC<TV>* symbolic=nullptr;
    CAHN_HILLIARD<TV>* ch=nullptr;
    CAHN_HILLIARD_HELMHOLTZ<TV>* chh=nullptr;
    CAHN_HILLIARD_OPTIMIZER<TV>* cho=nullptr;
    std::string code;

    bool solve_cahn_hilliard=true;
    bool st_force=true;
    bool use_viscosity=true;
    bool fill_old=false;
    bool slip_boundary=false;
    bool noslip_boundary=false;
    bool free_boundary=false;
    VECTOR<bool,TV::m> periodic;
    VECTOR<BC_TYPE,TV::m*2> side_bc_type; // -x, +x, -y, +y, -z, +z
    TV gravity;
    bool use_advection=true;
    bool old_advection=false;
    bool solve_pressure=true;
    bool use_ch_adv_flux=true;
    bool use_ch_solve_flux=true;
    bool ghost_output=false;
    bool show_flux=false;
    bool test_flux=false;
    bool show_st_force=false;
    bool suppress_output=false;
    bool error_output=false;
    bool test_vs_symbolic=false;
    bool symbolic_source=false;
    bool force_u=false;
    bool use_bound_flux=true;
    bool require_convergence=true;
    bool balanced=false;
    bool test_units=false;
    bool display_curvature=false;
    int curvature=0;
    bool use_ideal_curv=false;
    int st_type=0;
    T st_offset=0.2;
    T st_offset2=0.4;
    
    bool second_order_time=false;
    bool first_step=true;
    int g_type=0;

    bool keep_all=false;
    bool reflection_test=false;
    bool invalidate_ghost=false;
    bool div_free=false;
    bool dump_st_data=false;
    bool discrete_advection=false;
    bool use_weno_u=false;
 
    bool print_status=true;
    bool print_parameters=true;
    // bool print_gamma_stats=false;
    bool print_density_stats=false;
    bool print_momentum_stats=false;
    bool print_energy_stats=false;
    bool print_velocity_stats=false;
    bool print_velocity_error=false;
    bool print_solve_results=true;
    bool print_solve_residual=false;

    bool write_viscosity=false;
    bool write_pressure=false;

    ARRAY<T,FACE_INDEX<TV::m> > u_ab_keep;
    ARRAY<T,FACE_INDEX<TV::m> > rho_u_bd_keep;
    ARRAY<T,FACE_INDEX<TV::m> > u_bd_keep;
    ARRAY<T,FACE_INDEX<TV::m> > rho_u_ab_keep;
    ARRAY<T,FACE_INDEX<TV::m> > rho_bd_keep;
    ARRAY<T,FACE_INDEX<TV::m> > rho_ab_keep;
    ARRAY<T,IV> gamma_bd_keep;
    ARRAY<T,IV> gamma_ab_keep;
    ARRAY<T,FACE_INDEX<TV::m> > rho_star_keep;
    ARRAY<T,FACE_INDEX<TV::m> > rho_bar_star_keep;
    ARRAY<T,IV> gamma_star_keep;
    ARRAY<T,FACE_INDEX<TV::m> > rho_u_star_keep;
    ARRAY<T,FACE_INDEX<TV::m> > u_star_keep;
    ARRAY<T,FACE_INDEX<TV::m> > u_star_hat_keep;
    ARRAY<T,FACE_INDEX<TV::m> > u_star_star_keep;
    ARRAY<T,FACE_INDEX<TV::m> > u_star_star_visc_keep;
    ARRAY<T,FACE_INDEX<TV::m> > u_visc_prev_keep;
    ARRAY<T,FACE_INDEX<TV::m> > u_visc_keep;
    ARRAY<T,FACE_INDEX<TV::m> > corrections_keep;
    ARRAY<T,IV> p_diff_keep;
    VECTOR<ARRAY<T,FACE_INDEX<TV::m> >,TV::m> density_flux_keep;
    VECTOR<ARRAY<T,FACE_INDEX<TV::m> >,TV::m> u_avg_keep;
    ARRAY<T,FACE_INDEX<TV::m> > density_cell_flux_keep;
    ARRAY<T,FACE_INDEX<TV::m> > gamma_flux_keep;
    ARRAY<T,FACE_INDEX<TV::m> > gamma_advect_flux_keep;
    ARRAY<T,FACE_INDEX<TV::m> > gamma_solve_flux_keep;
    ARRAY<T,FACE_INDEX<TV::m> > st_force_keep;
    ARRAY<T,FACE_INDEX<TV::m> > grad_gamma_keep;
    
    GRID<TV> grid;
    VECTOR<GRID<TV>,TV::m> face_grids;
    PROJECTION_UNIFORM<TV> projection;
    ARRAY<T,IV> display_gamma;
    ARRAY<T,IV> gamma;
    ARRAY<T,IV> curvature_display;
    ARRAY<T,IV> gamma_old;
    ARRAY<T,IV> pressure;
    ARRAY<T,IV> pressure_old;
    ARRAY<T,IV> mass;
    ARRAY<T,IV> mass0;
    ARRAY<T,IV> mass1;
    ARRAY<T,IV> volume;
    ARRAY<T,FACE_INDEX<TV::m> > face_velocities;
    ARRAY<T,FACE_INDEX<TV::m> > display_velocities;
    ARRAY<T,FACE_INDEX<TV::m> > face_velocities_old;
    ARRAY<T,FACE_INDEX<TV::m> > density;
    ARRAY<T,FACE_INDEX<TV::m> > density_old;
    ARRAY<T,FACE_INDEX<TV::m> > viscosity;
    VECTOR<ARRAY<T,FACE_INDEX<TV::m> >,TV::m> momentum_flux;

    ARRAY<bool,FACE_INDEX<TV::m> > psi_N;
    ARRAY<T,FACE_INDEX<TV::m> > ch_psi_N_values;
    ARRAY<T,FACE_INDEX<TV::m> > p_psi_N_values;
    ARRAY<bool,IV> psi_D;
    ARRAY<T,IV> ch_psi_D_values;
    ARRAY<T,IV> p_psi_D_values;
    AVERAGING_UNIFORM<TV> avg;
    ADVECTION_CONSERVATIVE_WENO<TV,T> advection_scalar;
    BOUNDARY_PERIODIC<TV,T> boundary_scalar_p;
    
    CHNS_DRIVER(const STREAM_TYPE stream_type_input,PARSE_ARGS& parse_args);
    virtual ~CHNS_DRIVER();

    void Initialize_Grid(IV counts,RANGE<TV> domain);
    void Fill_Analytics();
    void Initialize();

    void Set_Boundary_Conditions();
    void Enforce_Boundary_Conditions(ARRAY<T,FACE_INDEX<TV::m> >& u);

    void Execute_Main_Program();
    void Simulate_To_Frame(int frame_input);
    void Step(int substep,T dt);

    void Discrete_Advection(const GRID<TV>& grid,ARRAY<T,FACE_INDEX<TV::m> >& out,const ARRAY<T,FACE_INDEX<TV::m> >& Z_face,const ARRAY<T,FACE_INDEX<TV::m> >& u_face,T dt,T time);
    void Add_Forces(ARRAY<T,FACE_INDEX<TV::m> >& u,ARRAY<T,IV>& gamma_ghost,const ARRAY<T,FACE_INDEX<TV::m> >& rho,T dt);
    void Compute_Density(ARRAY<T,FACE_INDEX<TV::m> >& density,const ARRAYS_ND_BASE<T,IV>& gamma_ghost,int ghost=0);
    void Compute_Density(ARRAY<T,IV>& density,const ARRAYS_ND_BASE<T,IV>& gamma);
    void Compute_Mass(const ARRAY<T,IV>& density_cell,const ARRAYS_ND_BASE<T,IV>& gamma);
    void Compute_Viscosity(ARRAY<T,FACE_INDEX<TV::m> >& viscosity,const ARRAYS_ND_BASE<T,IV>& gamma,int ghost);
    void Add_Viscosity(ARRAY<T,FACE_INDEX<TV::m> >& u_out,const ARRAY<T,FACE_INDEX<TV::m> >& u_in,T dt);
    void Project(ARRAY<T,FACE_INDEX<TV::m> >& u,const ARRAY<T,FACE_INDEX<TV::m> >& rho,T dt,T time);
    void Solve_Viscosity(ARRAY<T,FACE_INDEX<TV::m> >& u,T dt);
    
    void Write_Output_Files(const std::string& title);
    void Update_Display_Gamma(const ARRAY<T,IV>& input);
    void Update_Display_Velocity(const ARRAY<T,FACE_INDEX<TV::m> >& input);
    void Read_Output_Files();

    void Invalidate_Ghost(ARRAY<T,IV>& output,int ghost) const;
    void Invalidate_Ghost(ARRAY<T,FACE_INDEX<TV::m> >& output,int ghost) const;
    void Ghost_Fill(ARRAY<T,IV>& output,const ARRAY<T,IV>& input,int ghost,std::string n) const;
    void Ghost_Fill(ARRAY<T,FACE_INDEX<TV::m> >& output,const ARRAY<T,FACE_INDEX<TV::m> >& input,int ghost,std::string d,std::string n0,std::string n1,std::string n2) const;
    void Fill_Ghost(ARRAY<T,FACE_INDEX<TV::m> >& output,const ARRAY<T,FACE_INDEX<TV::m> >& input,int ghost,std::function<T(const IV&,int,int)> boundary_D=0,std::function<T(const IV&,int,int)> boundary_N=0) const;
    void Fill_Ghost(ARRAY<T,IV>& output,const ARRAY<T,IV>& input,int ghost,std::function<T(const IV&,int)> boundary_N=0) const;
    void Gradient(ARRAY<T,FACE_INDEX<TV::m> >& out,const ARRAYS_ND_BASE<T,IV>& in_ghost,int ghost) const;
    void Gradient(ARRAY<T,FACE_INDEX<TV::m> >& out,const ARRAY<T,IV>& in) const;

    void Compute_Surface_Tension(ARRAY<T,FACE_INDEX<TV::m> >& st_force,ARRAY<T,IV>& gamma_ghost,int type=0,std::function<T(T)> g=0,int curvature=0,bool display_curvature=false) const;
    void Compute_Curvature();
    void Compute_Divergence(const ARRAY<T,FACE_INDEX<TV::m> >& u,std::string title);
    void Print_Density_Stats(const ARRAY<T,FACE_INDEX<TV::m> >& rho,std::string title);
    void Print_Velocity_Stats(const ARRAY<T,FACE_INDEX<TV::m> >& u,std::string title);
    void Total_Momentum(const ARRAY<T,FACE_INDEX<TV::m> >& rho_u,std::string title,T time=0);
    void Total_Momentum(const ARRAY<T,FACE_INDEX<TV::m> >& face_velocities,const ARRAY<T,IV>& gamma_ghost,std::string title,T time=0);
    void Total_Momentum(const ARRAY<T,FACE_INDEX<TV::m> >& face_velocities,const ARRAY<T,FACE_INDEX<TV::m> >& density_ghost,std::string title,T time=0);
    void Print_Total_Energy(const ARRAY<T,IV>& gamma,const ARRAY<T,FACE_INDEX<TV::m> >& u,T time) const;
    template<class A>
    void Test_Analytic_T(const A& ch_data,const char* name,T t,int g=0,const char* alias=0,bool sub_avg=false);
    template<class A>
    void Test_Analytic_TV(const A& ch_data,const char* name,T t,int g=0,const char* alias=0);

    void Test_Periodic(const ARRAY<T,FACE_INDEX<TV::m> >& u,const char* step);

    void Test_Flux(const GRID<TV>& flux_grid,const ARRAYS_ND_BASE<T,IV>& base,const ARRAYS_ND_BASE<T,IV>& control,const ARRAY<T,FACE_INDEX<TV::m> >& flux,T dt,const std::string title);

    template<class A>
    void Keep(A& keep,const A& current)
    {if(keep_all) keep=current;}
    
//#####################################################################
};
}
#endif
