#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Arrays/FLOOD_FILL.h>
#include <Core/Boundaries/PERIODIC_BOUNDARY_HELPER.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/NODE_ITERATOR.h>
#include <Core/Krylov_Solvers/CONJUGATE_GRADIENT.h>
#include <Core/Krylov_Solvers/KRYLOV_SOLVER.h>
#include <Core/Krylov_Solvers/KRYLOV_VECTOR_BASE.h>
#include <Core/Krylov_Solvers/KRYLOV_VECTOR_WRAPPER.h>
#include <Core/Krylov_Solvers/MATRIX_SYSTEM.h>
#include <Core/Log/LOG.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Matrices/SPARSE_MATRIX_FLAT_MXN.h>
#include <Core/Nonlinear_Equations/NEWTONS_METHOD.h>
#include <Core/Poisson/COMPUTE_LAPLACE_MATRIX.h>
#include <Core/Poisson/LAPLACE.h>
#include <Core/Poisson/POISSON_UNIFORM.h>
#include <Core/Read_Write/OCTAVE_OUTPUT.h>
#include "CHN_OBJECTIVE.h"
#include "CHN_SOLVER.h"
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> CHN_SOLVER<TV>::
CHN_SOLVER(const GRID<TV>& grid_input,T e,T r_in,T k_in,T dt_in,bool linear,bool altF)
    :grid(grid_input),eta(e),r(r_in),k(k_in),dt(dt_in),linearized(linear),alt_F(altF)
{
}
//#####################################################################
// Function Solve
//#####################################################################
template<class TV> bool CHN_SOLVER<TV>::
Solve(ARRAYS_ND_BASE<T,IV>& u,ARRAYS_ND_BASE<T,IV>& v,const ARRAYS_ND_BASE<T,IV>& f,RANGE<IV> grid_domain,NEWTON_OPTIONS<T>& options)
{
    BOUNDARY_CONDITIONS<TV> bc;
    bc.psi_N=psi_N;
    bc.psi_N_values=psi_N_values;
    bc.psi_D=psi_D;
    bc.psi_D_values=psi_D_values;
    bc.grid=grid;
    for(int side=0;side<2*TV::m;side++)
    {
        if(periodic_boundary(side/2)) bc.bc(side)=BOUNDARY_CONDITIONS<TV>::BC_PERIODIC;
        else bc.bc(side)=BOUNDARY_CONDITIONS<TV>::BC_NEUMANN;
    }

    LAPLACE_MATRIX<TV> lm;
    
    RANGE<IV> valid_domain(grid_domain);
    RANGE<IV> solve_domain=valid_domain.Intersect(u.domain);
    ARRAY<ARRAY<T>> b_array;
    const ARRAY<T,FACE_INDEX<TV::m> >* beta_array=new ARRAY<T,FACE_INDEX<TV::m> >(grid_domain,init_new,beta);
    Compute_Laplace_Matrix<TV,T>(lm,b_array,f,bc,beta_array);
    delete beta_array;
    
    // Find_A(lm,solve_domain);
    
    bool converged=true;
    for(int color=0;color<lm.number_of_regions;color++)
        if(lm.matrix_index_to_cell_index_array(color).m>0)
        {
            options.enforce_compatibility=lm.enforce_compatibility(color);
            converged=Solve_Subregion(lm,lm.A_array(color),b_array(color),u,v,options,color);
        }

    if(psi_D && psi_D_values)
        for(RANGE_ITERATOR<TV::m> it(solve_domain,0,0);it.Valid();it.Next())
            if((*psi_D)(it.index)) u(it.index)=(*psi_D_values)(it.index);
    
    return converged;
}
//#####################################################################
// Function Solve_Subregion
//#####################################################################
template<class TV> bool CHN_SOLVER<TV>::
Solve_Subregion(LAPLACE_MATRIX<TV>& lm,SPARSE_MATRIX_FLAT_MXN<T>& A,ARRAY<T>& b,ARRAYS_ND_BASE<T,IV>& u,ARRAYS_ND_BASE<T,IV>& v,const NEWTON_OPTIONS<T>& options,const int color)
{
    // Newton Solver
    // Set up Objective Function
    //    use A as the Laplace operator, b as the rhs
    T eta2=eta*eta; 
    if(debug) LOG::printf("k=%P r=%P\n",k/eta2,r);
    CHN_OBJECTIVE<TV> obj(A,b,k/eta2,r,options.enforce_compatibility,alt_F,linearized);
    obj.use_preconditioner=use_preconditioner;
    if(!use_preconditioner) LOG::printf("WARNING: No preconditioning of Objective Function!\n");
    if(diff_test)
    {
        // LOG::printf("b: %P\n",b);
        ARRAY<T> x(b);
        MATRIX_SYSTEM<SPARSE_MATRIX_FLAT_MXN<T>,T,KRYLOV_VECTOR_WRAPPER<T,ARRAY<T>>> test_h=obj.system;
        KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> > test_x(x);
        obj.Test(test_x,test_h,"f1","df1");
        obj.Test_g();
    }

    T tolerance=max(options.relative_tolerance*b.Max_Abs(),options.absolute_tolerance);
    
    NEWTONS_METHOD<T> newtons_method;
    newtons_method.tolerance=tolerance;
    newtons_method.max_iterations=options.max_iterations;
    newtons_method.krylov_tolerance=options.krylov_tolerance;
    newtons_method.max_krylov_iterations=options.restart_iterations;
    // newtons_method.fail_on_krylov_not_converged=false;
    // newtons_method.require_one_iteration=true;
    newtons_method.use_cg=true;
    newtons_method.debug=options.show_results; // ?
    newtons_method.debug_substep_level=4;
    newtons_method.adaptive_tolerance=options.adaptive_tolerance;
    // solver.print_diagnostics=options.show_results;
    // solver.print_residuals=options.show_residual;
    // solver.restart_iterations=options.restart_iterations;
    // solver.ensure_convergence=options.ensure_convergence;
    if(print_system) newtons_method.print_matrix=octave_output.c_str();

    int number_of_unknowns=lm.matrix_index_to_cell_index_array(color).m;

    if(options.show_results) LOG::cout<<"solving "<<number_of_unknowns<<" cells to tolerance "<<tolerance<<std::endl;

    ARRAY<T> x(number_of_unknowns);
    for(int i=0;i<number_of_unknowns;i++) x(i)=u(lm.matrix_index_to_cell_index_array(color)(i));
    ARRAY<KRYLOV_VECTOR_BASE<T>*> vectors;
    KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> > kx;
    std::swap(kx.v,x);

    T objective_eval=0;
    obj.Compute(kx,nullptr,nullptr,&objective_eval);
    newtons_method.objective_error=options.objective_error==-1?
        1e3*std::numeric_limits<T>::epsilon():options.objective_error;
    
    auto [converged,iterations_used]=newtons_method.Newtons_Method(obj,obj.system,kx,vectors);
    std::swap(kx.v,x);

    if(print_system)
    {
        auto name=LOG::sprintf("%s-%i.txt",octave_output.c_str(),print_system_id++);
        OCTAVE_OUTPUT<T> o_out(name,true);
        o_out.Write("x",x);
    }
    
    for(int i=0;i<number_of_unknowns;i++) // move solution to u
    {
        u(lm.matrix_index_to_cell_index_array(color)(i))=obj.dg(x(i));
        v(lm.matrix_index_to_cell_index_array(color)(i))=x(i);        
    }
    
    vectors.Delete_Pointers_And_Clean_Memory();
    return converged;
}
//#####################################################################
// Function Set_Neumann_Boundaries
//#####################################################################
template<class TV> void CHN_SOLVER<TV>::
Set_Neumann_Boundaries(const ARRAY<bool,FACE_INDEX<TV::m> >& N, const ARRAY<T,FACE_INDEX<TV::m> >* N_values)
{
    psi_N=&N;
    psi_N_values=N_values;
}
//#####################################################################
// Function Set_Dirichlet_Boundaries
//#####################################################################
template<class TV> void CHN_SOLVER<TV>::
Set_Dirichlet_Boundaries(const ARRAYS_ND_BASE<bool,IV>& D, const ARRAYS_ND_BASE<T,IV>* D_values)
{
    psi_D=&D;
    psi_D_values=D_values;
}
//#####################################################################
// Function Set_Periodic_Boundaries
//#####################################################################
template<class TV> void CHN_SOLVER<TV>::
Set_Periodic_Boundaries(const VECTOR<bool,TV::m>& p)
{
    periodic_boundary=p;
}
//#####################################################################
// Function Set_Beta
//#####################################################################
template<class TV> void CHN_SOLVER<TV>::
Set_Beta(T b)
{
    beta=b;
}
//#####################################################################
namespace PhysBAM{
//template class CHN_SOLVER<VECTOR<float,1> >;
template class CHN_SOLVER<VECTOR<float,2> >;
template class CHN_SOLVER<VECTOR<float,3> >;
//template class CHN_SOLVER<VECTOR<double,1> >;
template class CHN_SOLVER<VECTOR<double,2> >;
template class CHN_SOLVER<VECTOR<double,3> >;
template struct NEWTON_OPTIONS<float>;
template struct NEWTON_OPTIONS<double>;
}
