//#####################################################################
// Copyright 2024 Jason Goulding
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class CHN_SOLVER
//#####################################################################
#ifndef __CHN_SOLVER__
#define __CHN_SOLVER__

#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Arrays/ARRAYS_ND_BASE.h>
#include <Core/Log/DEBUG_UTILITIES.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Matrices/SPARSE_MATRIX_FLAT_MXN.h>
namespace PhysBAM{

template<class TV> class GRID;
template<class T> class SPARSE_MATRIX_FLAT_MXN;
template<class TV> struct LAPLACE_MATRIX;

// Probably remove this later
template<class T>
struct NEWTON_OPTIONS
{
    int max_iterations=1000;
    int restart_iterations=200;
    bool show_results=false;
    bool show_residual=false;
    bool ensure_convergence=true;
    bool enforce_compatibility=true;
    T absolute_tolerance=1e-14;
    T relative_tolerance=1e-7;
    T objective_error=-1;
    T adaptive_tolerance=1e-3;
    T krylov_tolerance=1e-12;
};

template<class TV>
class CHN_SOLVER
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    
    T beta=1;
    const ARRAY<bool,FACE_INDEX<TV::m> >* psi_N=0;
    const ARRAY<T,FACE_INDEX<TV::m> >* psi_N_values=0; // User provides value in the normal direction
    const ARRAYS_ND_BASE<bool,IV>* psi_D=0; // move psi_D values out of u (double-check)
    const ARRAYS_ND_BASE<T,IV>* psi_D_values=0;
    VECTOR<bool,TV::m> periodic_boundary;
    ARRAY<int,IV> filled_region_colors;
    ARRAY<int> filled_region_touches_dirichlet;
    GRID<TV> grid;
    T eta=0.1;
    T r=0.01;
    T k=1;
    T dt=0.01;
    bool linearized=false;
    bool alt_F=false;

    void Find_A(LAPLACE_MATRIX<TV>& lm,const RANGE<IV>& u_size);
    bool Solve_Subregion(LAPLACE_MATRIX<TV>& lm,SPARSE_MATRIX_FLAT_MXN<T>& A,ARRAY<T>& b,ARRAYS_ND_BASE<T,IV>& u,ARRAYS_ND_BASE<T,IV>& v,const NEWTON_OPTIONS<T>& options,const int color=0);
    
public:
    bool debug=false;
    bool diff_test=false;
    bool print_system=false;
    bool diagonal_scaling=false;
    bool use_preconditioner=true;
    std::string octave_output; 
    int print_system_id=0;
    
    CHN_SOLVER(const GRID<TV>& grid_input,T e,T r_in,T k_in,T dt_in,bool linear,bool altF);
    
    void Set_Beta(T b);
    
    void Set_Neumann_Boundaries(const ARRAY<bool,FACE_INDEX<TV::m> >& N, const ARRAY<T,FACE_INDEX<TV::m> >* N_values=0);
    void Set_Dirichlet_Boundaries(const ARRAYS_ND_BASE<bool,IV>& D, const ARRAYS_ND_BASE<T,IV>* D_values);
    void Set_Periodic_Boundaries(const VECTOR<bool,TV::m>& p);
    
    bool Solve(ARRAYS_ND_BASE<T,IV>& u,ARRAYS_ND_BASE<T,IV>& v,const ARRAYS_ND_BASE<T,IV>& f,RANGE<IV> grid_domain,NEWTON_OPTIONS<T>& options=NEWTON_OPTIONS<T>());
    
//#####################################################################
};
}
#endif
