//#####################################################################
// Copyright 2022
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Log/LOG.h>
#include <Core/Matrices/MATRIX.h>
#include <Core/Ordinary_Differential_Equations/RUNGEKUTTA.h>
#include <Core/Symbolics/EXECUTABLE_PROGRAM.h>
#include <Core/Symbolics/PROGRAM_COMPILER.h>
#include <Core/Symbolics/PROGRAM_SCALAR.h>
#include "CHO_SYMBOLIC.h"

namespace PhysBAM{

//#####################################################################
// Constructor
//#####################################################################
template<class TV> CHO_SYMBOLIC<TV>::
CHO_SYMBOLIC(const std::string& analytic_solution,T r_in,T b_in,T F_scale,bool alt_F,bool use_b,bool linearized):r(r_in),b(b_in),Fscale(F_scale)
{
    typedef PROGRAM_SCALAR<T> PS;
    PROGRAM_COMPILER<T> pc;
    // pc.debug=true; // 2D variable names
    for(int i=0;i<TV::m;i++)
        idx_X(i)=pc.Add_Input(LOG::sprintf("%c","xyz"[i]).c_str()); //i0-i1?
    idx_t=pc.Add_Input("t"); // i2
    idx_dt=pc.Add_Input("dt"); // i3
    idx_lambda=pc.Add_Input("lambda"); // i4
    idx_eta=pc.Add_Input("eta"); // i5
    idx_M=pc.Add_Input("M"); // i6
    idx_second_order=pc.Add_Input("second_order"); // i13
    idx_t2=pc.Add_Input("t2"); // i14
    idx_t0=pc.Add_Input("t0"); // i15
    idx_sigma=pc.Add_Input("sigma"); // i17
    idx_m=pc.Add_Input("m"); // i18
    idx_s=pc.Add_Input("s"); // i19
    idx_kg=pc.Add_Input("kg"); // i20
    idx_speed_scale=pc.Add_Input("speed_scale"); // i22
    idx_r=pc.Add_Input("r");
    idx_b=pc.Add_Input("b");
    idx_Fscale=pc.Add_Input("Fscale");
    input_size=pc.Num_In();

    for(int i=0;i<TV::m;i++) pc.Add_Output(LOG::sprintf("%c","uvw"[i]).c_str());
    pc.Add_Output("p");
    pc.Add_Output("gamma");
    auto num_out=pc.Num_Out();
   
    pc.Parse(analytic_solution.c_str());
    auto pc2(pc);
    auto pc0(pc);
    pc2.Rename_Variable(idx_t,idx_t2); // time t+dt
    pc0.Rename_Variable(idx_t,idx_t0); // time t-dt
    pc.Append(pc2);
    pc.Append(pc0);
    pc.Optimize();
    
    PS ps_speed_scale={&pc,idx_speed_scale};
    PS ps_m={&pc,idx_m};
    PS ps_s={&pc,idx_s};
    PS ps_kg={&pc,idx_kg};

    VECTOR<PS,TV::m> ps_u;
    VECTOR<PS,TV::m> ps_u2;
    VECTOR<PS,TV::m> ps_u0;
    for(int i=0;i<TV::m;i++)
    {
        ps_u(i)=Make_Output(PS{&pc,MEM_OUT(i)}*ps_speed_scale);
        ps_u2(i)=Make_Output(PS{&pc,MEM_OUT(i+Value(num_out))}*ps_speed_scale);
        ps_u0(i)=Make_Output(PS{&pc,MEM_OUT(i+2*Value(num_out))}*ps_speed_scale);
    }
    Add("u",ps_u);
    Add("u2",ps_u2);
    Add("u0",ps_u0);
 
    PS ps_gamma={&pc,MEM_OUT(TV::m+1)};
    PS ps_gamma2={&pc,MEM_OUT(TV::m+1+Value(num_out))};
    PS ps_gamma0={&pc,MEM_OUT(TV::m+1+2*Value(num_out))};

    Add("gamma",ps_gamma);
    Add("gamma2",ps_gamma2);
    Add("gamma0",ps_gamma0);
    
    PS ps_dt={&pc,idx_dt};
    PS ps_lambda={&pc,idx_lambda};
    PS ps_eta={&pc,idx_eta};
    PS ps_M={&pc,idx_M};
    PS ps_second_order={&pc,idx_second_order};
    PS ps_t2={&pc,idx_t2};
    PS ps_t0={&pc,idx_t0};
    PS ps_sigma={&pc,idx_sigma};
    PS ps_r={&pc,idx_r};
    PS ps_b={&pc,idx_b};
    PS ps_Fscale={&pc,idx_Fscale};

    (void)ps_dt;
    (void)ps_lambda;
    (void)ps_eta;
    (void)ps_M;
    (void)ps_second_order;
    (void)ps_t2;
    (void)ps_t0;
    (void)ps_sigma;
    (void)ps_m;
    (void)ps_s;
    (void)ps_kg;
    (void)ps_r;
    (void)ps_b;
    (void)ps_Fscale;
    
    VECTOR<MEM_IN,TV::m> axii;
    for(int i=0;i<TV::m;i++) axii(i)=MEM_IN(i);

#define MAKE(name,ps) auto ps_##name=Make_Output(ps);Add(#name,ps_##name);
    MAKE(6,PS::Number(pc,(T)6));
    MAKE(4,PS::Number(pc,(T)4));
    MAKE(3,PS::Number(pc,(T)3));
    MAKE(2,PS::Number(pc,(T)2));
    MAKE(1,PS::Number(pc,(T)1));
    MAKE(0,PS::Number(pc,(T)0));
    MAKE(epsilon,PS::Number(pc,(T)1e-8));

    // cond(a,b,c)
    // a ? b : c
    // MAKE(test_cond,cond(ps_2<ps_1,ps_0,ps_1));
    auto dF=[&](const auto& g)
    {
        if(linearized) return ps_Fscale*g/(ps_eta*ps_eta);
        else if(alt_F)
        {
            return ps_Fscale*cond(g<ps_r,
                ps_6*(g-ps_r)*(ps_2*ps_r*ps_r-ps_4*ps_r*g+g)/g,
                cond(g<(ps_1/ps_2),
                    -ps_2*(ps_r-g)*(ps_3-ps_2*ps_r-ps_4*g)-ps_4*sqr(ps_r-g),
                    cond(g<(ps_1-ps_r),
                        -(-ps_2*(ps_r-(ps_1-g))*(ps_3-ps_2*ps_r-ps_4*(ps_1-g))-ps_4*sqr(ps_r-(ps_1-g))),
                        -(ps_6*((ps_1-g)-ps_r)*(ps_2*ps_r*ps_r-ps_4*ps_r*(ps_1-g)+(ps_1-g))/(ps_1-g))
                    )
                )
            )/(ps_eta*ps_eta);
        }
        else return ps_Fscale*cond(g<ps_r,
            ps_6*ps_r*(ps_1-ps_2*ps_r)*(ps_1-ps_r/g),
            cond(g<(ps_1/ps_2),
                -ps_2*(ps_r-g)*(ps_3-ps_2*ps_r-ps_4*g)-ps_4*sqr(ps_r-g),
                cond(g<=(ps_1-ps_r),
                    -(-ps_2*(ps_r-(ps_1-g))*(ps_3-ps_2*ps_r-ps_4*(ps_1-g))-ps_4*sqr(ps_r-(ps_1-g))),
                    -ps_6*ps_r*(ps_1-ps_2*ps_r)*(ps_1-ps_r/(ps_1-g))
                )
            )
        )/(ps_eta*ps_eta);
    };
    MAKE(order,ps_1+ps_second_order/ps_2);
    MAKE(dth,ps_dt/ps_order);
    ps_b=use_b?ps_b:sqrt(ps_M*ps_lambda/ps_dth);
    MAKE(a,ps_b*ps_dth);

    MAKE(df_gamma0,dF(ps_gamma0));
    MAKE(df_gamma,dF(ps_gamma));
    MAKE(df_gamma2,dF(ps_gamma2));
    MAKE(gamma_ab,ps_second_order*(ps_gamma-ps_gamma0)+ps_gamma);
    MAKE(gamma_bd,(ps_second_order*(ps_gamma-ps_gamma0/ps_2)+ps_gamma)/ps_order);
    MAKE(df_ab_gamma,ps_second_order*(ps_df_gamma-ps_df_gamma0)+ps_df_gamma);

    // source term for gamma
    MAKE(src_difft,Diff(ps_gamma2,idx_t2)/ps_s);
    MAKE(src_inert,ps_src_difft+Divergence(ps_u2*ps_gamma2,axii)/ps_m);
    MAKE(src_xi,ps_lambda*(ps_df_gamma2-Laplacian(ps_gamma2,axii)/(ps_m*ps_m)));
    MAKE(src_ch,ps_src_inert-ps_M*Laplacian(ps_src_xi,axii)/(ps_m*ps_m));

    // scheme   
    MAKE(lap_df_ab,Laplacian(ps_df_ab_gamma,axii));
    MAKE(u_ab,ps_second_order*(ps_u-ps_u0)+ps_u);
    MAKE(S,sqrt(ps_M*ps_lambda/ps_dth));
    MAKE(grad_gamma,Gradient(ps_gamma,axii));
    MAKE(grad_gamma_ab,Gradient(ps_gamma_ab,axii));
    MAKE(grad_gamma_bd,Gradient(ps_gamma_bd,axii));
    MAKE(grad_gamma2,Gradient(ps_gamma2,axii));
    MAKE(grad_gamma0,Gradient(ps_gamma0,axii));
    MAKE(grad_df_gamma2,Gradient(ps_df_gamma2,axii));
    MAKE(grad_df_ab_gamma,Gradient(ps_df_ab_gamma,axii));
    MAKE(gamma_star,ps_gamma_bd-ps_dth*Divergence(ps_u_ab*ps_gamma_ab,axii));
    MAKE(grad_gamma_star,Gradient(ps_gamma_star,axii));
    MAKE(gamma_C,ps_gamma2+ps_a*ps_df_gamma2);
    MAKE(gamma_B,ps_gamma_C-ps_S*ps_dth*Laplacian(ps_gamma_C,axii)+ps_a*ps_df_gamma2);
    MAKE(gamma_A,ps_gamma_B-(ps_2*ps_gamma_ab-ps_a*ps_S*ps_dth*ps_lap_df_ab));
    MAKE(gamma_star_alt,ps_gamma_A-ps_S*ps_dth*Laplacian(ps_gamma_A,axii)+ps_2*ps_gamma_ab-ps_dth*ps_M*ps_lambda*ps_lap_df_ab-ps_dth*ps_src_ch);
    MAKE(gamma_D,ps_gamma_star-(ps_2*ps_gamma_ab-ps_dth*ps_M*ps_lambda*ps_lap_df_ab)+ps_dth*ps_src_ch);
    MAKE(gamma_bd_alt,ps_gamma_star_alt+ps_dth*Divergence(ps_u_ab*ps_gamma_ab,axii));

    // Gradients for Solves (neumann boundaries)
    MAKE(grad_gamma_C,Gradient(ps_gamma_C,axii));
    MAKE(grad_gamma_A,Gradient(ps_gamma_A,axii));    
    MAKE(bc_xi,ps_lambda*(ps_df_gamma2-Laplacian(ps_gamma2,axii)+ps_2*ps_S*(ps_gamma2-ps_gamma_ab)/(ps_M*ps_lambda)-ps_a*Laplacian(ps_df_gamma2-ps_df_ab_gamma,axii)-(ps_M*ps_lambda-ps_a*ps_S)/(ps_M*ps_lambda)*(ps_df_gamma2-ps_df_ab_gamma)));
    MAKE(grad_bc_xi,Gradient(ps_bc_xi,axii));
    
    pc.Optimize();
#undef MAKE
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> CHO_SYMBOLIC<TV>::
~CHO_SYMBOLIC()
{
}
//#####################################################################
// Function Add
//#####################################################################
template<class TV> void CHO_SYMBOLIC<TV>::
Add(const std::string& name,const PROGRAM_SCALAR<T>& ps)
{
    t_map.Set(name,new EXECUTABLE_PROGRAM<T>(ps));
}
//#####################################################################
// Function Add
//#####################################################################
template<class TV> template<int d> void CHO_SYMBOLIC<TV>::
Add(const std::string& name,const VECTOR<PROGRAM_SCALAR<T>,d>& ps)
{
    tv_map.Set(name,new EXECUTABLE_PROGRAM<TV>(ps));
}
//#####################################################################
// Function Add
//#####################################################################
// template<class TV> template<int m,int n> void CH_SYMBOLIC<TV>::
// Add(const std::string& name,const MATRIX<PROGRAM_SCALAR<T>,m,n>& ps)
// {
//     mat_map.Set(name,new EXECUTABLE_PROGRAM<MATRIX<T,TV::m,TV::m> >(ps));
// }
//#####################################################################
// Function Set_Step
//#####################################################################
template<class TV> void CHO_SYMBOLIC<TV>::
Set_Step(T time, T dt)
{
    inputs.Resize(input_size,init_all);
    inputs.Subset(idx_X)=TV();
    inputs(idx_t)=time/time_scale;
    inputs(idx_dt)=dt/time_scale;
    inputs(idx_lambda)=lambda;
    inputs(idx_eta)=eta;
    inputs(idx_M)=M;
    inputs(idx_second_order)=second_order;
    inputs(idx_t2)=(time+dt)/time_scale;
    inputs(idx_t0)=(time-dt)/time_scale;
    inputs(idx_sigma)=sigma;
    inputs(idx_m)=distance_scale;
    inputs(idx_s)=time_scale;
    inputs(idx_kg)=mass_scale;
    inputs(idx_speed_scale)=speed_scale;
    inputs(idx_r)=r;
    inputs(idx_b)=b;
    inputs(idx_Fscale)=Fscale;
}
template class CHO_SYMBOLIC<VECTOR<float,2> >;
template class CHO_SYMBOLIC<VECTOR<double,2> >;
template class CHO_SYMBOLIC<VECTOR<float,3> >;
template class CHO_SYMBOLIC<VECTOR<double,3> >;
}
