//
// Created by mhrnaz on 5/1/23.
//
//#####################################################################
// Function main
//#####################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Grids/GRID.h>
#include <Core/Grids_Uniform_Computations/MARCHING_CUBES.h>
#include <Core/Images/TEX_FILE.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Meshes/MESH_OBJECT.h>
#include <Core/Meshes/SEGMENTED_CURVE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;

ARRAY<std::string> dir;
ARRAY<int> res;
int casenum=-1;
ARRAY<std::string> types={"circularity","COM","velocity"};
ARRAY<std::string> symbols={"psi","y","v"};
const int res_num=4;
ARRAY<T> data;
ARRAY<T> times;
ARRAY<T> ref;
ARRAY<T> errors;
std::string fig_dir;
std::string ref_dir="";
std::string tex_file="";
void search_replace(std::string file_name,std::string old_word,std::string new_word)
{
    std::ifstream fin(file_name);
    // Check if the file was opened successfully
    if(!fin)
    {
        std::cerr<<"Error opening file"<<std::endl;
        return;
    }
    // Read the contents of the file into a string
    std::string text(
            (std::istreambuf_iterator<char>(fin)),
            std::istreambuf_iterator<char>());
    fin.close();
    // Perform the search and replace operation on the string
    std::string search=old_word;
    std::string replace=new_word;
    size_t pos=0;
    while((pos=text.find(search, pos))!=std::string::npos)
    {
        text.replace(pos,search.length(),replace);
        pos+=replace.length();
    }
    // Open the file for writing
    std::ofstream fout(file_name,std::ios::trunc);
    // Check if the file was opened successfully
    if(!fout)
    {
        std::cerr<<"Error opening file"<<std::endl;
        return;
    }
    // Write the modified string back to the file
    fout<<text;
    fout.close();
}
void compute_error()
{
    std::ifstream in;
    for(int j=0;j<types.Size();j++)
    {
        std::string error_filepath=fig_dir+"/txt/"+types(j)+"_error.txt";
        std::ofstream error_file(error_filepath);
        for(int i=0;i<dir.m;i++)
        {
            data.Resize(0);
            times.Resize(0);
            ref.Resize(0);
            //Open the file with the data
            std::string data_filepath=fig_dir+"/txt/"+types(j)+"_"+dir(i)+".txt";
            LOG::cout<<"opening "<<data_filepath<<std::endl;
            in.open(data_filepath);
            T t,value;
            std::string header;
            if(in.is_open())
            {
                getline(in,header); // header line not used
                while(in>>t>>value) // get time and value
                {
                    data.Append(value);
                    times.Append(t);
                }
                in.close();
            }
            
            //Looks for a file with this format to use as a reference (type_ref_casenum-e.g velocity_ref_6.txt)
            std::string ref_filepath=fig_dir+"/txt/"+types(j)+"_reference_"+ref_dir+".txt";
            LOG::cout<<"opening reference file "<<ref_filepath<<std::endl;
            in.open(ref_filepath);
            if (in.is_open())
            {
                LOG::cout<<"\treference file is open"<<std::endl;
                getline(in,header);
                LOG::cout<<"\tresolution = "<<res(i)<<std::endl;
                while(in>>t>>value)
                {
                    ref.Append(value);
                }
                in.close();
            }
            LOG::cout<<"ref size = "<<ref.Size()<<std::endl;
            LOG::cout<<"data size = "<<data.Size()<<std::endl;
            PHYSBAM_ASSERT(ref.Size()==data.Size());
            T error=0;
            for (int i=0;i<ref.Size();++i)
            {
                T diff=ref(i)-data(i);
                error+=diff*diff;
            }
            error=std::sqrt(error/ref.Size());
            LOG::cout<<"error="<<error<<std::endl;
            LOG::cout<<"writing error to "<<error_filepath<<std::endl;
            error_file<<res(i)<<" "<<error<<std::endl;
            std::string placeholder="L2-"+std::to_string(res(i))+"-"+std::to_string(casenum)+"-"+symbols(j);
            LOG::cout<<"searching for \""<<placeholder<<"\" and replacing with error values..."<<std::endl;
            search_replace(fig_dir+"/tex/"+tex_file,placeholder,LOG::sprintf("%.2e",error));
        }
    }
}
void compute_order()
{
    for(int i=0;i<types.Size();++i)
    {
        errors.Resize(0);
        ARRAY<int> resolutions;
        std::ifstream in;
        std::string error_filepath=fig_dir+"/txt/"+types(i)+"_error.txt";
        LOG::cout<<"computing order in file: "<<error_filepath<<std::endl;
        in.open(error_filepath);
        int resolution;
        T error;
        if(in.is_open())
        {
            while(in>>resolution>>error)
            {
                // LOG::cout<<resolution<<" "<<error<<std::endl;
                resolutions.Append(resolution);
                errors.Append(error);
            }
            in.close();
        }
        // LOG::cout<<errors<<std::endl;
        for(int j=1;j<errors.Size();++j)
        {
            T order=std::log2(errors(j-1)/errors(j));
            std::string placeholder="L2-ORDER-"+std::to_string(resolutions(j))+"-"+std::to_string(casenum)+"-"+symbols(i);
            LOG::cout<<"searching for \""<<placeholder<<"\" and replacing with order value \""<<order<<"\""<<std::endl;
            search_replace(fig_dir+"/tex/"+tex_file,placeholder,LOG::sprintf("%.2f",order));
        }
    }
}
int main(int argc,char *argv[])
{
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;

    parse_args.Add("-tex",&tex_file,"string-file name","tex file name");
    parse_args.Add("-i",&dir,"string-directory name","directory name to read data from");
    parse_args.Add("-ref",&ref_dir,"string-directory name","directory name to read reference data from");
    parse_args.Add("-res",&res,"scalar-array","resolution");
    parse_args.Add("-case",&casenum,"scalar","case");
    parse_args.Add("-fig",&fig_dir,"dir","base figure directory");

    parse_args.Parse();
    PHYSBAM_ASSERT(casenum>0);
    compute_error();
    compute_order();
    
    return 0;
}
