#include <Core/Basic_Geometry/SPHERE.h>
#include <Core/Log/TEST_UNITS.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include "CHNS_DRIVER.h"

using namespace PhysBAM;

template<class TV> void Build_Domain(CHNS_DRIVER<TV>& driver,int test,PARSE_ARGS& parse_args);
template<class TV> void Fill_Arrays(CHNS_DRIVER<TV>& driver,int test,PARSE_ARGS& parse_args);

template<class TV> void Execute_Main_Program(STREAM_TYPE& stream_type,PARSE_ARGS& parse_args)
{ 
    typedef VECTOR<int,TV::m> IV;
    typedef typename TV::SCALAR T;

    CHNS_DRIVER<TV> driver(stream_type,parse_args);
    
    int test=0;
    parse_args.Add("-test",&test,"int","test number");
    parse_args.Parse(true);
    Build_Domain(driver,test,parse_args);
    Fill_Arrays(driver,test,parse_args);
    parse_args.Parse(); // Final parse to check for bad arguments
    // DRIVER MAIN LOOP
    driver.Execute_Main_Program();
}

int main(int argc,char *argv[])
{
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);
    typedef double T;
    typedef double RW;
    STREAM_TYPE stream_type=STREAM_TYPE::type_double;
    PARSE_ARGS parse_args(argc,argv);
    bool opt_3d=false;
    LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;
    LOG::cout<<std::setprecision(16);
    parse_args.Add("-3d",&opt_3d,"run in 3 dimensions");
    parse_args.Parse(true);
    if(opt_3d) Execute_Main_Program<VECTOR<T,3> >(stream_type,parse_args);
    else Execute_Main_Program<VECTOR<T,2> >(stream_type,parse_args);
    LOG::Finish_Logging();
    return 0;
}

template<class TV> void Build_Domain(CHNS_DRIVER<TV>& driver,int test,PARSE_ARGS& parse_args)
{
    typedef VECTOR<int,TV::m> IV;
    typedef typename TV::SCALAR T;
    RANGE<TV> grid_range;
    IV cell_counts;

    TV range_scale;
    TV range_min_corner;
    bool use_min_corner=false;
    TV range_max_corner;
    bool use_max_corner=false;
    int res=32;
    IV res_v;
    bool use_res_v=false;

    parse_args.Add("-res",&res,"resolution","grid resolution");
    parse_args.Add("-res_v",&res_v,&use_res_v,"integer vector","grid resolution vector");
    parse_args.Add("-range_scale",&range_scale,"vector","scale range of simulation");
    parse_args.Add("-range_min",&range_min_corner,&use_min_corner,"vector","minimum corner of the grid domain");
    parse_args.Add("-range_max",&range_max_corner,&use_max_corner,"vector","maximum corner of the grid domain");
    parse_args.Parse(true);
    
    T distance_scale=driver.distance_scale;
    
    switch(test)
    {
        case 1: // test source
            {
                if(use_max_corner) grid_range.max_corner=range_max_corner*distance_scale;
                else grid_range.max_corner=TV()+pi*distance_scale;
                if(use_min_corner) grid_range.min_corner=range_min_corner*distance_scale;
                else grid_range.min_corner=TV()-pi*distance_scale;

                if(use_res_v) cell_counts=res_v;
                else cell_counts=IV()+res;
                break;
            }
        case 16: // Air Bubble Test
            {
                if(use_max_corner) grid_range.max_corner=range_max_corner*distance_scale;
                else
                {
                    grid_range.max_corner(0)=1*distance_scale;
                    grid_range.max_corner(1)=2*distance_scale;
                }
                if(use_min_corner) grid_range.min_corner=range_min_corner*distance_scale;
                else grid_range.min_corner=TV();

                if(use_res_v) cell_counts=res_v;
                else
                {
                    cell_counts(0)=res;
                    cell_counts(1)=2*res;
                }            
                break;
            }
        case 17: // Rayleigh-Taylor Test
            {
                if(use_max_corner) grid_range.max_corner=range_max_corner*distance_scale;
                else
                {
                    grid_range.max_corner(0)=1*distance_scale;
                    grid_range.max_corner(1)=4*distance_scale;
                }
                if(use_min_corner) grid_range.min_corner=range_min_corner*distance_scale;
                else grid_range.min_corner=TV();

                if(use_res_v) cell_counts=res_v;
                else
                {
                    cell_counts(0)=res;
                    cell_counts(1)=4*res;
                }            
                break;
            }
        case 18: // Dam Break Test
            {
                T a=0.05715;
                if(use_max_corner) grid_range.max_corner=range_max_corner*distance_scale;
                else
                {
                    grid_range.max_corner(0)=4*a*distance_scale;
                    grid_range.max_corner(1)=2*a*distance_scale;
                }
                if(use_min_corner) grid_range.min_corner=range_min_corner*distance_scale;
                else grid_range.min_corner=TV();

                if(use_res_v) cell_counts=res_v;
                else
                {
                    cell_counts(0)=2*res;
                    cell_counts(1)=res;
                }            
                break;
            }
        case 31: 
            {
                if(use_max_corner) grid_range.max_corner=range_max_corner*distance_scale;
                else
                {
                    grid_range.max_corner(0)=0.5;
                    grid_range.max_corner(1)=1;
                }
                if(use_min_corner) grid_range.min_corner=range_min_corner*distance_scale;
                else grid_range.min_corner=TV();

                if(use_res_v) cell_counts=res_v;
                else
                {
                    cell_counts(0)=res/2;
                    cell_counts(1)=res;
                }
                break;
            }
        default:
            {
                RANGE<TV> default_range(-TV::All_Ones_Vector()*distance_scale*range_scale(0),TV::All_Ones_Vector()*distance_scale*range_scale(1));
                if(use_max_corner) default_range.max_corner=range_max_corner*distance_scale;
                if(use_min_corner) default_range.min_corner=range_min_corner*distance_scale;
                grid_range=default_range;

                cell_counts=IV()+res;
                if(use_res_v) cell_counts=res_v;
            }
    }
    
    if(driver.print_parameters) LOG::cout<<"Grid range: "<<grid_range<<std::endl;
    driver.Initialize_Grid(cell_counts,grid_range);    
}
template<class TV> void Fill_Arrays(CHNS_DRIVER<TV>& driver,int test,PARSE_ARGS& parse_args)
{
    typedef VECTOR<int,TV::m> IV;
    typedef typename TV::SCALAR T;

    int seed=1235;
    bool invert_gamma=false;
    T default_gamma=0;
    bool random_gamma=false;
    T random_scale=1;
    bool use_periodic=false;
    bool custom_contact=false;
    T gamma_min=0;
    parse_args.Add("-seed",&seed,"int","rng seed");
    parse_args.Add("-gamma",&default_gamma,"default_gamma","The default concentration for grid cells");
    parse_args.Add("-random_gamma",&random_scale,&random_gamma,"float","The default concentration for grid cells will be random");
    parse_args.Add("-invert_gamma",&invert_gamma,"invert the gamma distribution");
    parse_args.Add("-fill_old",&driver.fill_old,"use the provided analytic solution to fill information at time -dt");
    parse_args.Add("-periodic",&use_periodic,"The domain should be periodic in all directions");
    parse_args.Add("-periodic_v",&driver.periodic,"bool vector","Define the axii where the domain should be periodic");
    parse_args.Add("-contact",&driver.ch->side_contact_angle,&custom_contact,"side vector","Define the contact angle for each side of the domain");
    parse_args.Add("-slip",&driver.slip_boundary,"Symbolic equation is consistent with a Free-Slip Boundary");
    parse_args.Add("-free",&driver.free_boundary,"Symbolic equation is consistent with a Free Boundary");
    parse_args.Add("-gamma_min",&gamma_min,"scalar","Offset gamma from 0 and 1 during setup.");
    parse_args.Parse(true);

    if(use_periodic) driver.periodic.Fill(true);
    driver.side_bc_type.Fill(CHNS_DRIVER<TV>::BC_SLIP);

    T max_gamma=1;
    T min_gamma=0;
    if(gamma_min==0&&driver.cho!=nullptr) //ensure "0" maps to "r" and "1" to "1-r"
    {
        max_gamma-=driver.cho->r;
        min_gamma+=driver.cho->r;
    }
    else
    {
        max_gamma-=gamma_min;
        min_gamma+=gamma_min;
    }
    
    GRID<TV>& grid=driver.grid;
    ARRAY<T,IV>& gamma=driver.gamma;
    ARRAY<T,IV>& pressure=driver.pressure;
    ARRAY<T,FACE_INDEX<TV::m>>& face_velocities=driver.face_velocities;
    
    RANDOM_NUMBERS<T> random(seed);
    switch(test)
    {
        case 0: // Gamma in whole field (random or default)
            {
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T g=random_gamma ? (T).5+random_scale*(random.Get_Number()-(T)0.5) : default_gamma;
                    gamma(it.Cell_Index())=g;
                }

                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 1: // test source
            {
                if(driver.symbolic)
                {
                    driver.Fill_Analytics();
                }
                break;
            }
        case 2: // sin wave gamma distribution
            {
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T x=grid.Cell_X(it.index)(0);
                    gamma(it.Cell_Index())=(sin(2*pi*x)+1)/2;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 3: // Gamma concentrated in top half
            {
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T y=grid.Cell_X(it.index)(1);
                    if(y>=(T)0.5*grid.domain.Edge_Lengths()(1)) gamma(it.Cell_Index())=1;
                    else gamma(it.Cell_Index())=0;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 4: // Gamma in center box
            {
                RANGE<TV> box((T)0.5*grid.domain);
                box+=(TV()+(T)0.25);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(box.Lazy_Inside(grid.Cell_X(it.index))) gamma(it.index)=1;
                    else gamma(it.index)=0;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 5: // Checkerboard
            {
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T g=random_gamma ? random.Get_Number() : default_gamma;
                    if((it.index(0)/2+it.index(1)/2)%2) gamma(it.Cell_Index())=g+0.01;
                    else gamma(it.Cell_Index())=g-0.01;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 6: // Lopsized Spheres (2)
            {
                TV b1c,b2c;
                b1c(0)=(T)0.33;
                b1c(1)=(T)0.5;
                b2c(0)=(T)0.75;
                b2c(1)=(T)0.5;
                SPHERE<TV> ball1(b1c,(T)0.2);
                SPHERE<TV> ball2(b2c,(T)0.1);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(ball1.Lazy_Inside(grid.Cell_X(it.index)) || ball2.Lazy_Inside(grid.Cell_X(it.index)))
                        gamma(it.index)=1;
                    else gamma(it.index)=0;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 7: // Equal Spheres (4)
            {
                TV b1c,b2c,b3c,b4c;
                b1c(0)=(T)0.25;
                b1c(1)=(T)0.25;
                b2c(0)=(T)0.75;
                b2c(1)=(T)0.75;
                b3c(0)=(T)0.75;
                b3c(1)=(T)0.25;
                b4c(0)=(T)0.25;
                b4c(1)=(T)0.75;
                SPHERE<TV> ball1(b1c,(T)0.15);
                SPHERE<TV> ball2(b2c,(T)0.15);
                SPHERE<TV> ball3(b3c,(T)0.15);
                SPHERE<TV> ball4(b4c,(T)0.15);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(ball1.Lazy_Inside(grid.Cell_X(it.index)) || ball2.Lazy_Inside(grid.Cell_X(it.index)) || ball3.Lazy_Inside(grid.Cell_X(it.index)) || ball4.Lazy_Inside(grid.Cell_X(it.index)))
                        gamma(it.index)=1;
                    else gamma(it.index)=0;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 8: // UnEqual Spheres (4)
            {
                TV b1c,b2c,b3c,b4c;
                b1c(0)=(T)0.25;
                b1c(1)=(T)0.25;
                b2c(0)=(T)0.75;
                b2c(1)=(T)0.75;
                b3c(0)=(T)0.75;
                b3c(1)=(T)0.25;
                b4c(0)=(T)0.30;
                b4c(1)=(T)0.75;
                SPHERE<TV> ball1(b1c,(T)0.1);
                SPHERE<TV> ball2(b2c,(T)0.1);
                SPHERE<TV> ball3(b3c,(T)0.1);
                SPHERE<TV> ball4(b4c,(T)0.1);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(ball1.Lazy_Inside(grid.Cell_X(it.index)) || ball2.Lazy_Inside(grid.Cell_X(it.index)) || ball3.Lazy_Inside(grid.Cell_X(it.index)) || ball4.Lazy_Inside(grid.Cell_X(it.index)))
                        gamma(it.index)=1;
                    else gamma(it.index)=0;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 9: // Half-Gamma in center box
            {
                RANGE<TV> box((T)0.5*RANGE<TV>::Unit_Box());
                box+=(TV()+(T)0.25);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(box.Lazy_Inside(grid.Cell_X(it.index))) gamma(it.index)=(T)0.75;
                    else gamma(it.index)=(T)0.25;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 10: // normal gamma distribution
            {
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T x=grid.Cell_X(it.index)(0);
                    T y=grid.Cell_X(it.index)(1);
                    T sigma=1;
                    T mean=grid.domain.Center()(0);
                    T a=0.5;
                    gamma(it.Cell_Index())=a*exp(-0.5*sqr((x-mean)/sigma))*exp(-0.5*sqr((y-mean)/sigma));
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 11: // Gamma in whole field (random or default) (constant v)
            {
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T g=random_gamma ? (T).5+(T)0.1*(random.Get_Number()-(T)0.5) : default_gamma;
                    gamma(it.Cell_Index())=g;
                }
                TV half_edge=grid.domain.Edge_Lengths()/2;
                for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    face_velocities(it.face)=1;
                }
                pressure.Fill(0);
                break;
            }
        case 12: // Gamma in whole field (cosine field) (cos*sin velocity)
            {
                TV full_edge=grid.domain.Edge_Lengths();
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    gamma(it.index)=((cos(2*pi/full_edge(0)*grid.Cell_X(it.index)(0)-pi)+1)/2)*((cos(2*pi/full_edge(1)*grid.Cell_X(it.index)(1)-pi)+1)/2);
                }

                for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(it.face.axis==0)
                        face_velocities(it.face)=1*sin((2*pi/full_edge(0))*grid.Face_X(it.face)(0))*cos((2*pi/full_edge(1))*grid.Face_X(it.face)(1));
                    else if(it.face.axis==1)
                        face_velocities(it.face)=-1*cos((2*pi/full_edge(0))*grid.Face_X(it.face)(0))*sin((2*pi/full_edge(1))*grid.Face_X(it.face)(1));
                }
                pressure.Fill(0);
                driver.slip_boundary=true;
                break;
            }
        case 13: // Gamma in whole field (random or default) (cos*sin velocity)
            {
                TV full_edge=grid.domain.Edge_Lengths();
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T g=random_gamma ? (T).5+(T)0.1*(random.Get_Number()-(T)0.5) : default_gamma;
                    gamma(it.Cell_Index())=g;
                }

                full_edge*=1;
                T shift=0;
                // full_edge*=2; // *2 to make the width of the spiral twice as large
                // T shift=pi/2; // pi/2 to center on origin
                for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(it.face.axis==0)
                        face_velocities(it.face)=1*sin((2*pi/full_edge(0))*grid.Face_X(it.face)(0)+shift)*cos((2*pi/full_edge(1))*grid.Face_X(it.face)(1)+shift);
                    else if(it.face.axis==1)
                        face_velocities(it.face)=-1*cos((2*pi/full_edge(0))*grid.Face_X(it.face)(0)+shift)*sin((2*pi/full_edge(1))*grid.Face_X(it.face)(1)+shift);
                }
                pressure.Fill(0);
                driver.slip_boundary=true;
                break;
            }
        case 14: // Gamma in whole field (checkerboard) (cos*sin velocity)
            {

                TV full_edge=grid.domain.Edge_Lengths();
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    gamma(it.index)=((int)grid.Cell_X(it.index)(0))%2;
                }

                for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(it.face.axis==0)
                        face_velocities(it.face)=sin((2*pi/full_edge(0))*grid.Face_X(it.face)(0))*cos((2*pi/full_edge(1))*grid.Face_X(it.face)(1));
                    else if(it.face.axis==1)
                        face_velocities(it.face)=-cos((2*pi/full_edge(0))*grid.Face_X(it.face)(0))*sin((2*pi/full_edge(1))*grid.Face_X(it.face)(1));
                }
                pressure.Fill(0);
                driver.slip_boundary=true;
                break;
            }
        case 15: // Periodic advection test
            {
                for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                    face_velocities(it.face)=1;
                
                SPHERE<TV> ball(TV()+0.5*grid.domain.Edge_Lengths(),(T)0.1);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=-ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=(1+tanh(distance/driver.ch->eta))/2;
                    gamma(it.index)=(max_gamma-min_gamma)*profile+min_gamma;
                }
                pressure.Fill(0);
                break;
            }
        case 16: // Air Bubble Test
            {
                driver.side_bc_type(0)=CHNS_DRIVER<TV>::BC_SLIP;
                driver.side_bc_type(1)=CHNS_DRIVER<TV>::BC_SLIP;
                driver.side_bc_type(2)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(3)=CHNS_DRIVER<TV>::BC_NOSLIP;
                SPHERE<TV> ball(TV()+0.5,(T)0.25);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=(1+tanh(distance/driver.ch->eta))/2;
                    gamma(it.index)=(max_gamma-min_gamma)*profile+min_gamma;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 17: // Rayleigh-Taylor Test
            {
                driver.side_bc_type(0)=CHNS_DRIVER<TV>::BC_PERIODIC;
                driver.side_bc_type(1)=CHNS_DRIVER<TV>::BC_PERIODIC;
                driver.side_bc_type(2)=CHNS_DRIVER<TV>::BC_SLIP;
                driver.side_bc_type(3)=CHNS_DRIVER<TV>::BC_SLIP;
                
                auto boundary_func=[&grid](TV X){
                    return 0.1*cos(2*pi*X.x)+grid.domain.max_corner(1)/2;
                };
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    TV X=grid.Cell_X(it.index);
                    if(X.y>boundary_func(X)) gamma(it.index)=max_gamma;
                    else gamma(it.index)=min_gamma;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 18: // Dam Break Test
            {
                RANGE<TV> water(grid.domain);
                water.max_corner*=0.25;
                water.max_corner(1)*=2;

                for(int side=0;side<TV::m*2;side++)
                    driver.side_bc_type(side)=CHNS_DRIVER<TV>::BC_NOSLIP;
                
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    TV X=grid.Cell_X(it.index);
                    if(water.Lazy_Inside(X)) gamma(it.index)=max_gamma;
                    else gamma(it.index)=min_gamma;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 19: // Rising Bubble Test
            {
                TV center;
                center.y=0.5;
                SPHERE<TV> ball(center,(T)0.25);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=-ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=(1+tanh(distance/driver.ch->eta))/2;
                    gamma(it.index)=profile;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 20: // Steady Drop Test (Hyperbolic Tangent)
            {
                driver.side_bc_type(0)=CHNS_DRIVER<TV>::BC_SLIP;
                driver.side_bc_type(1)=CHNS_DRIVER<TV>::BC_SLIP;
                driver.side_bc_type(2)=CHNS_DRIVER<TV>::BC_SLIP;
                driver.side_bc_type(3)=CHNS_DRIVER<TV>::BC_SLIP;
                SPHERE<TV> ball(TV()+0.5,(T)0.2);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=-ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=(1+tanh(distance/driver.ch->eta))/2;
                    gamma(it.index)=(max_gamma-min_gamma)*profile+min_gamma;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 21: // Horizontal Shear Test
            {
                driver.side_bc_type.Fill(CHNS_DRIVER<TV>::BC_PERIODIC);
                driver.periodic.Fill(true);
                T y0=0.5;
                T y1=0.25;
                T y2=0.75;
                // T delta1=(T)1/30;
                T delta1=sqrt(2)*driver.ch->eta;
                T delta2=0.05;
                T k=2*pi;
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T y=grid.Cell_X(it.index)(1);
                    T profile=min_gamma;
                    if(y<=y0) profile=(1+tanh((y-y1)/delta1))/2;
                    else profile=(1+tanh((y2-y)/delta1))/2;
                    gamma(it.index)=(max_gamma-min_gamma)*profile+min_gamma;
                }
                pressure.Fill(0);
                for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T x=grid.Face_X(it.face)(0);
                    T y=grid.Face_X(it.face)(1);
                    if(it.face.axis==0)
                    {
                        if(y<=y0) face_velocities(it.face)=tanh((y-y1)/delta1);
                        else face_velocities(it.face)=tanh((y2-y)/delta1);
                    }
                    else
                    {
                        face_velocities(it.face)=delta2*sin(k*x);
                    }
                }
                break;
            }
        case 22: // Steady Drop Test (cubic)
            {
                driver.side_bc_type(0)=CHNS_DRIVER<TV>::BC_SLIP;
                driver.side_bc_type(1)=CHNS_DRIVER<TV>::BC_SLIP;
                driver.side_bc_type(2)=CHNS_DRIVER<TV>::BC_SLIP;
                driver.side_bc_type(3)=CHNS_DRIVER<TV>::BC_SLIP;
                SPHERE<TV> ball(TV()+0.5,(T)0.2);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=-ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=0;
                    T range=2*driver.ch->eta;
                    if(distance>range) profile=1;
                    else if(distance<-range) profile=0;
                    else profile=clamp(0.25*pow(-distance,(T)3)/pow(range,(T)3)-0.75*(-distance)/range+0.5,(T)0,(T)1);
                    gamma(it.index)=profile;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 23: // Periodic advection test (random)
            {
                for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                    face_velocities(it.face)=1;
                
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T g=random_gamma ? random.Get_Uniform_Number(0,1) : default_gamma;
                    gamma(it.Cell_Index())=g;
                }
                pressure.Fill(0);
                break;
            }
        case 24: // Droplet and Pool
            {
                RANGE<TV> pool(grid.domain);
                pool.max_corner(1)*=0.25;

                TV center;
                center.x=0.5;
                center.y=0.7;
                SPHERE<TV> droplet(center,(T)0.2);

                for(int side=0;side<TV::m*2;side++)
                    driver.side_bc_type(side)=CHNS_DRIVER<TV>::BC_NOSLIP;
                
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    TV X=grid.Cell_X(it.index);
                    if(pool.Lazy_Inside(X)) gamma(it.index)=1;
                    else
                    {
                        T distance=-droplet.Signed_Distance(grid.Cell_X(it.index));
                        T profile=0;
                        T range=2*driver.ch->eta;
                        if(distance>range) profile=1;
                        else if(distance<-range) profile=0;
                        else profile=clamp(0.25*pow(-distance,(T)3)/pow(range,(T)3)-0.75*(-distance)/range+0.5,(T)0,(T)1);
                        gamma(it.index)=profile;
                    }
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 25: // Horizontal Shear Test (test)
            {
                driver.side_bc_type.Fill(CHNS_DRIVER<TV>::BC_PERIODIC);
                driver.periodic.Fill(true);
                T y0=0.5;
                T y1=0.25;
                T y2=0.75;
                T delta1=(T)1/30;
                T delta2=0.05;
                T k=2*pi;
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T y=grid.Cell_X(it.index)(1);
                    if(y<=y0) gamma(it.index)=(1+tanh((y-y1)/delta1))/2;
                    else gamma(it.index)=(1+tanh((y2-y)/delta1))/2;
                }
                pressure.Fill(0);
                for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T x=grid.Face_X(it.face)(0);
                    T y=grid.Face_X(it.face)(1);
                    if(it.face.axis==0)
                    {
                        if(y<=y0) face_velocities(it.face)=tanh((y-y1)/delta1);
                        else face_velocities(it.face)=tanh((y2-y)/delta1);
                    }
                    else
                    {
                        face_velocities(it.face)=delta2*sin(k*x);
                    }
                }
                gamma.Fill(1);
                break;
            }
        case 26: // Single Sphere - volume test
            {
                TV b1c;
                b1c(0)=(T)0.5;
                b1c(1)=(T)0.5;
                SPHERE<TV> ball1(b1c,(T)0.1);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(ball1.Lazy_Inside(grid.Cell_X(it.index)))
                    {
                        gamma(it.index)=1;
                        driver.volume(it.index)=1;
                    }
                    else
                    {
                        gamma(it.index)=0;
                        driver.volume(it.index)=0.9;
                    }
                }
                if(driver.chh) driver.chh->volume=&driver.volume;
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 27: // UnEqual Spheres - volume test
            {
                TV b1c,b2c,b3c,b4c;
                b1c(0)=(T)0.25;
                b1c(1)=(T)0.25;
                b2c(0)=(T)0.75;
                b2c(1)=(T)0.75;
                b3c(0)=(T)0.75;
                b3c(1)=(T)0.25;
                b4c(0)=(T)0.30;
                b4c(1)=(T)0.75;
                SPHERE<TV> ball1(b1c,(T)0.1);
                SPHERE<TV> ball2(b2c,(T)0.1);
                SPHERE<TV> ball3(b3c,(T)0.1);
                SPHERE<TV> ball4(b4c,(T)0.1);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(ball1.Lazy_Inside(grid.Cell_X(it.index)) || ball2.Lazy_Inside(grid.Cell_X(it.index)) || ball3.Lazy_Inside(grid.Cell_X(it.index)) || ball4.Lazy_Inside(grid.Cell_X(it.index)))
                    {
                        gamma(it.index)=1;
                        driver.volume(it.index)=1;
                    }
                    else
                    {
                        gamma(it.index)=0;
                        driver.volume(it.index)=0.9;
                    }
                }
                driver.chh->volume=&driver.volume;
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 28: // two different Sphere - volume test extreme
            {
                TV b1c,b2c;
                b1c(0)=(T)0.5;
                b1c(1)=(T)0.5;
                b2c(0)=(T)0.6;
                b2c(1)=(T)0.6;
                SPHERE<TV> ball1(b1c,(T)0.1);
                SPHERE<TV> ball2(b2c,(T)0.1);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    if(ball1.Lazy_Inside(grid.Cell_X(it.index)))
                    {
                        gamma(it.index)=1;
                        driver.volume(it.index)=1-(grid.Cell_X(it.index)-b1c).Magnitude()/0.1;
                    }
                    else if(ball2.Lazy_Inside(grid.Cell_X(it.index)))
                    {
                        gamma(it.index)=0;
                        driver.volume(it.index)=1-(grid.Cell_X(it.index)-b2c).Magnitude()/0.1;
                    }
                    else
                    {
                        gamma(it.index)=0.5;
                        driver.volume(it.index)=0;
                    }
                }
                driver.chh->volume=&driver.volume;
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 29: // two different Sphere - volume test same
            {
                TV b1c,b2c;
                b1c(0)=(T)0.5;
                b1c(1)=(T)0.5;
                b2c(0)=(T)0.6;
                b2c(1)=(T)0.6;
                SPHERE<TV> ball1(b1c,(T)0.1);
                SPHERE<TV> ball2(b2c,(T)0.1);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    driver.volume(it.index)=1;
                    if(ball1.Lazy_Inside(grid.Cell_X(it.index)))
                    { gamma(it.index)=1; }
                    else if(ball2.Lazy_Inside(grid.Cell_X(it.index)))
                    { gamma(it.index)=0; }
                    else { gamma(it.index)=0.5; }
                }
                driver.chh->volume=&driver.volume;
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 30: // Gamma in whole field (1D random)
            {
                IV min_corner=grid.Cell_Indices().min_corner;
                IV max_corner=grid.Cell_Indices().max_corner;
                for(int x=min_corner(0);x<max_corner(0);x++)
                {
                    T g=random_gamma ? (T).5+random_scale*(random.Get_Number()-(T)0.5) : default_gamma;
                    for(int y=min_corner(1);y<max_corner(1);y++)
                    {
                        IV index;
                        index(0)=x;
                        index(1)=y;
                        gamma(index)=g;
                    }
                }

                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 31:
            {
                driver.side_bc_type.Fill(CHNS_DRIVER<TV>::BC_PERIODIC);
                driver.periodic.Fill(true);
                T y0=0.3;
                T y1=0.7;
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T y=grid.Cell_X(it.index)(1);
                    if(y>y0&&y<y1) gamma(it.index)=max_gamma;
                    else gamma(it.index)=min_gamma;
                }
                pressure.Fill(0);
                for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T y=grid.Face_X(it.face)(1);
                    if(it.face.axis==0)
                    {
                        if(y>y0&&y<y1) face_velocities(it.face)=0;
                        else face_velocities(it.face)=1;
                    }
                    else face_velocities(it.face)=0;
                }
                break;
            }
        case 32: // Contact angles
            {
                driver.side_bc_type(0)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(1)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(2)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(3)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.ch->side_contact_angle(0)=90; // -x
                driver.ch->side_contact_angle(1)=90; // +x
                driver.ch->side_contact_angle(2)=90; // -y not used?
                driver.ch->side_contact_angle(3)=90; // +y
                custom_contact=true;

                TV center;
                center.y=grid.domain.Edge_Lengths()(1)/3;
                SPHERE<TV> ball(center,grid.domain.Edge_Lengths()(0)/4);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=(1+tanh(distance/driver.ch->eta))/2;
                    gamma(it.index)=(max_gamma-min_gamma)*profile+min_gamma;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 33: // Contact angles
            {
                driver.side_bc_type(0)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(1)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(2)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(3)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.ch->side_contact_angle(0)=150; // -x
                driver.ch->side_contact_angle(1)=30; // +x
                driver.ch->side_contact_angle(2)=90; // -y not used?
                driver.ch->side_contact_angle(3)=90; // +y
                custom_contact=true;
                
                TV center;
                center.y=grid.domain.Edge_Lengths()(1)/3;
                SPHERE<TV> ball(center,grid.domain.Edge_Lengths()(0)/4);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=(1+tanh(distance/driver.ch->eta))/2;
                    gamma(it.index)=(max_gamma-min_gamma)*profile+min_gamma;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 34: // Contact angles
            {
                driver.side_bc_type(0)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(1)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(2)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(3)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.ch->side_contact_angle(0)=120; // -x
                driver.ch->side_contact_angle(1)=150; // +x
                driver.ch->side_contact_angle(2)=90; // -y not used?
                driver.ch->side_contact_angle(3)=90; // +y
                custom_contact=true;
                
                TV center;
                center.y=grid.domain.Edge_Lengths()(1)/3;
                SPHERE<TV> ball(center,grid.domain.Edge_Lengths()(0)/4);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=(1+tanh(distance/driver.ch->eta))/2;
                    gamma(it.index)=(max_gamma-min_gamma)*profile+min_gamma;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 35: // Contact angles
            {
                driver.side_bc_type(0)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(1)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(2)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(3)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.ch->side_contact_angle(0)=60; // -x
                driver.ch->side_contact_angle(1)=30; // +x
                driver.ch->side_contact_angle(2)=90; // -y not used?
                driver.ch->side_contact_angle(3)=120; // +y
                custom_contact=true;
                
                TV center;
                center.y=grid.domain.Edge_Lengths()(1)/3;
                SPHERE<TV> ball(center,grid.domain.Edge_Lengths()(0)/4);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=(1+tanh(distance/driver.ch->eta))/2;
                    gamma(it.index)=(max_gamma-min_gamma)*profile+min_gamma;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        case 36: // Equilibrium Drop
            {
                driver.side_bc_type(0)=CHNS_DRIVER<TV>::BC_PERIODIC;
                driver.side_bc_type(1)=CHNS_DRIVER<TV>::BC_PERIODIC;
                driver.side_bc_type(2)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.side_bc_type(3)=CHNS_DRIVER<TV>::BC_NOSLIP;
                driver.periodic.Fill(true);
                driver.periodic(1)=false;

                TV center;
                // center.y=grid.domain.Edge_Lengths()(1)/3;
                SPHERE<TV> ball(center,1);
                for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
                {
                    T distance=-ball.Signed_Distance(grid.Cell_X(it.index));
                    T profile=(1+tanh(distance/driver.ch->eta))/2;
                    gamma(it.index)=(max_gamma-min_gamma)*profile+min_gamma;
                }
                pressure.Fill(0);
                face_velocities.Fill(0);
                break;
            }
        default:
            PHYSBAM_FATAL_ERROR("Invalid Test Number");          
    }
    if(invert_gamma) gamma.array=1-gamma.array;
    if(!custom_contact) driver.ch->side_contact_angle.Fill(pi/2);
    else
    {
        for(int i=0;i<TV::m*2;i++)
            driver.ch->side_contact_angle(i)=driver.ch->side_contact_angle(i)*pi/180;
    }
}
