//##############################################################################
// August 2023 - Mehrnaz Ayazi, Jason Goulding
//##############################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>
#include <fstream>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;

std::string mode="";
std::string fig_dir;
ARRAY<std::string> dir;
ARRAY<int> frame;
//#####################################################################
// Function Dump_Velocity_data
//#####################################################################
void Dump_Velocity_data()
{
    std::string u_L2="\\multirow{2}{*}{u} & $L_2$";
    std::string u_Linf=" & $L_{\\infty}$";
    std::string v_L2="\\multirow{2}{*}{v} & $L_2$";
    std::string v_Linf=" & $L_{\\infty}$";
    for(int i=0;i<dir.m;i++)
    {        
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);

        viewer_dir.Set(frame(i));
        u_Linf+=" & ";
        v_Linf+=" & ";
        u_L2+=" & ";
        v_L2+=" & ";
        if(viewer_dir.Valid_Frame(true))
        {
            ARRAY<T,FACE_INDEX<TV::m> > velocity;
            Read_From_File(viewer_dir.Frame_File("velocities"),velocity);

            TV L2_error;
            TV Linf_error;
            for(FACE_ITERATOR<TV> it(grid,0,INTERIOR_REGION);it.Valid();it.Next())
            {
                T e=velocity(it.face)-1;
                L2_error(it.face.axis)+=sqr(e)*grid.Cell_Size();
                Linf_error(it.face.axis)=max(abs(e),Linf_error(it.face.axis));
            }
            for(int axis=0;axis<TV::m;axis++) // Always periodic
            {
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis);it.Valid();it.Next())
                {
                    T e=velocity(it.face)-1;
                    L2_error(it.face.axis)+=sqr(e)*grid.Cell_Size();
                    Linf_error(it.face.axis)=max(abs(e),Linf_error(it.face.axis));
                }
            }

            u_Linf+=LOG::sprintf("%.2e",Linf_error.x);
            v_Linf+=LOG::sprintf("%.2e",Linf_error.y);
            u_L2+=LOG::sprintf("%.2e",sqrt(L2_error.x));
            v_L2+=LOG::sprintf("%.2e",sqrt(L2_error.y));
        }
        else LOG::printf("ERROR: Invalid Frame: %P, %P\n",frame(i),fig_dir+"/"+dir(i));
    }
    std::string filename=fig_dir+"/tex/advection_table_"+mode+".tex";
    LOG::cout<<"creating tex file "<<filename<<std::endl;
    std::ofstream stream(filename);
    std::string begin= "\\begin{tabular}{| c | c | c | c | c | c |}\n "
                       "\\hline \n "
                       "Variable & Norm & $\\rho_1/\\rho_0=1$  & $\\rho_1/\\rho_0=10^3$ & $\\rho_1/\\rho_0=10^6$ & $\\rho_1/\\rho_0=10^{9}$  \\\\ \n "
                       "\\hline";
    LOG::fprintf(stream,"%P",begin);
    LOG::fprintf(stream,"%P \\\\ \n",u_L2);
    LOG::fprintf(stream,"%P \\\\ \n",u_Linf);
    LOG::fprintf(stream,"\\hline \n");
    LOG::fprintf(stream,"%P \\\\ \n",v_L2);
    LOG::fprintf(stream,"%P \\\\ \n",v_Linf);
    LOG::fprintf(stream,"\\hline \n");
    LOG::fprintf(stream," \\end{tabular}");
}
//#####################################################################
// Function main
//#####################################################################
int main(int argc,char *argv[])
{
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    // LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;

    VECTOR<T,2> range_scale={0,1};
    std::string boundary_file="";

    parse_args.Add("-dir",&dir,"string-array","directory name to read data from");
    parse_args.Add("-frame",&frame,"int-array","frames to gather data (one per directory)");
    parse_args.Add("-mode",&mode,"string","category of test");//With or without surface tension
    parse_args.Add("-fig",&fig_dir,"dir","base figure directory");

    parse_args.Parse();
    Dump_Velocity_data();
    // LOG::Finish_Logging();
    return 0;
}
