//##############################################################################
// August 2023 - Mehrnaz Ayazi, Jason Goulding
//##############################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Basic_Geometry/SPHERE.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <fstream>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;

std::string fig_dir;
ARRAY<std::string> dir;
ARRAY<int> frame;
bool min_shift=false;
bool left_shift=false;
bool ideal=false;
T r=0.1;
T sigma=1;
//#####################################################################
// Function Dump_Pressure_data
//#####################################################################
void Dump_Ideal_Pressure_data()
{
    std::string filenamex="pressure_ideal_x.txt";
    LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filenamex<<std::endl;
    std::ofstream streamx(fig_dir+"/txt/"+filenamex);
    LOG::fprintf(streamx,"x pressure\n");

    // std::ofstream streamy;
    // std::string filenamey="pressure_ideal_y.txt";
    // LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filenamey<<std::endl;
    // streamy.open(fig_dir+"/txt/"+filenamey);
    // LOG::fprintf(streamy,"y pressure\n");
    
    GRID<TV> grid(IV()+128,RANGE<TV>::Unit_Box());
    RANGE<IV> domain=grid.Cell_Indices();
    ARRAY<T,IV> pressure(domain);
    SPHERE<TV> drop(grid.domain.Center(),r);
    T p_jump=sigma/r;
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        if(drop.Lazy_Inside(grid.Cell_X(it.index))) pressure(it.index)=p_jump;
    }

    int half_y=domain.Center()(0);
    for(int i=domain.min_corner(0);i<domain.max_corner(0);i++)
    {
        IV index={i,half_y};
        T x=grid.Cell_X(index)(0);
        T p=pressure(index);
        LOG::fprintf(streamx,"%P %.16P\n",x,p);
    }

    // int half_x=domain.Center()(1);
    // for(int i=domain.min_corner(1);i<domain.max_corner(1);i++)
    // {
    //     IV index={half_x,i};
    //     T y=grid.Cell_X(index)(1);
    //     T p=pressure(index);
    //     LOG::fprintf(streamy,"%P %.16P\n",y,p);
    // }
}
//#####################################################################
// Function Dump_Pressure_data
//#####################################################################
void Dump_Pressure_data()
{
    for(int i=0;i<dir.m;i++)
    {
        std::string filenamex="pressure_"+dir(i)+"_x.txt";
        LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filenamex<<std::endl;
        std::ofstream streamx(fig_dir+"/txt/"+filenamex);
        LOG::fprintf(streamx,"x pressure\n");

        // std::ofstream streamy;
        // std::string filenamey="pressure_"+dir(i)+"_y.txt";
        // LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filenamey<<std::endl;
        // streamy.open(fig_dir+"/txt/"+filenamey);
        // LOG::fprintf(streamy,"y pressure\n");
        
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);

        viewer_dir.Set(frame(i));
        if(viewer_dir.Valid_Frame(true))
        {
            ARRAY<T,IV> pressure;
            Read_From_File(viewer_dir.Frame_File("pressure"),pressure);
            T min_p=1e10;
            // for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
            // {
            //     T p=pressure(it.index);
            //     min_p=p<min_p?p:min_p;
            // }

            RANGE<IV> domain=grid.Cell_Indices();
            int half_y=domain.Center()(0);
            if(left_shift)
            {
                IV index={0,half_y};
                min_p=pressure(index);
            }
            else if(min_shift)
            {
                for(int i=domain.min_corner(0);i<domain.max_corner(0);i++)
                {
                    IV index={i,half_y};
                    T p=pressure(index);
                    min_p=p<min_p?p:min_p;
                }
            }
            else min_p=0;
            for(int i=domain.min_corner(0);i<domain.max_corner(0);i++)
            {
                IV index={i,half_y};
                T x=grid.Cell_X(index)(0);
                T p=pressure(index)-min_p;
                LOG::fprintf(streamx,"%P %.16P\n",x,p);
            }

            // int half_x=domain.Center()(1);
            // for(int i=domain.min_corner(1);i<domain.max_corner(1);i++)
            // {
            //     IV index={half_x,i};
            //     T y=grid.Cell_X(index)(1);
            //     T p=pressure(index)-(min_shift?min_p:0);
            //     LOG::fprintf(streamy,"%P %.16P\n",y,p);
            // }
        }
        else LOG::printf("ERROR: Invalid Frame: %P, %P\n",frame(i),fig_dir+"/"+dir(i));
    }
}
//#####################################################################
// Function main
//#####################################################################
int main(int argc,char *argv[])
{
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    // LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;

    VECTOR<T,2> range_scale={0,1};
    std::string boundary_file="";

    parse_args.Add("-dir",&dir,"string-array","directory name to read data from");
    parse_args.Add("-frame",&frame,"int-array","frames to gather data (one per directory)");
    parse_args.Add("-min_shift",&min_shift,"shift pressure up by minimum pressure value");
    parse_args.Add("-left_shift",&left_shift,"shift pressure up by leftmost pressure value");
    parse_args.Add("-ideal",&ideal,"compute ideal pressure values");
    parse_args.Add("-r",&r,"scalar","drop radius for ideal pressure");
    parse_args.Add("-sigma",&sigma,"scalar","surface tension coefficient for ideal pressure");
    parse_args.Add("-fig",&fig_dir,"dir","base figure directory");

    parse_args.Parse();
    if(ideal) Dump_Ideal_Pressure_data();
    else Dump_Pressure_data();
    // LOG::Finish_Logging();
    return 0;
}
