//##############################################################################
// August 2023 - Mehrnaz Ayazi, Jason Goulding
//##############################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Grids/GRID.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>
#include <fstream>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;

AVERAGING_UNIFORM<TV> avg;
std::string fig_dir;
ARRAY<std::string> dir;
ARRAY<int> frame;
T threshold=1e-6;
//#####################################################################
// Function Dump_Surface_Tension_data
//#####################################################################
void Dump_Surface_Tension_data()
{
    for(int i=0;i<dir.m;i++)
    {
        std::string filename_mag="st_mag_"+dir(i)+".txt";
        LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filename_mag<<std::endl;
        std::ofstream stream_mag(fig_dir+"/txt/"+filename_mag);

        std::string filename_angle="st_angle_"+dir(i)+".txt";
        LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filename_angle<<std::endl;
        std::ofstream stream_angle(fig_dir+"/txt/"+filename_angle);
        
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);

        viewer_dir.Set(frame(i));
        if(viewer_dir.Valid_Frame(false))
        {
            ARRAY<T,FACE_INDEX<TV::m> > st_force;
            ARRAY<T,FACE_INDEX<TV::m> > grad_gamma;
            Read_From_File(viewer_dir.Frame_File("st_force"),st_force);
            Read_From_File(viewer_dir.Frame_File("grad_gamma"),grad_gamma);

            RANGE<IV> domain=grid.Cell_Indices();
            TV center=grid.domain.Center();
            for(int i=domain.min_corner.x;i<domain.max_corner.x;i++)
            {
                for(int j=domain.min_corner.y;j<domain.max_corner.y;j++)
                {
                    IV index={i,j};
                    TV X=grid.Cell_X(index);

                    TV st=avg.Face_To_Cell_Vector(grid,index,st_force);
                    TV norm=-(X-center);
                    // TV grad=avg.Face_To_Cell_Vector(grid,index,grad_gamma);
                    // T angle=(st.Magnitude_Squared()<threshold||grad.Magnitude_Squared()<threshold)?0:TV::Angle_Between(st,grad);
                    T angle=(st.Magnitude_Squared()<threshold||norm.Magnitude_Squared()<threshold)?0:st.Angle_Between(norm);

                    LOG::fprintf(stream_mag,"%P %P %.16P\n",X.x,X.y,st.Magnitude());
                    LOG::fprintf(stream_angle,"%P %P %.16P\n",X.x,X.y,angle);
                }
                LOG::fprintf(stream_mag,"\n");
                LOG::fprintf(stream_angle,"\n");
            }
        }
        else LOG::printf("ERROR: Invalid Frame: %P, %P\n",frame(i),fig_dir+"/"+dir(i));
    }
}
//#####################################################################
// Function main
//#####################################################################
int main(int argc,char *argv[])
{
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    // LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;

    VECTOR<T,2> range_scale={0,1};
    std::string boundary_file="";

    parse_args.Add("-dir",&dir,"string-array","directory name to read data from");
    parse_args.Add("-frame",&frame,"int-array","frames to gather data (one per directory)");
    parse_args.Add("-threshold",&threshold,"scalar","ignore gamma gradients smaller than this value");
    parse_args.Add("-fig",&fig_dir,"dir","base figure directory");

    parse_args.Parse();
    Dump_Surface_Tension_data();
    // LOG::Finish_Logging();
    return 0;
}
